//===========================================================================
//
// Port of selected old ISC Clunkdraw routines.  Those C routines were
// in turn based on the Pascal imaging model of the original Macintosh,
// so in a way things have come full circle in 30+ years.
//
// An 8 MHz 68010 is not really all that fast, and heavy optimization
// was required for the orginal Clunkdraw, including an extreme
// preference for assembly language.  (Just like the Macintosh, and
// for all the same reasons.)  This is not necessary here, both
// because the RPi's ARM is quite fast, and because we simply don't
// need to draw that much at a time here.  Clunkdraw took several years
// to develop.  I spent only a few days on this...
//
// In other words the implementation is naive, the performance is
// poor, and that simply doesn't matter here.  We only have monochrome,
// no colors or gray scale.
//
// Clunkdraw genesis:  Start with the elegant and efficient imaging
// model of the original Macintosh's Quickdraw, whose implementation
// validated the design on machines of that class.  (Which the first
// Clunkdraw target also was.)  Implement that design in a mix of C
// and assembly language, on both 68010 (monochrome) and 34010 (color)
// processors, over the course of years.  The early versions were
// extremely clunky, which inspired the name.  The bulk of the time was
// spent making the assembly language parts fast, and bug-free.  Let
// decades go by, long past the time when the Clunkdraw products had
// become obsolete.  Enter a need for good-looking monochrome graphics
// output on an e-Paper peripheral, one that is _not_ integrated into
// the Raspberry Pi host GUI system.  Find the supplied demonstration
// programs disgusting, particularly the text, and remember fondly the
// very nice monochrome text abilities of Quickdraw, and Clunkdraw.  Dust
// off Clunkdraw, and port to Pascal, quickly.  Count on the fast ARM
// processor and small screen size to make up for the lack of the
// high-speed assembly-language parts.  Do the job in days, not years.
//
//===========================================================================

//{$DEFINE DEBUG}

unit clunkdraw;

{$mode objfpc}{$H+}

interface

type
   OSErr = integer;

// Transfer modes, used by font and pattern operations.

// Every time a pixel is modified by Clunkdraw, a "transfer mode" (either
// explicit or implicit) is used to determine the new state of the pixel
// based on its input(s).  The transfer mode is normally one of eight Boolean
// (logical) operations.  For each bit of the item to be drawn the transfer
// mode is applied, with this bit and the bit already at the appropriate
// spot of the destination as inputs.  The result is then placed into the
// bit image at the destination.
//
// There are four basic operations --- COPY, OR, XOR, and BIC ('bit
// clear').  The COPY operation merely replaces the pixels of the
// destination with the pixels of the source, tromping over the destination
// without regard for what was already there.  The OR, XOR, and BIC
// operations leave pixels that lie underneath the white part of the source
// (0 bits) unchanged, and affect pixels that lie underneath the black part
// of the source (1 bits) as follows: OR replaces these pixels with
// foreground pixels (nominally black), thus 'overlaying' the destination
// with the black parts of the source; XOR inverts (switches foreground
// pixels to background pixels and vice versa) the pixels under the black
// part; and BIC erases them to background pixels (nominally white).
//
// Each of the basic four operations has a variant in which every bit of
// the source is inverted before the operation is performed.  Thus, there
// are eight operations in all.  Each mode is defined by name:
//
//                               Result
//      Transfer Mode  Black Source  White Source
//      -------------  ------------  ------------
//       srcCopy       Force Black   Force White
//       srcOr         Force Black   Leave Alone
//       srcXor        Invert        Leave Alone
//       srcBic        Force White   Leave Alone
//       notSrcCopy    Force White   Force Black
//       notSrcOr      Leave Alone   Force Black
//       notSrcXor     Leave Alone   Invert
//       notSrcBic     Leave Alone   Force White
//
// Where 'Black' and 'White' are used in the tables above it is more
// correct to speak of Foreground and Background, respectively.  'Invert'
// swaps foreground and background.  However, this distinction doesn't really
// matter in a non-color environment.
//
   xferMode = (srcCopy,    // 0 Copy from source into window.
               srcOr,      // 1 Or from source into window.
               srcXor,     // 2 Xor from source into window.
               srcBic,     // 3 Clear bits set in source in window.
               notSrcCopy, // 4 Copy inverse of source into window.
               notSrcOr,   // 5 Or inverse of source into window.
               notSrcXor,  // 6 Xor inverse of source into window.
               notSrcBic); // 7 Clear bits with zero in source in window.

   ModeStore = packed record
      case Boolean of
        False : (e : xferMode);
        True  : (i : smallInt); // Sets the size of the union.
   end;

// Text styles.

// normal
//     The normal, unembellished style.  (The absence of all other styles.)
//     It is considerably faster in drawing than any other style (especially
//     when combined with the srcOr transfer mode).
//
// bold
//     Each character is 'smeared' one bit to the right.  Bold characters
//     are thus wider (and bolder) than normal.  Small characters, when
//     bolded, may smear into blobs.  You get better results by using a
//     font that has been designed to be bold, rather than applying a bold
//     style to a normal font.  As font sizes get larger, the bolding
//     effect diminishes.
//
// fixedBold
//     The same as bold, but the character width is forced to remain
//     the same as normal.  Primarily for use with fixed-pitch fonts when
//     emulating older ASCII (or EBCDIC) terminals.  Besides the risk of
//     normal bold blobbing, wide characters can smear into each other and
//     turn into even larger illegible blobs.
//
// underline
//     Draws a line below the base line of the characters.  Pixels in the
//     underline will be eliminated wherever they would touch a descender
//     or the baseline pixels of the font itself.  The latter case only
//     happens on particularly small fonts.
//
// italic
//     Adds an italic (actually an oblique) slant to the characters.
//     Character bits above the base line will lean to the right, and
//     those below will shift to the left.  Because of this lean, if the
//     last character is particularly tall it may hang out farther into
//     space than you might have thought, and can even run into
//     non-italicized characters that might follow.
//
// ghosted
//     Makes the text look 'ghostly', by knocking out every other pixel.
//     The results can be particularly illegible unless you are using a
//     thick-stroked font (one that has pixels to spare), or you've called
//     for a bold style as well.
//
// outline
//     Draws a hollow, outlined character rather than a solid one.  The
//     results can be rather poor on small sizes, and get better as the
//     sizes get larger.
//
// shadow
//     Also draws a hollow, outlined character rather than a solid one but
//     the outline is thickened below and to the right of the character to
//     achieve the effect of a shadow.
//
// strikeout
//     Puts a solid strikeout line through the middle of the characters.
//
   StyleItem = (bold, fixedBold, underline, italic, ghosted, outline, shadow, strikeout);
   Style = set of StyleItem;
   StyleStore = packed record
      case Boolean of
        False : (s : Style);
        True  : (i : smallInt); // Sets the size of the union.
   end;

   Point = packed record        // Graphics point structure.
      v,                        // 0: Vertical component.
      h : smallInt;             // 2: Horizontal component.
   end;
   PPoint = ^Point;

   Rect = packed record         // Graphics Rectangle structure.
      case Boolean of
         False :
           (top,                // 0: Topmost coordinate, etc.
            left,               // 2:
            bottom,             // 4:
            right : smallInt);  // 6:
         True :
           (topLeft,            // 0:
            botRight : Point);  // 4:
   end;
   PRect = ^Rect;

   BitMap = packed record       // System Bitmap structure.
       baseAddr : PByte;        // 0: Pointer to bit image.
       rowBytes : smallInt;     // 4: Number of bytes per row of each plane.
       bounds : Rect;           // 6: Logical Bounding rectangle of planes.
   end;

   Fixed = longInt;             // Fixed-point data type.
//
// Actually a Fixed is the following structure.  We just call it a long so that
// it may be passed by _value_ rather than reference, and so that addition,
// subtraction, comparison, and multiplication and division by integers may
// be done directly.
//
// typedef struct {             // Big-endian equivalent structure to Fixed.
//      short Integer;          // Integral part of fixed-point number.
//      short Fraction;         // Fractional part of fixed-point number.
// } Fixed;
//

   Region = packed record
      rgnSize : word;           // Size of total region in bytes (min 10).
      rgnBBox : Rect;           // Bounding box of region.
      rgnData : array [0..0] of smallInt; // Optional region data (may be large).
   end;

   RgnPtr = ^Region;
   RgnHandle =  ^RgnPtr;

   Ptr = Pointer;
   Handle = ^Ptr;

   RegionHdr = packed record
      rgnSize : word;           // Size of total region in bytes (min 10).
      rgnBBox :  Rect;          // Bounding box of region.
   end;

   Polygon = packed record
      polySize   : word;        // Size of total polygon in bytes (min 10).
      polyBBox   : Rect;        // Bounding box of region.
      polyPoints : array [0..0] of Point; // Polygon data points (may be large).
   end;

   PolyPtr = ^Polygon;
   PolyHandle = ^PolyPtr;

   PolyHdr = packed record
      polySize : word;          // Size of total polygon in bytes (min 10).
      polyBBox : Rect;          // Bounding box of region.
   end;

   Pattern = packed record
      bytes : array [0..7] of Byte;
   end;

   Icon = packed record
      data : array [0..127] of Byte;
   end;

   FontInfo = packed record
      ascent : smallInt;        // 0: Height above baseline of font.
      descent : smallInt;       // 2: Height below baseline of font.
      widMax : smallInt;        // 4: Maximum character width.
      leading : smallInt;       // 6: Row-to-row spacing of this font.
   end;

   // A drawing context.
   GrafPort = packed record
      device : smallInt;        //  0: Device specific information. */
      portBits : BitMap;        //  2: grafPort's bit map. */
                                //  2: portBits.baseAddr */
                                //  6: portBits.rowBytes */
                                //  8: portBits.bounds.top */
                                // 10: portBits.bounds.left */
                                // 12: portBits.bounds.bottom */
                                // 14: portBits.bounds.right */
      portRect : Rect;          // 16: grafPort's rectangle. */
                                // 16: portRect.top */
                                // 18: portRect.left */
                                // 20: portRect.bottom */
                                // 22: portRect.right */
      visRgn : RgnHandle;       // 24: Visible region. */
      clipRgn : RgnHandle;      // 28: Clipping region. */
      bkPat : Pattern;          // 32: Background pattern. */
      fillPat : Pattern;        // 40: Fill pattern. */
      pnLoc : Point;            // 48: Pen location. */
                                // 48: pnLoc.v */
                                // 50: pnLoc.h */
      pnSize : Point;           // 52: Pen size. */
                                // 52: pnSize.v */
                                // 54: pnSize.h */
      pnMode : modeStore;       // 56: Pen's transfer mode. */
      pnPat : Pattern;          // 58: Pen pattern. */
      pnVis : smallInt;         // 66: Pen visibility. */
      txFont : smallInt;        // 68: Font number for text. */
      fontPtr : ^FontRec;       // 70: Pointer to font table. */
      txFace : styleStore;      // 74: Text's character style. */
      txMode : modeStore;       // 76: Text's transfer mode. */
      txSize : smallInt;        // 78: Font size for text. */
      spExtra : Fixed;          // 80: Extra space for ASCII space chars. */
      fgColor : smallInt;       // 84: Foreground color. */
      bkColor : smallInt;       // 86: Background color. */
      colrBit : smallInt;       // 88: Color plane(s) in use OR plane mask. */
      txRotate : smallInt;      // 90: Text rotation value, degrees CCW. */
      picSave : Ptr;            // 92: Picture being saved (n/a). */
      rgnSave : RgnHandle;      // 96: Region being saved. */
      polySave : PolyHandle;    // 100: Polygon being saved. */
      txExtra : Fixed;          // 104: Extra space for text chars. */
   end;

   GrafPtr = ^GrafPort;

   PenState = packed record     // A pen state save record.
      pnLoc  : Point;           //  0: Pen's location.
      pnSize : Point;           //  4: Pen's size.
      pnMode : modeStore;       //  8: Pen's drawing mode.
      pnPat  : Pattern;         // 10: Pen's pattern.
   end;

// typedef struct {
//     unsigned short fontType; /*  0: Font type, 1 = ISC 8-bit index, */
//                              /*                2 = ISC 16-bit index, */
//                              /*                others = Macintosh. */
//     unsigned short firstChar;/*  2: ASCII code of first character. */
//     unsigned short lastChar; /*  4: ASCII code of last character. */
//     short widMax;            /*  6: Maximum character width. */
//     short kernMax;           /*  8: Negative of max character kern. */
//     short nDescent;          /* 10: Negative of descent value. */
//     short fRectWidth;        /* 12: Width of font rectangle. */
//     short fRectHeight;       /* 14: Height of font rectangle. */
//     unsigned short owTLoc;   /* 16: Offset (word, from owTLoc) to */
//                              /*     offset/width table. */
//     short ascent;            /* 18: Height above baseline of font. */
//     short descent;           /* 20: Height below baseline of font. */
//     short leading;           /* 22: Row-to-row spacing of this font. */
//     unsigned short rowWords; /* 24: Row-to-row offset (in words) of the */
//                              /*     following source bitmap. */
//                              /* 26: The bitmap, and both tables. Actually */
//                              /*     is variable size, but we can't */
//                              /*     declare that in C. */
// } FontRec;
   FontRec = packed record      // Macintosh FontRec type.  Big-Endian.
      fontType : word;
      firstChar, lastChar : word;
      widMax, kernMax : smallInt;
      nDescent, fRectWidth, fRectHeight : smallInt;
      owTLoc : word;
      ascent, descent, leading : smallInt;
      rowWords : word;
   end;
   PFontRec = ^FontRec;

   FontSizes = 0..256;

   FontCtrl = packed record
      fontNum : smallInt;       // Font family number.
      fontSize : smallInt;      // Size of this font in points.
      fontPtr : ^FontRec;       // Pointer to the memory image.
      memSize : longInt;        // Size of the memory image.
      next : ^FontCtrl;         // Pointer to next in chain.
      used : Boolean;           // TRUE when font's been used since a purge.
   end;

const
   // Pre-defined basic patterns.
   white  : Pattern = (bytes:($00, $00, $00, $00, $00, $00, $00, $00));
   black  : Pattern = (bytes:($FF, $FF, $FF, $FF, $FF, $FF, $FF, $FF));
   gray   : Pattern = (bytes:($AA, $55, $AA, $55, $AA, $55, $AA, $55));
   ltGray : Pattern = (bytes:($88, $22, $88, $22, $88, $22, $88, $22));
   dkGray : Pattern = (bytes:($DD, $77, $DD, $77, $DD, $77, $DD, $77));

   function  FixRound(f : Fixed) : smallInt;
   function  Int2Fix(i : integer) : Fixed;
   function  HiWord(f : Fixed) : word;
   function  Fix2Int(f : Fixed) : smallInt;
   function  LoWord(f : Fixed) : word;
   function  Fix2Frac(f : Fixed) : integer;
   function  FixRatio(dividend, divisor : smallInt) : Fixed;
   function  FixMul(multiplicand, multiplier : Fixed) : Fixed;
   function  StrToFixedU32(s : string; var send : PChar; scale : integer;
                           trunc : Boolean) : longword;
   function  StrToFixedS32(s : string; var send : PChar; scale : integer;
                           trunc : Boolean) : longint;
   function  StrToFixedU64(s : string; var send : PChar; scale : integer;
                           trunc : Boolean) : qword;
   function  StrToFixedS64(s : string; var send : PChar; scale : integer;
                           trunc : Boolean) : int64;
   function  FixedU32ToStr(fixed, scale : dword; dtrunc : integer;
                           commas : Boolean) : string;
   function  FixedU64ToStr(fixed, scale : qword; dtrunc : integer;
                           commas : Boolean) : string;
   function  FixedS32ToStr(fixed : longint; scale, dtrunc : integer;
                           commas : Boolean) : string;
   function  FixedS64ToStr(fixed : int64; scale : qword; dtrunc : integer;
                           commas : Boolean) : string;
   function  MulS(multiplicand, multiplier : smallInt) : longint;
   function  MulU(multiplicand, multiplier : word) : longword;
   function  DivS(dividend : longint; divisor : smallInt) : smallInt;
   function  DivU(dividend : longword; divisor : word) : longword;
   function  DivModS(dividend : longint; divisor : smallInt) : longword;
   function  DivModU(dividend : longword; divisor : word) : longword;
   function  ModS(dividend : longint; divisor : smallInt) : smallInt;
   function  ModU(dividend : longword; divisor : word) : word;
   function  ScaleS(multiplicand, multiplier, divisor : smallInt) : smallInt;
   function  ScaleU(multiplicand, multiplier, divisor : word) : word;
   function  ScaleRS(multiplicand, multiplier, divisor : smallInt) : smallInt;
   function  ScaleRU(multiplicand, multiplier, divisor : word) : word;

   function  NewHandle(size : longInt) : Handle;
   procedure DisposeHandle(var h : Handle);
   procedure SetHandleSize(h : Handle; size : PtrUInt);
   function  GetHandleSize(h : handle) : PtrUInt;

   function  NewRgn() : RgnHandle;
   procedure OpenRgn;
   procedure CloseRgn(var dstRgn : RgnHandle);
   procedure DisposeRgn(rgn : RgnHandle);
   procedure CopyRgn(srcRgn : RgnHandle; dstRgn : RgnHandle);
   procedure SetEmptyRgn(rgn : RgnHandle);
   procedure SetRectRgn(rgn : RgnHandle; left, top, right, bottom : smallInt);
   procedure RectRgn(rgn : RgnHandle; r : Rect);
   procedure SetClip(rgn : RgnHandle);
   procedure GetClip(rgn : RgnHandle);
   procedure ClipRect(r : Rect);
   procedure OffsetRgn(rgn : RgnHandle; dh, dv : smallInt);
   function  PtInRgn(pt : Point; rgn : RgnHandle) : Boolean;
   function  RectInRgn(r : Rect; rgn : RgnHandle) : Boolean;
   function  EqualRgn(rgn1, rgn2 : RgnHandle) : Boolean;
   function  EmptyRgn(rgn : RgnHandle) : Boolean;
   procedure PaintRgn(rgn : RgnHandle);
   procedure EraseRgn(rgn : RgnHandle);
   procedure InvertRgn(rgn : RgnHandle);
   procedure FillRgn(rgn : RgnHandle; pat : Pattern);

   procedure InitGraf;
   procedure FlushGraf;
   procedure FlushGraf(r : Rect);
   procedure EndGraf;
   procedure OpenPort(port : GrafPtr);
   procedure InitPort(port : GrafPtr);
   procedure ClosePort(port : GrafPtr);
   procedure SetPort(port : GrafPtr);
   procedure GetPort(var port : GrafPtr);
   procedure GrafDevice(device : smallInt);
   procedure SetPortBits(bm : BitMap);
   procedure PortSize(width, height : smallInt);
   procedure MovePortTo(leftGlobal, topGlobal : smallInt);
   procedure SetOrigin(h, v : smallInt);
   procedure BackPat(pat : Pattern);
   procedure AddPt(srcPt : Point; var dstPt : Point);
   procedure SubPt(srcPt : Point; var dstPt : Point);
   procedure SetPt(var pt : Point; h, v : smallInt);
   procedure OffsetPt(var pt : Point; deltah, deltav : smallInt);
   function  EqualPt(pt1, pt2 : Point) : Boolean;
   procedure ScalePt(var pt : Point; srcRect, dstRect : Rect);
   procedure MapPt(var pt : Point; srcRect, dstRect : Rect);
   procedure LocalToGlobal(var pt : Point);
   procedure GlobalToLocal(var pt : Point);
   procedure SetRect(var rect : Rect; left, top, right, bottom : smallInt);
   procedure DumpPt(p : Point; s : string);
   procedure DumpRect(r : Rect; s : string);
   procedure DumpPat(pat : Pattern; s : string);
   procedure DumpBmapHdr(bmap : BitMap; s : string);
   procedure DumpBmap(bmap : BitMap; s : string);
   procedure DumpRgn(rgn : RgnHandle; s : string);
   procedure DumpPort(p : GrafPtr);
   function  RectWidth(r : rect) : smallInt;
   function  RectHeight(r : rect) : smallInt;
   procedure OffsetRect(var rect : Rect; deltah, deltav : smallInt);
   procedure InsetRect(var rect : Rect; deltah, deltav : smallInt);
   function  SectRect(srect1, srect2 : Rect; var destrect : Rect) : Boolean;
   procedure UnionRect(srect1, srect2 : Rect; var destrect : Rect);
   function  PtInRect(point : Point; r : Rect) : Boolean;
   procedure Pt2Rect(point1, point2 : Point; var drect : Rect);
   procedure RotRect(sr : Rect; var dr : Rect);
   function  EqualRect(rect1, rect2 : Rect) : Boolean;
   function  EmptyRect(r : Rect) : Boolean;
   function  PinRect(theRect : Rect; thePt : Point) : Point;
   procedure MapRect(var r : Rect; srcRect, dstRect : Rect);
   procedure FrameRect(r : Rect);
   procedure   iSetBit(x, y : smallInt); // Internal (inter-module) use only
   procedure iClearBit(x, y : smallInt); // Internal (inter-module) use only
   procedure  iFlipBit(x, y : smallInt); // Internal (inter-module) use only
   procedure PaintRect(r : Rect);
   procedure EraseRect(r : Rect);
   procedure InvertRect(r : Rect);
   procedure FillRect(r : Rect; fillPat : Pattern);
   procedure HidePen;
   procedure ShowPen;
   procedure GetPen(var pen : Point);
   procedure GetPenState(var penRec : PenState);
   procedure SetPenState(penrec : PenState);
   procedure PenSize(width, height : smallInt);
   procedure PenPat(pat : Pattern);
   procedure PenMode(mode : xferMode);
   procedure PenNormal;
   procedure MoveTo(x, y : smallInt);
   procedure Move(x, y : smallInt);
   procedure SetPixel(h, v : smallInt);
   function  GetPixel(h, v : smallInt) : Boolean;
   procedure iBuildLine(r : Rect; negative : Boolean; p : PsmallInt); // Intrnl!
   procedure iBuildLRgn(r : Rect; neg : Boolean; line, p : PsmallInt);// Intrnl!
   procedure Line(x, y : smallInt);
   procedure LineTo(newx, newy : smallInt);

   procedure CopyBits(var srcBits, dstBits : BitMap; var srcRect, dstRect : Rect;
                      mode : xferMode; maskRgn : RgnHandle);
   procedure PlotIcon(var r : Rect; var iconptr : Icon);
   function  RotateBMapCCW(srcBMap : BitMap; var dstBMap : BitMap) : Boolean;
   function  RotateBMapCW(srcBMap : BitMap; var dstBMap : BitMap) : Boolean;

var
   // Drawing context, such as it is.
   thePort : GrafPtr;           // Pointer to current graphics port.

implementation

uses
   Clunkpoly,           // Unfortunate dependency due to the line drawer.
   sysUtils, ePaper,
{$IFDEF Unix}
   Unix, baseunix;
{$ENDIF}

{$IFDEF Windows}
   Windows, ShellAPI;
{$ENDIF}

const
   NHANDLES = 250;      // Number of Handles we can, umm, _handle_.
                        // Supports around 50 Windows.

   EndMarker = $7FFF;   // End-of-region/polygon-data marker value.

var
   handles : ^Ptr = nil;

   screenBits : BitMap;         // The raw hardware.

   procedure HLine(width : smallInt); forward;
   procedure VLine(height : smallInt); forward;
   procedure iAdd2Poly(x, y : smallInt); forward;

//
// Math.  Much of this was originally designed to allow efficient
//        access to 68010 machine instructions.
//

// Given a positive fixed-point number, FixRound rounds it to the nearest
// integer and returns the result.  If the value is halfway between two
// integers (.5), it's rounded up.
//
function FixRound(f : Fixed) : smallInt;
begin
   FixRound := (f + $8000) >> 16;
end;

// Int2Fix converts an integer to a fixed point number with a zero
// fractional part.
//
function Int2Fix(i : integer) : Fixed;
begin
   Int2Fix := i << 16;
end;

// HiWord returns the high-order word of the given long integer.  One of the
// uses of this function is to extract the integer part of a fixed-point number.
//
function HiWord(f : Fixed) : word;
begin
   HiWord := f >> 16;
end;

function Fix2Int(f : Fixed) : smallInt;
begin
   Fix2Int := HiWord(f);
end;

// LoWord returns the low-order word of the given long integer.  One of the
// uses of this function is to extract the fractional part of a fixed-point
// number.
//
function LoWord(f : Fixed) : word;
begin
   LoWord := f and $FFFF;
end;

function Fix2Frac(f : Fixed) : integer;
begin
   Fix2Frac := LoWord(f);
end;

// FixRatio returns the signed fixed-point ratio of dividend / divisor.
//
function FixRatio(dividend, divisor : smallInt) : Fixed;
begin
   if divisor = 0 then begin
      FixRatio := $7FFFFFFF;
      if dividend < 0 then
         FixRatio := -FixRatio;
   end else begin
      FixRatio := (Int32(dividend)<<16) div Int32(divisor);
   end;
end;

// FixMul returns the signed fixed-point product of a and b.  The result is
// computed modulo 65536, truncated, and signed according to the signs of a
// and b.  A positive overflow returns $7FFFFFFF, a negative returns $80000001.
// On the original 68010 this result had to be made out of an intermediate
// 64-bit product elaborately constructed with the 68010's 16x16=32 multiplier.
// It's a lot simpler here!
//
function FixMul(multiplicand, multiplier : Fixed) : Fixed;
var product : Int64;
begin
   product := Int64(multiplicand) * Int64(multiplier);
   FixMul := product >> 16;
end;

// Determine whether our fixed-point scale factor is binary.  (Or else decimal.)
function isBinaryScale(scale : dword) : Boolean;
begin
   scale := scale and (scale - 1); // Clear the most-significant set bit.
   isBinaryScale := scale = 0;     // No bits left?  Then was binary.
end;

function isBinaryScale(scale : qword) : Boolean;
begin
   scale := scale and (scale - 1); // Clear the most-significant set bit.
   isBinaryScale := scale = 0;     // No bits left?  Then was binary.
end;

function strtoul(s : string; var send : PChar) : longword;
var ii : integer;
    ch : Char;
    started : Boolean;
begin
   strtoul := 0;
   send := PChar(s);
   started := False;
   ii := 1;
   while ii <= length(s) do begin
      ch := s[ii];
      if not started and (ch = ' ') then begin
         inc(ii);
         continue;
      end;
      started := True;
      if (ch < '0') or (ch > '9') then
         break;
      strtoul := (strtoul * 10) + (ord(ch) - ord('0'));
      inc(ii);
   end;
   send := PChar(@s[ii]);
end;

// Clunkdraw fixed-point is a binary type.  They can also be used more
// as a decimal type, exemplified by the later conversion routines.
// They don't really mix, you'd usually use one kind or the other.
// (Clunkdraw's usual type would have a scale of 65536, which is illegal here.)

// Parse a fixed-point unsigned integer.  Modeled after C's strtoul.
//
// Examples: strToFixedU32("1.23",  0,   1, 1) ==   1 (ones)
//           strToFixedU32("1.23",  0,  10, 1) ==  12 (tenths)
//           strToFixedU32("1.23",  0, 100, 1) == 123 (hundredths)
//           strToFixedU32("1.234", 0, 100, 1) == 123 (hundredths)
//           strToFixedU32("1.2",   0, 100, 1) == 120 (hundredths)
//
function StrToFixedU32(s : string; var send : PChar; scale : integer; trunc : Boolean) : longword;
var cp : Pchar;
    ch : Char;
begin
   cp := PChar(s);
   if (scale <= 0) or (((scale mod 10) <> 0) and (scale <> 1)) then begin
      send := cp;
      strToFixedU32 := 0;       // Scale is integer powers of 10 only!
      exit;
   end;
   strToFixedU32 := strtoul(s, cp) * scale; // Integer part.
   ch := cp^;
   inc(cp);
   if ch = '.' then begin       // Fractional part? */
      repeat
         ch := cp^;
         if (scale <= 1) or (ord(ch) = 0) then break;
         if (scale mod 10) <> 0 then // Scale is integer powers of 10 only!
            break;
         scale := scale div 10;
         if (ch >= '0') and (ch <= '9') then begin
            strToFixedU32 += (ord(ch) - ord('0')) * scale;
            inc(cp);            // Consume the digit.
         end else               // Not a digit, we're done.
            break;
      until False;
      repeat
         ch := cp^;
         if not trunc or (ch < '0') or (ch > '9') then
            break;
         inc(cp);               // Discard excess digits.
      until False;
   end else                     // Not a decimal point,
      dec(cp);                  // so don't consume it.
   send := cp;
end;

function strtoull(s : string; var send : PChar) : qword;
var ii : integer;
    ch : Char;
    started : Boolean;
begin
   strtoull := 0;
   send := PChar(s);
   started := False;
   ii := 1;
   while ii <= length(s) do begin
      ch := s[ii];
      if not started and (ch = ' ') then begin
         inc(ii);
         continue;
      end;
      started := True;
      if (ch < '0') or (ch > '9') then
         break;
      strtoull := (strtoull * 10) + (ord(ch) - ord('0'));
      inc(ii);
   end;
   send := PChar(@s[ii]);
end;

function StrToFixedU64(s : string; var send : PChar; scale : integer; trunc : Boolean) : qword;
var cp : Pchar;
    ch : Char;
begin
   cp := PChar(s);
   if (scale <= 0) or (((scale mod 10) <> 0) and (scale <> 1)) then begin
      send := cp;
      strToFixedU64 := 0;       // Scale is integer powers of 10 only!
      exit;
   end;
   strToFixedU64 := strtoull(s, cp) * scale; // Integer part.
   ch := cp^;
   inc(cp);
   if ch = '.' then begin       // Fractional part? */
      repeat
         ch := cp^;
         if (scale <= 1) or (ord(ch) = 0) then break;
         if (scale mod 10) <> 0 then // Scale is integer powers of 10 only!
            break;
         scale := scale div 10;
         if (ch >= '0') and (ch <= '9') then begin
            strToFixedU64 += (ord(ch) - ord('0')) * scale;
            inc(cp);            // Consume the digit.
         end else               // Not a digit, we're done.
            break;
      until False;
      repeat
         ch := cp^;
         if not trunc or (ch < '0') or (ch > '9') then
            break;
         inc(cp);               // Discard excess digits.
      until False;
   end else                     // Not a decimal point,
      dec(cp);                  // so don't consume it.
   send := cp;
end;

function strtol(s : string; var send : PChar) : longint;
var ii : integer;
    ch : Char;
    negative, started : Boolean;
begin
   strtol := 0;
   send := PChar(s);
   started := False;
   negative := False;
   ii := 1;
   while ii <= length(s) do begin
      ch := s[ii];
      if not started and (ch = ' ') then begin
         inc(ii);
         continue;
      end;
      started := True;
      if not negative and (ch = '-') then begin
         negative := True;
         continue;
      end;
      if (ch < '0') or (ch > '9') then
         break;
      strtol := (strtol * 10) + (ord(ch) - ord('0'));
      inc(ii);
   end;
   if negative then
      strtol := -strtol;
   send := PChar(@s[ii]);
end;

function StrToFixedS32(s : string; var send : PChar; scale : integer; trunc : Boolean) : longint;
var cp : Pchar;
    ch : Char;
begin
   cp := PChar(s);
   if (scale <= 0) or (((scale mod 10) <> 0) and (scale <> 1)) then begin
      send := cp;
      strToFixedS32 := 0;       // Scale is integer powers of 10 only!
      exit;
   end;
   strToFixedS32 := strtol(s, cp) * scale; // Integer part.
   ch := cp^;
   inc(cp);
   if ch = '.' then begin       // Fractional part? */
      repeat
         ch := cp^;
         if (scale <= 1) or (ord(ch) = 0) then break;
         if (scale mod 10) <> 0 then // Scale is integer powers of 10 only!
            break;
         scale := scale div 10;
         if (ch >= '0') and (ch <= '9') then begin
            if strToFixedS32 < 0 then
               strToFixedS32 -= (ord(ch) - ord('0')) * scale
            else
               strToFixedS32 += (ord(ch) - ord('0')) * scale;
            inc(cp);            // Consume the digit.
         end else               // Not a digit, we're done.
            break;
      until False;
      repeat
         ch := cp^;
         if not trunc or (ch < '0') or (ch > '9') then
            break;
         inc(cp);               // Discard excess digits.
      until False;
   end else                     // Not a decimal point,
      dec(cp);                  // so don't consume it.
   send := cp;
end;

function strtoll(s : string; var send : PChar) : int64;
var ii : integer;
    ch : Char;
    negative, started : Boolean;
begin
   strtoll := 0;
   send := PChar(s);
   started := False;
   negative := False;
   ii := 1;
   while ii <= length(s) do begin
      ch := s[ii];
      if not started and (ch = ' ') then begin
         inc(ii);
         continue;
      end;
      started := True;
      if not negative and (ch = '-') then begin
         negative := True;
         continue;
      end;
      if (ch < '0') or (ch > '9') then
         break;
      strtoll := (strtoll * 10) + (ord(ch) - ord('0'));
      inc(ii);
   end;
   if negative then
      strtoll := -strtoll;
   send := PChar(@s[ii]);
end;

function StrToFixedS64(s : string; var send : PChar; scale : integer; trunc : Boolean) : int64;
var cp : Pchar;
    ch : Char;
begin
   cp := PChar(s);
   if (scale <= 0) or (((scale mod 10) <> 0) and (scale <> 1)) then begin
      send := cp;
      strToFixedS64 := 0;       // Scale is integer powers of 10 only!
      exit;
   end;
   strToFixedS64 := strtoll(s, cp) * scale; // Integer part.
   ch := cp^;
   inc(cp);
   if ch = '.' then begin       // Fractional part? */
      repeat
         ch := cp^;
         if (scale <= 1) or (ord(ch) = 0) then break;
         if (scale mod 10) <> 0 then // Scale is integer powers of 10 only!
            break;
         scale := scale div 10;
         if (ch >= '0') and (ch <= '9') then begin
            if strToFixedS64 < 0 then
               strToFixedS64 -= (ord(ch) - ord('0')) * scale
            else
               strToFixedS64 += (ord(ch) - ord('0')) * scale;
            inc(cp);            // Consume the digit.
         end else               // Not a digit, we're done.
            break;
      until False;
      repeat
         ch := cp^;
         if not trunc or (ch < '0') or (ch > '9') then
            break;
         inc(cp);               // Discard excess digits.
      until False;
   end else                     // Not a decimal point,
      dec(cp);                  // so don't consume it.
   send := cp;
end;

// Turn a fixed-point U32 into a string.
//
// Examples: fixedU32ToStr(   123,     1 ...) == "123"
//           fixedU32ToStr(   123,    10 ...) == "12.3"
//           fixedU32ToStr(   123,   100 ...) == "1.23"
//           fixedU32ToStr(   123,  1000 ...) == "0.123"
//           fixedU32ToStr($A8000, 65536 ...) == "10.5"
//           fixedU32ToStr(     3,     2 ...) == "1.5"
//
// The commas flag will result in the integer part having
// commas every three digits.  Use of non-decimal scales can
// result in a lot of unwanted precision after the decimal point,
// use of dtrunc is advised.
//
function FixedU32ToStr(fixed, scale : dword; dtrunc : integer; commas : Boolean) : string;
var log, ii, digits : integer;
    quotient, remainder : dword;
    chars : array[1..13] of Char; // Stack of output characters.
begin
   FixedU32ToStr := '';
   try
      quotient := fixed div scale;
      if not commas then begin
         FixedU32ToStr += IntToStr(quotient);
      end else begin            // We want comma'd integer conversion.
         remainder := quotient;
         log := 1;
         while remainder > 9 do begin  // Calculate int(log(fixed/scale))
            remainder := remainder div 10;
            inc(log);
         end;
         ii := log + ((log-1) div 3); // Length of digits and their commas.
         digits := 0;
         log := ii;
         while log <> 0 do begin // Work backwards, optimizes digit math.
            if digits = 3 then begin // Done three digits?  Time for a comma!
               chars[log] := ',';
               digits := 0;
            end else begin      // Do a digit.
               chars[log] := chr((quotient mod 10) + ord('0'));
               quotient := quotient div 10;
               inc(digits);
            end;
            dec(log);
         end;
         repeat                 // Now emit the comma'd string in forward order.
            inc(log);
            FixedU32ToStr += chars[log];
         until log = ii;
      end;
      if (scale <> 1) then begin // Potential fractional part?
         FixedU32ToStr += '.';
         remainder := fixed mod scale;  // The fractional part.
         if isBinaryScale(scale) then begin
            digits := 0;
            repeat
               if scale > remainder then
                  remainder *= 10;
               ii := remainder div scale;
               FixedU32ToStr += chr(ii + ord('0'));
               inc(digits);
               if (dtrunc > 0) and (digits >= dtrunc) then break; // Truncate?
               remainder := remainder mod scale;
            until remainder = 0;
         end else begin         // Scale is a power of 10.
            log := 0;
            ii := scale;
            while ii > 1 do begin // Calculate int(log(scale));
               ii := ii div 10;
               inc(log);
            end;
            ii := log;
            while ii <> 0 do begin // Work backwards, optimizes digit math.
               chars[ii] := chr((remainder mod 10) + ord('0')); // Do a digit.
               remainder := remainder div 10;
               dec(ii);
            end;
            digits := 0;
            repeat                 // Now emit the string in forward order.
               inc(digits);
               FixedU32ToStr += chars[digits];
               if (dtrunc > 0) and (digits >= dtrunc) then break; // Truncate?
            until digits = log;
         end;
      end;
   except
      FixedU32ToStr += '***';
      exit;
   end;
end;

function FixedU64ToStr(fixed, scale : qword; dtrunc : integer; commas : Boolean) : string;
var log, ii, digits : integer;
    quotient, remainder : qword;
    chars : array[1..26] of Char; // Stack of output characters.
begin
   FixedU64ToStr := '';
   try
      quotient := fixed div scale;
      if not commas then begin
         FixedU64ToStr += IntToStr(quotient);
      end else begin            // We want comma'd integer conversion.
         remainder := quotient;
         log := 1;
         while remainder > 9 do begin  // Calculate int(log(fixed/scale))
            remainder := remainder div 10;
            inc(log);
         end;
         ii := log + ((log-1) div 3); // Length of digits and their commas.
         digits := 0;
         log := ii;
         while log <> 0 do begin // Work backwards, optimizes digit math.
            if digits = 3 then begin // Done three digits?  Time for a comma!
               chars[log] := ',';
               digits := 0;
            end else begin      // Do a digit.
               chars[log] := chr((quotient mod 10) + ord('0'));
               quotient := quotient div 10;
               inc(digits);
            end;
            dec(log);
         end;
         repeat                 // Now emit the comma'd string in forward order.
            inc(log);
            FixedU64ToStr += chars[log];
         until log = ii;
      end;
      if (scale <> 1) then begin // Potential fractional part?
         FixedU64ToStr += '.';
         remainder := fixed mod scale;  // The fractional part.
         if isBinaryScale(scale) then begin
            digits := 0;
            repeat
               if scale > remainder then
                  remainder *= 10;
               ii := remainder div scale;
               FixedU64ToStr += chr(ii + ord('0'));
               inc(digits);
               if (dtrunc > 0) and (digits >= dtrunc) then break; // Truncate?
               remainder := remainder mod scale;
            until remainder = 0;
         end else begin         // Scale is a power of 10.
            log := 0;
            ii := scale;
            while ii > 1 do begin  // Calculate int(log(scale));
               ii := ii div 10;
               inc(log);
            end;
            ii := log;
            while ii <> 0 do begin // Work backwards, optimizes digit math.
               chars[ii] := chr((remainder mod 10) + ord('0')); // Do a digit.
               remainder := remainder div 10;
               dec(ii);
            end;
            digits := 0;
            repeat                 // Now emit the string in forward order.
               inc(digits);
               FixedU64ToStr += chars[digits];
               if (dtrunc > 0) and (digits >= dtrunc) then break; // Truncate?
            until digits = log;
         end;
      end;
   except
      FixedU64ToStr += '***';
      exit;
   end;
end;

// Turn a fixed-point S32 into a string.
//
function FixedS32ToStr(fixed : longint; scale, dtrunc : integer; commas : Boolean) : string;
begin
   if fixed < 0 then
      fixedS32ToStr := '-' + FixedU32ToStr(-fixed, scale, dtrunc, commas)
   else
      fixedS32ToStr := FixedU32ToStr(longword(fixed), scale, dtrunc, commas);
end;

// Turn a fixed-point S64 into a string.
//
function FixedS64ToStr(fixed : int64; scale : qword; dtrunc : integer; commas : Boolean) : string;
begin
   if fixed < 0 then
      fixedS64ToStr := '-' + FixedU64ToStr(-fixed, scale, dtrunc, commas)
   else
      fixedS64ToStr := FixedU64ToStr(qword(fixed), scale, dtrunc, commas);
end;

//
// Precision-changing math.  Designed to the 68010 instruction set.
//
function MulS(multiplicand, multiplier : smallInt) : longint;
begin
   MulS := multiplicand * multiplier;
end;

function MulU(multiplicand, multiplier : word) : longword;
begin
   MulU := multiplicand * multiplier;
end;

function DivS(dividend : longint; divisor : smallInt) : smallInt;
begin
   DivS := dividend div divisor;
end;

function DivU(dividend : longword; divisor : word) : longword;
begin
   DivU := dividend div divisor;
end;

// This was a single machine instruction on the 68010.
function DivModS(dividend : longint; divisor : smallInt) : longword;
begin
   DivModS := (DivS(dividend, divisor) and $FFFF) or
              (ModS(dividend, divisor) << 16);
end;

// This was a single machine instruction on the 68010.
function DivModU(dividend : longword; divisor : word) : longword;
begin
   DivModU := (DivU(dividend, divisor) and $FFFF) or
              (ModU(dividend, divisor) << 16);
end;

function ModS(dividend : longint; divisor : smallInt) : smallInt;
begin
   ModS := dividend mod divisor;
end;

function ModU(dividend : longword; divisor : word) : word;
begin
   ModU := dividend mod divisor;
end;

function ScaleS(multiplicand, multiplier, divisor : smallInt) : smallInt;
var intermediate : longint;
begin
   intermediate := multiplicand * multiplier;
   ScaleS := intermediate div divisor;
end;

function ScaleU(multiplicand, multiplier, divisor : word) : word;
var intermediate : longword;
begin
   intermediate := multiplicand * multiplier;
   ScaleU := intermediate div divisor;
end;

function ScaleRS(multiplicand, multiplier, divisor : smallInt) : smallInt;
var intermediate : longint;
begin
   intermediate := multiplicand * multiplier;
   ScaleRS := intermediate div divisor;
   if ((intermediate mod divisor) << 1) >= divisor then
      inc(ScaleRS);
end;

function ScaleRU(multiplicand, multiplier, divisor : word) : word;
var intermediate : longword;
begin
   intermediate := multiplicand * multiplier;
   ScaleRU := intermediate div divisor;
   if ((intermediate mod divisor) << 1) >= divisor then
      inc(ScaleRU);
end;

// Initialize Clunkdraw, up to and including creating the monochrome
// bitmap storage buffer for the raw (480 tall, 280 wide) 3.7" E-Paper
// display hardware.  This particular panel is physically in portrait
// orientation, but is used here in landscape orientation, so we've
// embedded rotation into our low-level access to it.  It also supports
// 4-level grayscale, but that is very slow.  Using it in 1-bit monochrome
// mode is much faster, and is all Clunkdraw really supports anyhow.
//
procedure InitGraf;
var ImageSize : PtrUInt;
begin
   //                      left top     right           bottom
   SetRect(screenBits.bounds, 0, 0, EPD_3IN7_WIDTH, EPD_3IN7_HEIGHT);
   screenBits.baseAddr := nil;
   screenBits.rowBytes := (screenBits.bounds.right + 7) div 8;
   ImageSize := screenBits.rowBytes * screenBits.bounds.bottom;
   screenBits.baseAddr := getMem(ImageSize);
   if screenBits.baseAddr = nil then begin
      WriteLn('Failed to allocate image memory...');
      exit;
   end;
   FillByte(screenBits.baseAddr^, ImageSize, 0);
   New(thePort);
   if thePort = nil then begin
      WriteLn('Failed to allocate GrafPort memory...');
      exit;
   end;
   FillByte(thePort^, sizeof(GrafPort), 0);
   OpenPort(thePort);
   // Now connect up to the e-Paper display.
   EPD_SPI_Init;
   EPD_3IN7_4Gray_Init;  // Wipe out everything in the hardware.
   EPD_3IN7_4Gray_Clear;
   EPD_3IN7_1Gray_Init;  // Now hook up in monochrome mode.
   EPD_3IN7_1Gray_Clear;
end;

// Push the entire image buffer to the display hardware.
//
procedure FlushGraf;
begin
   if thePort = nil then exit;
///   EPD_3IN7_1Gray_Init;
   if thePort^.device = 1 then
      EPD_3IN7_1Gray_Display(thePort^.portBits.baseAddr);
///   EPD_3IN7_Sleep;
end;

// Push part of the image buffer to the display hardware.  Very limited!
//
procedure FlushGraf(r : Rect);
var tp : GrafPtr;
    scratch : Rect;
begin
   tp := thePort;
   if tp = nil then exit;
   if tp^.device = 1 then begin
      if SectRect(r, tp^.portRect, scratch) then begin
         RotRect(scratch, scratch); // Move to true display coordinates.
         scratch.left := 0; // Conform to the limits of the partial display fn.
         scratch.top := 0;
         scratch.right := EPD_3IN7_WIDTH;
         EPD_3IN7_1Gray_Display_Part(tp^.portBits.baseAddr, scratch.left,
                                     scratch.top, scratch.right, scratch.bottom);
      end;
   end;
end;

procedure EndGraf;
begin
   if thePort = nil then exit;
   Dispose(thePort);
   thePort := nil;
   freeMem(screenBits.baseAddr);
   screenBits.baseAddr := nil;
   screenBits.rowBytes := 0;
   SetRect(screenBits.bounds, 0, 0, 0, 0);

   EPD_3IN7_Sleep;
   EPD_Delay_ms(2000); //important, at least 2s
   EPD_SPI_Exit;
end;

procedure OpenPort(port : GrafPtr);
begin
   InitPort(port);
   port^.visRgn := nil;         // We currently only engage clip regions if we
   port^.clipRgn := nil;        // have brought in the optional windowing system.
end;

procedure InitPort(port : GrafPtr);
begin
   SetPort(port);
   GrafDevice(1);               // e-Paper device, rotated.
   SetPortBits(screenBits);
   // Waveshare's 3.7" e-Paper display is really a portrait display, but we
   // want landscape.  Pretend.  (Our low-level drawing code is modified too.)
   RotRect(screenBits.bounds, port^.portRect); // Our fiction, rotated reality.
   MoveTo(0, 0);
   port^.pnVis := 0;            // Pen is visible (>= 0).
   PenNormal();
// TextFace([]);                // Plain text.
   port^.txFace.i := 0;
// TextSize(12);                // Our default font size.
   port^.txSize := 12;          // (Fonts unit is optional, don't use the API.)
// TextFont(0);                 // Our default font.
   port^.txFont := 0;           // (Fonts unit is optional, ditto.)
   port^.fontPtr := nil;
   port^.txMode.e := srcOr;
   port^.txRotate := 0;         // No text rotation.
   port^.spExtra := 0;
   port^.txExtra := 0;
   port^.fgColor := 1;          // Foreground = black.
   port^.bkColor := 0;          // Background = white.
   port^.bkPat := white;
   port^.fillPat := black;
   port^.colrBit := 0;          // No color support at all.
end;

procedure ClosePort(port : GrafPtr);
begin
   if port^.visRgn <> nil then
      DisposeRgn(port^.visRgn);
   port^.visRgn := nil;

   if port^.clipRgn <> nil then
      DisposeRgn(port^.clipRgn);
   port^.clipRgn := nil;
end;

procedure SetPort(port : GrafPtr);
begin
   thePort := port;
end;

procedure GetPort(var port : GrafPtr);
begin
   port := thePort;
end;

procedure GrafDevice(device : smallInt);
begin
   if thePort = nil then exit;
   thePort^.device := device;
end;

procedure SetPortBits(bm : BitMap);
var tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   tp^.portBits := bm;
   tp^.portRect := bm.bounds;
end;

procedure PortSize(width, height : smallInt);
var tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   tp^.portRect.right := tp^.portRect.left + width;
   tp^.portRect.bottom := tp^.portRect.top + height;
end;

procedure MovePortTo(leftGlobal, topGlobal : smallInt);
var width, height : smallInt;
    tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   width := tp^.portRect.right - thePort^.portRect.left;
   height := tp^.portRect.bottom - thePort^.portRect.top;
   if tp^.device = 1 then begin // Handle weird e-paper rotated coords.
      tp^.portRect.top := tp^.portBits.bounds.left + topGlobal;
      tp^.portRect.left := tp^.portBits.bounds.top + leftGlobal;
   end else begin
      tp^.portRect.top := tp^.portBits.bounds.top + topGlobal;
      tp^.portRect.left := tp^.portBits.bounds.left + leftGlobal;
   end;
   tp^.portRect.bottom := tp^.portRect.top + height;
   tp^.portRect.right := tp^.portRect.left + width;
end;

procedure SetOrigin(h, v : smallInt);
var deltah, deltav : smallInt;
    tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   deltah := h - tp^.portRect.left;
   deltav := v - tp^.portRect.top;
   if tp^.device = 1 then // Handle weird e-paper rotated coords.
      OffsetRect(tp^.portBits.bounds, deltav, deltah)
   else
      OffsetRect(tp^.portBits.bounds, deltah, deltav);
   OffsetRect(tp^.portRect, deltah, deltav);
   OffsetRgn(tp^.visRgn, deltah, deltav);
end;

procedure BackPat(pat : Pattern);
begin
   if thePort = nil then exit;
   thePort^.bkPat := pat;
end;

procedure HidePen;
begin
   if thePort = nil then exit;
   thePort^.pnVis -= 1;
end;

procedure ShowPen;
begin
   if thePort = nil then exit;
   thePort^.pnVis += 1;
end;

// Returns (in pen) the current drawing pen position.
//
procedure GetPen(var pen : Point);
begin
   if thePort = nil then exit;
   pen := thePort^.pnLoc;
end;

// Returns (in penRec) the current drawing pen's state.
//
procedure GetPenState(var penRec : PenState);
begin
   if thePort = nil then exit;
   penRec.pnLoc := thePort^.pnLoc;
   penRec.pnSize := thePort^.pnSize;
   penRec.pnMode := thePort^.pnMode;
   penRec.pnPat := thePort^.pnPat;
end;

// Sets the drawing pen's state variables to match those stored in
// PenState.  A handy way to set all pen variables at once.
//
procedure SetPenState(penrec : PenState);
begin
   if thePort = nil then exit;
   thePort^.pnLoc := penRec.pnLoc;
   thePort^.pnSize := penRec.pnSize;
   thePort^.pnMode := penRec.pnMode;
   thePort^.pnPat := penRec.pnPat;
end;

// Set the drawing pen's size to be WIDTH pixels wide and HEIGHT
// pixels high.  All subsequent calls to Line, LineTo, and all
// routines that draw framed shapes will use the new pen size.  If the
// pen size is less than or equal to zero in either dimension, then no
// drawing will be performed.  The default size is 1x1.
//
procedure PenSize(width, height : smallInt);
begin
   if thePort = nil then exit;
   thePort^.pnSize.v := height;
   thePort^.pnSize.h := width;
end;

procedure dumpPen;
begin
   writeLn('pnPat: $', hexStr(thePort^.pnPat.bytes[0], 2),
           ',$',hexStr(thePort^.pnPat.bytes[1], 2),
           ',$',hexStr(thePort^.pnPat.bytes[2], 2),
           ',$',hexStr(thePort^.pnPat.bytes[3], 2),
           ',$',hexStr(thePort^.pnPat.bytes[4], 2),
           ',$',hexStr(thePort^.pnPat.bytes[5], 2),
           ',$',hexStr(thePort^.pnPat.bytes[6], 2),
           ',$',hexStr(thePort^.pnPat.bytes[7], 2));
end;

// Set drawing pen's pattern to PAT.  All subsequent calls to
// Line and LineTo, all routines that draw framed shapes, and
// all PaintXXX routines will use the new pattern.  The default
// pattern is black.
//
procedure PenPat(pat : Pattern);
begin
   if thePort = nil then exit;
   thePort^.pnPat := pat;
end;

// Set drawing pen's transfer mode to be MODE.  MODE must be one of
// the standard logical transfer modes:  srcCopy, srcOr, srcXor, srcBic,
// notSrcCopy, notSrcOr notSrcXor, or notSrcBic.  The default mode is srcCopy.
//
procedure PenMode(mode : xferMode);
begin
   if thePort = nil then exit;
   thePort^.pnMode.e := mode;
end;

// An easy way to reset the pen to its default characteristics.
//
procedure PenNormal;
begin
   PenSize(1, 1);
   PenPat(black);
   PenMode(srcCopy);
end;

//
// Calculations with Points.
//

procedure AddPt(srcPt : Point; var dstPt : Point);
begin
   dstPt.h += srcPt.h;
   dstPt.v += srcPt.v;
end;

procedure SubPt(srcPt : Point; var dstPt : Point);
begin
   dstPt.h -= srcPt.h;
   dstPt.v -= srcPt.v;
end;

procedure SetPt(var pt : Point; h, v : smallInt);
begin
   pt.h := h;
   pt.v := v;
end;

procedure OffsetPt(var pt : Point; deltah, deltav : smallInt);
begin
   pt.h += deltah;
   pt.v += deltav;
end;

function EqualPt(pt1, pt2 : Point) : Boolean;
begin
   EqualPt := (pt1.h = pt2.h) and (pt1.v = pt2.v);
end;

procedure ScalePt(var pt : Point; srcRect, dstRect : Rect);
var val : smallInt;
begin
   val := ScaleS(pt.h, dstRect.right - dstRect.left,
                       srcRect.right - srcRect.left);
   if val < 1 then
      val := 1;
   pt.h := val;
   val := ScaleRS(pt.v, dstRect.bottom - dstRect.top,
                        srcRect.bottom - srcRect.top);
   if val < 1 then
      val := 1;
   pt.v := val;
end;

procedure MapPt(var pt : Point; srcRect, dstRect : Rect);
var val : smallInt;
begin
   val := pt.h - srcRect.left;
   val := ScaleRS(val, dstRect.right - dstRect.left,
                       srcRect.right - srcRect.left);
   pt.h := val + dstRect.left;
   val := pt.v - srcRect.top;
   val := ScaleRS(val, dstRect.bottom - dstRect.top,
                       srcRect.bottom - srcRect.top);
   pt.v := val + dstRect.top;
end;

procedure LocalToGlobal(var pt : Point);
var tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   if tp^.device = 1 then begin // Handle weird e-paper rotated coords.
      pt.h -= tp^.portBits.bounds.top;
      pt.v -= tp^.portBits.bounds.left;
   end else begin
      pt.h -= tp^.portBits.bounds.left;
      pt.v -= tp^.portBits.bounds.top;
   end;
end;

procedure GlobalToLocal(var pt : Point);
var tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   if tp^.device = 1 then begin // Handle weird e-paper rotated coords.
      pt.h += tp^.portBits.bounds.top;
      pt.v += tp^.portBits.bounds.left;
   end else begin
      pt.h += tp^.portBits.bounds.left;
      pt.v += tp^.portBits.bounds.top;
   end;
end;

//
// Calculations with Rectangles.
//
procedure SetRect(var rect : Rect; left, top, right, bottom : smallInt);
begin
   rect.left := left;
   rect.top := top;
   rect.right := right;
   rect.bottom := bottom;
end;

procedure DumpPt(p : Point; s : string);        // Dump out a Point.
begin                                           // For debugging.
   WriteLn(s, 'h: ', p.h, ', ', s, 'v: ', p.v);
end;

procedure DumpRect(r : Rect; s : string);       // Dump out a Rectangle.
begin                                           // For debugging.
   WriteLn(s, 'l: ', r.left, ', ', s, 't: ', r.top, ', ',
           s, 'r: ', r.right, ', ', s, 'b: ', r.bottom, ';  ',
           s, 'h: ', r.right - r.left, ', ', s, 'v: ', r.bottom - r.top);
end;

procedure DumpPat(pat : Pattern; s : string);
begin
   WriteLn(s, ' $', HexStr(pat.bytes[0], 2),
             ', $', HexStr(pat.bytes[1], 2),
             ', $', HexStr(pat.bytes[2], 2),
             ', $', HexStr(pat.bytes[3], 2),
             ', $', HexStr(pat.bytes[4], 2),
             ', $', HexStr(pat.bytes[5], 2),
             ', $', HexStr(pat.bytes[6], 2),
             ', $', HexStr(pat.bytes[7], 2));
end;

procedure DumpBmapHdr(bmap : BitMap; s : string); // Dump out a bitmap header.
begin
   WriteLn(s);
   DumpRect(bmap.bounds, s);
   Write(s, 'baseAddr: $', HexStr(PtrUInt(bmap.baseAddr), 8), ' ');
   WriteLn(s, 'rowBytes: ', bmap.rowBytes,
              ' (size: ', RectHeight(bmap.bounds) * bmap.rowBytes, ')');
end;

procedure DumpBmap(bmap : BitMap; s : string); // Dump out a bitmap, visually.
var x, y, z : smallInt;                        // (Best if it's not too big!)
    p : PByte;                                 // For debugging. There are TWO
    bits : string;                             // of these bmap dumpers, beware!
begin
   DumpBmapHdr(bmap, s);
   p := bmap.baseAddr;
   for y := bmap.bounds.top to bmap.bounds.bottom - 1 do begin
      Write('  ', HexStr(y, 2), ': ');
      for x := 0 to bmap.rowBytes - 1 do begin
         bits := BinStr(p^, 8);
         for z := 1 to length(bits) do begin
            if bits[z] = '1' then bits[z] := '*' else bits[z] := '.';
            if (bmap.bounds.right - bmap.bounds.left) = (x * 8 + z) then
               if bits[z] = '.' then bits[z] := '|';
         end;
         if (bits[8] = '.') and (x <> (bmap.rowBytes - 1)) then
            bits[8] := ' '; // So we can see byte boundaries.
         Write(bits);
         inc(p);
      end;
      WriteLn;
   end;
end;

//
// Dump out a region.  If there are scanlines they'll look like:
// Y: Y1 X: X1..X2, X3..X4... $7FFF
// ...
// Y: $7FFF
//
procedure DumpRgn(rgn : RgnHandle; s : string);
var ps : ^smallInt;
   ii : integer;
begin
   if (rgn = nil) or (rgn^ = nil) then exit;
   ii := rgn^^.rgnSize;
   WriteLn(IntToStr(ii) + '-byte region ' + s); Write(' ');
   DumpRect(rgn^^.rgnBBox, s);
   ii := (ii - sizeof(RegionHdr)) div sizeof(smallInt);
   ps := @rgn^^.rgnData;
   if ii > 0 then begin
      while (ii > 0) {and (ps^ <> EndMarker)} do begin
         Write('Y: ', ps^, ': X: ');
         dec(ii);
         inc(ps);
         while (ii > 0) and (ps^ <> EndMarker) do begin
            Write(ps[0], '..', ps[1], ', ');
            ii -= 2;
            ps += 2;
         end;
         WriteLn('$7FFF');
         dec(ii);
         inc(ps);
      end;
      WriteLn('Y: $7FFF');
   end;
end;

procedure DumpPort(p : GrafPtr);
begin
   Write('Device ', p^.device, ', Bitmap bounds ');
   DumpRect(p^.portBits.bounds, 'pbb');
   Write('Port  ');
   DumpRect(p^.portRect, 'pr');
   DumpRgn(p^.clipRgn, 'Clipping');
   DumpRgn(p^.visRgn, 'Visible');
   DumpPat(p^.pnPat, 'pnPat');
   DumpPat(p^.bkPat, 'bkPat');
   DumpPt(p^.pnLoc, 'pnLoc');
   WriteLn('Font ', p^.txFont, ' Size ', p^.txSize);
   if p^.fontPtr <> nil then begin
      WriteLn('Ascent ', BEtoN(p^.fontPtr^.ascent),
              ' Descent ', BEtoN(p^.fontPtr^.descent));
   end;
end;

// Rectangle Width
//
function RectWidth(r : rect) : smallInt;
begin
   RectWidth := r.right - r.left;
end;

// Rectangle Height
//
function RectHeight(r : rect) : smallInt;
begin
   RectHeight := r.bottom - r.top;
end;

// Shift rect by deltah, deltav.
//
procedure OffsetRect(var rect : Rect; deltah, deltav : smallInt);
begin
    rect.left += deltah;
    rect.right += deltah;
    rect.top += deltav;
    rect.bottom += deltav;
end;

// Inset rect by deltah, deltav.
//
procedure InsetRect(var rect : Rect; deltah, deltav : smallInt);
begin
   rect.left += deltah;
   rect.right -= deltah;
   rect.top += deltav;
   rect.bottom -= deltav;
   if EmptyRect(rect) then
      SetRect(rect, 0, 0, 0, 0);
end;

// Return TRUE if rectangles intersect, also returns the intersection rect.
//
function SectRect(srect1, srect2 : Rect; var destrect : Rect) : Boolean;
var scratch : Rect;
begin
   scratch := srect1;
   if (srect2.top > scratch.top) then
      scratch.top := srect2.top;
   if (srect2.bottom < scratch.bottom) then
      scratch.bottom := srect2.bottom;
   if (srect2.left > scratch.left) then
      scratch.left := srect2.left;
   if (srect2.right < scratch.right) then
      scratch.right := srect2.right;
   if ((scratch.bottom <= scratch.top) or
       (scratch.right <= scratch.left)) then begin
      SetRect(destrect, 0, 0, 0, 0);
      SectRect := False;
      exit;
   end;
   destrect := scratch;
   SectRect := True;
end;

// Return rectangle that encloses the source rectangles.
//
procedure UnionRect(srect1, srect2 : Rect; var destrect : Rect);
var scratch : Rect;
begin
   scratch := srect1;
   if (srect2.top < scratch.top) then
      scratch.top := srect2.top;
   if (srect2.bottom > scratch.bottom) then
      scratch.bottom := srect2.bottom;
   if (srect2.left < scratch.left) then
      scratch.left := srect2.left;
   if (srect2.right > scratch.right) then
      scratch.right := srect2.right;
   destrect := scratch;
end;

// PtInRect determines whether the pixel _below_and_to_the_right_ of the given
// coordinate point is enclosed in the specified rectangle, and returns TRUE
// if so or FALSE if not.
//
// Returns TRUE if the point is enclosed by the rectangle.
//
function PtInRect(point : Point; r : Rect) : Boolean;
begin
   PtInRect := ((point.h >= r.left) and (point.h < r.right) and
                (point.v >= r.top)  and (point.v < r.bottom));
end;

// Pt2Rect returns the smallest rectangle that encloses the two given points.
//
procedure Pt2Rect(point1, point2 : Point; var drect : Rect);
var scratch : Rect;
begin
   scratch.top := point1.v;
   scratch.bottom := point2.v;
   if (scratch.bottom < scratch.top) then begin
      scratch.top := point2.v;
      scratch.bottom := point1.v;
   end;
   scratch.left := point1.h;
   scratch.right := point2.h;
   if (scratch.right < scratch.left) then begin
      scratch.left := point2.h;
      scratch.right := point1.h;
   end;
   drect := scratch;
end;

// RotRect rotates a rectangle structure 90 degrees, effectively
// swapping its coordinate axes.  Used to talk with the 3.7"
// E-Paper display, and in bitmap rotation.
//
procedure RotRect(sr : Rect; var dr : Rect);
begin
   SetRect(dr, sr.top, sr.left, sr.bottom, sr.right);
end;

// Return TRUE if rectangles are equal.
//
function EqualRect(rect1, rect2 : Rect) : Boolean;
begin
   EqualRect := (rect1.top = rect2.top) and
                (rect1.bottom = rect2.bottom) and
                (rect1.left = rect2.left) and
                (rect1.right = rect2.right);
end;

// Return TRUE if rectangle is empty.
//
function EmptyRect(r : Rect) : Boolean;
begin
   EmptyRect := (r.bottom <= r.top) or
                (r.right <= r.left);
end;

// Pin the point within the rectangle.
function PinRect(theRect : Rect; thePt : Point) : Point;
begin
   PinRect := thePt;
   if (thePt.h < theRect.left) then
      PinRect.h := theRect.left
   else if (thePt.h > theRect.right) then
      PinRect.h := theRect.right - 1;
   if (thePt.v < theRect.top) then
      PinRect.v := theRect.top
   else if (thePt.v > theRect.bottom) then
      PinRect.v := theRect.right - 1;
end;

procedure MapRect(var r : Rect; srcRect, dstRect : Rect);
begin
   MapPt(r.topLeft, srcRect, dstRect);
   MapPt(r.botRight, srcRect, dstRect);
end;

// FrameRect draws an outline just inside the specified rectangle, using the
// current grafPort's pen pattern, mode, and size.  The outline is as wide as
// the pen width and as tall as the pen height.  It's drawn with the pnPat,
// according to the transfer mode specified by pnMode.  The pen location is
// not changed by this procedure.
//
// We do this by using four 'stick' rectangles as a frame.  (With small
// rectangles and fat pens, we might only need one.  Because of the XOR
// transfer modes, we don't want to hit any pixel of the frame more than
// once.)
//
procedure FrameRect(r : Rect);
var tlline, brline : Rect;
    pnSize : Point;
    penSave : PenState;
    tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   if (tp^.pnVis < 0) then exit;
   if (tp^.pnSize.h < 1) or (tp^.pnSize.v < 1) then exit;
   GetPenState(penSave);
   tlline := r;  // Make sure frame lines don't squeeze together.  We use two
   brline := r;  // rectangles, first top and bottom, then left and right;
   pnSize := tp^.pnSize;
   tlline.bottom := tlline.top + pnSize.v;
   brline.top := brline.bottom - pnSize.v;
   if SectRect(r, tlline, tlline) then // Stay within the original bounds.
      if (tp^.pnMode.e = srcCopy) then
         FillRect(tlline, tp^.pnPat) // FillRect's faster than PaintRect
      else                          // and they are the same operation when
         PaintRect(tlline);        // using srcCopy mode.
   if brline.top < tlline.bottom then // Squashed together?  Don't strike
      brline.top := tlline.bottom;    // any pixels twice.
   if SectRect(r, brline, brline) then // Stay within the original bounds.
      if (tp^.pnMode.e = srcCopy) then
         FillRect(brline, tp^.pnPat) // FillRect's faster than PaintRect
      else                          // and they are the same operation when
         PaintRect(brline);        // using srcCopy mode.
   if brline.top > tlline.bottom then begin // Any verticals necessary?
      SetRect(tlline, tlline.left, tlline.bottom,
                      tlline.left + pnSize.h, brline.top);
      if SectRect(r, tlline, tlline) then // Stay within the original bounds.
         if (tp^.pnMode.e = srcCopy) then
            FillRect(tlline, tp^.pnPat) // FillRect's faster than PaintRect
         else                          // and they are the same operation when
            PaintRect(tlline);        // using srcCopy mode.
      SetRect(brline, brline.right - pnSize.h, tlline.top,
                      brline.right, tlline.bottom);
      if SectRect(r, brline, brline) then // Stay within the original bounds.
         if (tp^.pnMode.e = srcCopy) then
            FillRect(brline, tp^.pnPat) // FillRect's faster than PaintRect
         else                          // and they are the same operation when
            PaintRect(brline);        // using srcCopy mode.
   end;
   SetPenState(penSave);
end;

// Plot bits on the e-Paper display.  Handle mapping from the real hardware
// layout to the 90-degree rotated drawing model we would like to use.
// (480 pixels wide & 280 pixels high, matching what the demo programs do.)
// These routines work in 0-based physical coordinates, which means the
// layer above can merge patterns cleanly, etc.
procedure iSetBit(x, y : smallInt);
var mask : byte;
    d : Pbyte;
    tp : GrafPtr;
    wid, hei : smallInt;
begin
   tp := thePort; // (Don't need to validate this, it was done much earlier.)
   if tp^.device = 1 then begin // 3.7 inch E-paper.
      wid := tp^.portBits.bounds.bottom - tp^.portBits.bounds.top;
      hei := tp^.portBits.bounds.right - tp^.portBits.bounds.left;
      x := wid - 1 - x;
      if (y < 0) or (y >= hei) or         // Stay on
         (x < 0) or (x >= wid) then exit; // the e-Paper!
      x *= (7 + hei) div 8;
      d := tp^.portBits.baseAddr + x + (y div 8);
      mask := 1 << (7 - (y and 7));
   end else begin               // Normal (offscreen) monochrome bitmap.
      wid := tp^.portBits.bounds.right - tp^.portBits.bounds.left;
      hei := tp^.portBits.bounds.bottom - tp^.portBits.bounds.top;
      if (y < 0) or (y >= hei) or         // Stay on
         (x < 0) or (x >= wid) then exit; // the bitmap!
      d := tp^.portBits.baseAddr + (y * tp^.portBits.rowBytes) + (x div 8);
      mask := 1 << (7 - (x and 7));
   end;
   d^ := d^ or mask;
end;

procedure iClearBit(x, y : smallInt);
var mask : byte;
    d : Pbyte;
    tp : GrafPtr;
    wid, hei : smallInt;
begin
   tp := thePort; // (Don't need to validate this, it was done much earlier.)
   if tp^.device = 1 then begin // 3.7 inch E-paper.
      wid := tp^.portBits.bounds.bottom - tp^.portBits.bounds.top;
      hei := tp^.portBits.bounds.right - tp^.portBits.bounds.left;
      x := wid - 1 - x;
      if (y < 0) or (y >= hei) or         // Stay on
         (x < 0) or (x >= wid) then exit; // the e-Paper!
      x *= (7 + hei) div 8;
      d := tp^.portBits.baseAddr + x + (y div 8);
      mask := 1 << (7 - (y and 7));
   end else begin               // Normal (offscreen) monochrome bitmap.
      wid := tp^.portBits.bounds.right - tp^.portBits.bounds.left;
      hei := tp^.portBits.bounds.bottom - tp^.portBits.bounds.top;
      if (y < 0) or (y >= hei) or         // Stay on
         (x < 0) or (x >= wid) then exit; // the bitmap!
      d := tp^.portBits.baseAddr + (y * tp^.portBits.rowBytes) + (x div 8);
      mask := 1 << (7 - (x and 7));
   end;
   d^ := d^ and not mask;
end;

procedure iFlipBit(x, y : smallInt);
var mask : byte;
    d : Pbyte;
    tp : GrafPtr;
    wid, hei : smallInt;
begin
   tp := thePort; // (Don't need to validate this, it was done much earlier.)
   if tp^.device = 1 then begin // 3.7 inch E-paper.
      wid := tp^.portBits.bounds.bottom - tp^.portBits.bounds.top;
      hei := tp^.portBits.bounds.right - tp^.portBits.bounds.left;
      x := wid - 1 - x;
      if (y < 0) or (y >= hei) or         // Stay on
         (x < 0) or (x >= wid) then exit; // the e-Paper!
      x *= (7 + hei) div 8;
      d := tp^.portBits.baseAddr + x + (y div 8);
      mask := 1 << (7 - (y and 7));
   end else begin               // Normal (offscreen) monochrome bitmap.
      wid := tp^.portBits.bounds.right - tp^.portBits.bounds.left;
      hei := tp^.portBits.bounds.bottom - tp^.portBits.bounds.top;
      if (y < 0) or (y >= hei) or         // Stay on
         (x < 0) or (x >= wid) then exit; // the bitmap!
      d := tp^.portBits.baseAddr + (y * tp^.portBits.rowBytes) + (x div 8);
      mask := 1 << (7 - (x and 7));
   end;
   d^ := d^ xor mask;
end;

// Find the beginning of the region scanline that covers the given
// Y (local) coordinate.  For use on clipping regions that begin
// above where we start drawing.  (By design we should never begin
// drawing where the region hasn't started yet, due to the SectRect
// we have done with all the bounding rectangles.)
//
procedure fastForward(var ps : PsmallInt; y : integer);
var p : ^smallInt;
begin
   p := ps;
   while p^ <> EndMarker do begin
      if p^ >= y then
         break;
      ps := p;          // This scanline is not yet too far.  Move to it
      inc(p);           // and skip to the next one.
      while p^ <> EndMarker do
         p += 2;        // Skip each X segment.
      inc(p);           // Skip the End-of-X marker.
   end;
end;

// PaintRect paints the specified rectangle with the current grafPort's pen
// pattern and mode.  The rectangle is filled with the pnPat, according to
// the transfer mode specified by pnMode.  The pen location is not changed
// by this procedure.
//
procedure PaintRect(r : Rect);
var br : Rect;          // Bounding rectangle, in global coordinates.
    pbr : Rect;
    co, vo, x, y : integer;
    patBit : Boolean;
    tp : GrafPtr;
    cps, vps : ^smallInt; // Clip and Vis Rgn data pointers; Nil if inactive.
    cStartX, cEndX, vStartX, vEndX : smallInt;
begin
   tp := thePort;
   if tp = nil then exit;
   if (tp^.pnVis < 0) then exit;
   co := 0; cEndX := 0;
   vo := 0; vEndX := 0;
   pbr := tp^.portBits.bounds;  // We need the logical portBits.bounds for
   if tp^.device = 1 then       // clipping, and moving to global coordinates.
      RotRect(pbr, pbr);
   if SectRect(r, tp^.portRect, br) then begin
      if SectRect(br, pbr, br) then begin
         cps := nil;
         if (tp^.clipRgn <> nil) and (tp^.clipRgn^ <> nil) then begin
            if not SectRect(br, tp^.clipRgn^^.rgnBBox, br) then exit;
            if tp^.clipRgn^^.rgnSize <> sizeof(RegionHdr) then
               cps := @tp^.clipRgn^^.rgnData;
            co := 0;               // Offset to Y marker.
         end;
         vps := nil;
         if (tp^.visRgn <> nil) and (tp^.visRgn^ <> nil) then begin
            if not SectRect(br, tp^.visRgn^^.rgnBBox, br) then exit;
            if tp^.visRgn^^.rgnSize <> sizeof(RegionHdr) then
               vps := @tp^.visRgn^^.rgnData;
            vo := 0;               // Offset to Y marker.
         end;
         for y := br.top - pbr.top to br.bottom - pbr.top - 1 do begin
            if (cps <> nil) then begin
               if co = 0 then      // Find the first appropriate scanline.
                  fastForward(cps, y + pbr.top);
               if (cps[co] <> EndMarker) and (y >= (cps[co] - pbr.top)) then
                  cps += co;       // Time to move the Y?
               co := 1;            // For re-traversing X list for each Y.
               cStartX := -32767;  // Coords of active X segment of clip region;
               cEndX   := -32767;  // Minus infinity to start with.
            end;
            if (vps <> nil) then begin
               if vo = 0 then      // Find the first appropriate scanline.
                  fastForward(vps, y + pbr.top);
               if (vps[vo] <> EndMarker) and (y >= vps[vo] - pbr.top) then
                  vps += vo;       // Time to move the Y?
               vo := 1;            // For re-traversing X list for each Y.
               vStartX := -32767;  // Coords of active X segment of vis region;
               vEndX   := -32767;  // Minus infinity to start with.
            end;
            for x := br.left - pbr.left to br.right - pbr.left - 1 do begin
               while (cps <> nil) and
                     (x >= cEndX) and (cps[co] <> EndMarker) do begin
                  cStartX := cps[co] - pbr.left; // Get the X segment of clip rgn
                  cEndX := cps[co+1] - pbr.left; // for where we are right now.
                  co += 2;
               end;
               while (vps <> nil) and
                     (x >= vEndX) and (vps[vo] <> EndMarker) do begin
                  vStartX := vps[vo] - pbr.left; // Get the X segment of vis rgn
                  vEndX := vps[vo+1] - pbr.left; // for where we are right now.
                  vo += 2;
               end;
               if ((cps = nil) or ((x >= cStartX) and (x < cEndX))) and // Clip.
                  ((vps = nil) or ((x >= vStartX) and (x < vEndX))) then begin
                  patBit := (tp^.pnPat.bytes[y and 7] and
                             (1 << (7 - (x and 7)))) <> 0;
                  if tp^.pnMode.i > ord(srcBic) then
                     patBit := not patBit;
                  case tp^.pnMode.e of
                     notSrcCopy,
                     srcCopy:    if patBit then iSetBit(x, y)
                                           else iClearBit(x, y);
                     notSrcOr,
                     srcOr:      if patBit then iSetBit(x, y);
                     notSrcXor,
                     srcXor:     if patBit then iFlipBit(x, y);
                     notSrcBic,
                     srcBic:     if patBit then iClearBit(x, y);
                  else exit;
                  end;
               end;
            end;
            while (cps <> nil) and (cps[co] <> EndMarker) do // Get to end of X
               co += 2;            // segments, if not already there.
            inc(co);               // Point to next Y marker.
            while (vps <> nil) and (vps[vo] <> EndMarker) do // Get to end of X
               vo += 2;            // segments, if not already there.
            inc(vo);               // Point to next Y marker.
         end;
      end;
   end;
end;

// EraseRect paints the specified rectangle with the current grafPort's
// background pattern bkPat (in srcCopy mode).  The grafPort's pnPat and pnMode
// are ignored; the pen location is not changed.
//
procedure EraseRect(r : Rect);
begin
   if thePort = nil then exit;
   FillRect(r, thePort^.bkPat);
end;

// InvertRect inverts the pixels enclosed by the specified rectangle.  Every
// white pixel becomes black and every black pixel becomes white.  The
// grafPort's pnPat, pnMode and bkPat are ignored; the pen location is not
// changed.
//
procedure InvertRect(r : Rect);
var penSave : PenState;
begin
   if thePort = nil then exit;
   GetPenState(penSave);
   PenMode(srcXor);
   PenPat(black);
   PaintRect(r);
   SetPenState(penSave);
end;

// FillRect fills the specified rectangle with the given pattern (in srcCopy
// mode).  The grafPort's pnPat, pnMode, and bkPat are all ignored; the pen
// location is not changed.
//
procedure FillRect(r : Rect; fillPat : Pattern);
var br : Rect;          // Bounding rectangle, in global coordinates.
    pbr : Rect;
    co, vo, x, y : integer;
    patBit : Boolean;
    tp : GrafPtr;
    cps, vps : ^smallInt; // Clip and Vis Rgn data pointers; Nil if inactive.
    cStartX, cEndX, vStartX, vEndX : smallInt;
begin
   tp := thePort;
   if tp = nil then exit;
   if (tp^.pnVis < 0) then exit;
   co := 0; cEndX := 0;
   vo := 0; vEndX := 0;
   pbr := tp^.portBits.bounds;  // We need the logical portBits.bounds for
   if tp^.device = 1 then       // clipping, and moving to global coordinates.
      RotRect(pbr, pbr);
   if SectRect(r, tp^.portRect, br) then begin
      if SectRect(br, pbr, br) then begin
         cps := nil;
         if (tp^.clipRgn <> nil) and (tp^.clipRgn^ <> nil) then begin
            if not SectRect(br, tp^.clipRgn^^.rgnBBox, br) then exit;
            if tp^.clipRgn^^.rgnSize <> sizeof(RegionHdr) then
               cps := @tp^.clipRgn^^.rgnData;
            co := 0;               // Offset to Y marker.
         end;
         vps := nil;
         if (tp^.visRgn <> nil) and (tp^.visRgn^ <> nil) then begin
            if not SectRect(br, tp^.visRgn^^.rgnBBox, br) then exit;
            if tp^.visRgn^^.rgnSize <> sizeof(RegionHdr) then
               vps := @tp^.visRgn^^.rgnData;
            vo := 0;               // Offset to Y marker.
         end;
         for y := br.top - pbr.top to br.bottom - pbr.top - 1 do begin
            if (cps <> nil) then begin
               if co = 0 then      // Find the first appropriate scanline.
                  fastForward(cps, y + pbr.top);
               if (cps[co] <> EndMarker) and (y >= (cps[co] - pbr.top)) then
                  cps += co;       // Time to move the Y?
               co := 1;            // For re-traversing X list for each Y.
               cStartX := -32767;  // Coords of active X segment of clip region;
               cEndX   := -32767;  // Minus infinity to start with.
            end;
            if (vps <> nil) then begin
               if vo = 0 then      // Find the first appropriate scanline.
                  fastForward(vps, y + pbr.top);
               if (vps[vo] <> EndMarker) and (y >= vps[vo] - pbr.top) then
                  vps += vo;       // Time to move the Y?
               vo := 1;            // For re-traversing X list for each Y.
               vStartX := -32767;  // Coords of active X segment of vis region;
               vEndX   := -32767;  // Minus infinity to start with.
            end;
            for x := br.left - pbr.left to br.right - pbr.left - 1 do begin
               while (cps <> nil) and
                     (x >= cEndX) and (cps[co] <> EndMarker) do begin
                  cStartX := cps[co] - pbr.left; // Get the X segment of clip rgn
                  cEndX := cps[co+1] - pbr.left; // for where we are right now.
                  co += 2;
               end;
               while (vps <> nil) and
                     (x >= vEndX) and (vps[vo] <> EndMarker) do begin
                  vStartX := vps[vo] - pbr.left; // Get the X segment of vis rgn
                  vEndX := vps[vo+1] - pbr.left; // for where we are right now.
                  vo += 2;
               end;
               if ((cps = nil) or ((x >= cStartX) and (x < cEndX))) and // Clip.
                  ((vps = nil) or ((x >= vStartX) and (x < vEndX))) then begin
                  patBit := (fillPat.bytes[y and 7] and
                            (1 << (7 - (x and 7)))) <> 0;
                  if patBit then
                     iSetBit(x, y)
                  else
                     iClearBit(x, y);
               end;
            end;
            while (cps <> nil) and (cps[co] <> EndMarker) do // Get to end of X
               co += 2;            // segments, if not already there.
            inc(co);               // Point to next Y marker.
            while (vps <> nil) and (vps[vo] <> EndMarker) do // Get to end of X
               vo += 2;            // segments, if not already there.
            inc(vo);               // Point to next Y marker.
         end;
      end;
   end;
end;

//
// MoveTo()
//
// Move the point, i.e. current x and y.
//

// Moves the drawing pen position to location (X, Y).  No drawing is performed.
//
procedure MoveTo(x, y : smallInt);
var tp : GrafPtr;
    ph : PolyHandle;
begin
   tp := thePort;
   if tp = nil then exit;
   if tp^.rgnSave <> nil then begin // Accumulating a region?
      ph := tp^.polySave;
      if ph <> nil then begin   // If accumulating a polygon,
         ClosePoly;
         if ph^^.polySize > sizeof(Polygon) then // Is it
            iPolyIntoRgn(ph);   // big enough to use?
         KillPoly(ph);          // Yes, put it in too.
      end;
      OpenPoly;
   end;
   if tp^.polySave <> nil then  // If accumulating a polygon,
      iAdd2Poly(x, y);          // add in this point.
   tp^.pnLoc.v := y;
   tp^.pnLoc.h := x;
end;

// Moves the drawing pen position (X, Y) relative to its current position.
// Positive directions are to the right and down.  No drawing is performed.
//
procedure Move(x, y : smallInt);
var tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   MoveTo(tp^.pnLoc.h + x, tp^.pnLoc.v + y);
end;

//
// SetPixel()
//
// Function to plot a single point.
//
// Equivalent to:
//
//     MoveTo(H, V);
//     Line(0, 0);
//
// This call is slightly more efficient than its equivalent.  It is
// intended to be used for plotting a mathematical function or the
// like.  If you have a  large number of pixels to set you'll get
// *much* quicker results if you do your own point plotting in an
// off-screen bitmap (this isn't really all that hard to do) and then
// use CopyBits to show the result on the screen.
//
procedure SetPixel(h, v : smallInt);
var scratch : Rect;
    tp : GrafPtr;
begin
   tp := thePort;
   if tp = nil then exit;
   if (tp^.pnVis < 0) then exit;
   SetRect(scratch, h, v, h+1, v+1);
   FillRect(scratch, black);
end;

// GetPixel looks at the pixel associated with the given coordinate
// point and returns True if it's black or False if it's white.  The
// selected pixel is immediately below and to the right of the point
// whose coordinates are given in h and v, in the local coordinates of
// the current grafPort.  There's no guarantee that the specified pixel
// actually belongs to the port, however, it may have been drawn by a
// port overlapping the current one.  To see if the point indeed belongs
// to the current port, you could call PtInRgn(pt, thePort^.visRgn).
//
function GetPixel(h, v : smallInt) : Boolean;
var mask : byte;
    src : Pbyte;
    tp : GrafPtr;
begin
   GetPixel := False;
   tp := thePort;
   if tp = nil then exit;
   if tp^.device = 1 then begin // 3.7 inch E-paper?
      h := tp^.portBits.bounds.bottom - 1 - h;
      v := v - tp^.portBits.bounds.left;
      if (h < 0) or (v < 0) or (v >= tp^.portBits.bounds.right) or
         (h >= tp^.portBits.bounds.bottom) then exit;
      src := tp^.portBits.baseAddr;
      src += h * tp^.portBits.bounds.right div 8;
      src += v div 8;
      mask := 1 << (7 - (v and 7));
   end else begin            // Normal (offscreen) monochrome bitmap.
      v -= tp^.portBits.bounds.top;
      h -= tp^.portBits.bounds.left;
      if (h < 0) or (v < 0) or (h >= tp^.portBits.bounds.right) or
         (v >= tp^.portBits.bounds.bottom) then exit;
      src := tp^.portBits.baseAddr;
      src += v * tp^.portBits.rowBytes;
      src += h div 8;
      mask := 1 << (7 - (h and 7));
   end;
   GetPixel := (src^ and mask) <> 0;
end;

procedure iBuildLine(r : Rect; negative : Boolean; p : PsmallInt);
var x, y, deltax, deltay : smallInt;
   ii, jj, t1, t2, d : longint; // Avoid overflow.
begin
   x := 0;
   y := 0;
   deltax := r.right - r.left;
   deltay := r.bottom - r.top;

   if deltax = deltay then begin // Special case 45-degree diagonals.
      ii := 0;
      inc(deltax);
      while deltax > 0 do begin
         p[0] := ii;
         inc(p);
         inc(ii);
         dec(deltax);
      end;
   end else begin               // Non-45-degree diagonal line.
      d := deltay - deltax;     // Initial difference used as decision var.
      t1 := 2 * deltax;         // X step amount.
      t2 := 2 * deltay;         // Y step amount.

      jj := deltax + deltay;
      for ii := 0 to jj - 1 do begin
         if negative and (deltax > deltay) then begin // Hokey shit to make this
            p[y] := x;          // line drawer do the same lines as the
{$IFDEF DEBUG}			// optimized one.  This could use some work!
            Write('P[', y, '] = ', x, ', ');
{$ENDIF}
         end;
         if d > 0 then begin    // If next y closer to line,
            if deltay > deltax then begin
               p[y] := x;
{$IFDEF DEBUG}
               Write('P[', y, '] = ', x, ', ');
{$ENDIF}
            end;
            inc(y);             // plot next Y coordinate and
            d -= t1;            // step in X direction.
         end else begin         // Else next x is closer to line.
            if not negative and (deltax > deltay) then begin
               p[y] := x;
{$IFDEF DEBUG}
               Write('P[', y, '] = ', x, ', ');
{$ENDIF}
            end;
            inc(x);             // Plot next X coord and
            d += t2;            // step in Y direction.
         end;
      end;
      p[deltay] := deltax;
{$IFDEF DEBUG}
      WriteLn('P[', deltay, '] = ', deltax, '.');
{$ENDIF}
   end;
end;

//
// BuildLRgn builds a fake region data structure from a list of data points
// (provided by iBuildLine) that describe a line.
//
procedure iBuildLRgn(r : Rect;          // What to make a line region from.
                     neg : Boolean;     // Flag for negative-slope lines.
                     line,              // Line control point list.
                     p : PsmallInt      // Where to make it.
);
var
   ii, jj, start, stop, nullpt : smallInt;
   save1, save2, lastpt : ^smallInt;
   tempr : Rect;
begin
   save1 := p;
   inc(p);                      // Skip over the size field.
   tempr := r;
   inc(tempr.right);
   inc(tempr.bottom);
   PRect(p)^ := tempr;          // Dummy region's bounding box.
   inc(PRect(p));
   save2 := p;
   jj := tempr.bottom - tempr.top; // How many scanlines to do.

   nullpt := EndMarker;         // So we get the first region segment.
   lastpt := @nullpt;
   stop := 0;
   for ii := 0 to jj - 1 do begin
      if line^ <> lastpt^ then begin // Don't output unnecessary region segs.
         lastpt := line;
         p^ := tempr.top + ii;
         inc(p);
         start := stop;
         stop := line^;
         if neg then begin      // Handle reversed-slope lines.
            if (stop - start) < 1 then
               stop := start + 1;
            if ii = (jj - 1) then // On last pass thru, use edge fer sure.
               p^ := tempr.left
            else
               p^ := tempr.right - stop; // Start scanline.
            inc(p);
            p^ := tempr.right - start; // Stop scanline.
            inc(p);
         end else begin         // Positive sloped line.
            if (stop - start) < 1 then
               stop := start + 1
            else
               inc(stop);
            p^ := tempr.left + start; // Start scanline.
            inc(p);
            p^ := tempr.left + stop;  // Stop scanline.
            inc(p);
         end;
         p^ := EndMarker;       // End-of-scanline marker.
         inc(p);
      end;
      inc(line);
   end;
   if save2 = p then            // No actual data?
      save1^ := sizeof(RegionHdr) // Re-size it then.
   else begin
      p^ := EndMarker;          // End-of-region marker.
      inc(p);
      save1^ := PByte(p) - PByte(save1); // Plug in real size.
   end;
end;

//
// Internal routine to accumulate a point into a polygon.
//
procedure iAdd2Poly(x, y : smallInt);
var
   p : ^smallInt;
   polysize : integer;
   tp : GrafPtr;
begin
   tp := thePort;
   if (tp = nil) or (tp^.polySave = nil) then exit; // Sanity check.
   polysize := tp^.polySave^^.polySize;
   SetHandleSize(Handle(tp^.polySave), polysize + sizeof(Point));
   if tp^.polySave^ <> nil then begin
      p := @tp^.polySave^^.polySize;
      p^ += sizeof(Point);      // Mark with new size.
      p += polySize div sizeof(smallInt);
      PPoint(p)^.h := x;        // Put in the new point.
      PPoint(p)^.v := y;
   end else                     // Resize failed, close accumulation.
      tp^.polySave := nil;
end;

//
// Line() & LineTo()
//
// Functions to draw a line using Bresenham's algorithm.
//

// Like LineTo, but draws a line (X, Y) from the current pen position.
// Positive directions are to the right and down.
//
procedure Line(x, y : smallInt);
begin
   if thePort = nil then exit;
   LineTo(thePort^.pnLoc.h + x, thePort^.pnLoc.v + y);
end;

procedure VLine(height : smallInt);
var scratch : Rect;     // Convert to FillRect/PaintRect call.
begin
   if thePort = nil then exit;
   if (height >= 0) then begin
      scratch.top := thePort^.pnLoc.v;
      scratch.bottom := scratch.top + height;
   end else begin
      scratch.bottom := thePort^.pnLoc.v;
      scratch.top := scratch.bottom + height;
   end;
   scratch.left := thePort^.pnLoc.h;
   scratch.bottom += thePort^.pnSize.v;
   scratch.right := scratch.left + thePort^.pnSize.h;
   if (thePort^.pnMode.e = srcCopy) then
      FillRect(scratch, thePort^.pnPat) // Faster than PaintRect
   else                         // and they are the same operation when
      PaintRect(scratch);       // using srcCopy mode.
end;

procedure HLine(width : smallInt);
var scratch : Rect;     // Convert to FillRect/PaintRect call.
begin
   if thePort = nil then exit;
   if (width >= 0) then begin
      scratch.left := thePort^.pnLoc.h;
      scratch.right := scratch.left + width;
   end else begin
      scratch.right := thePort^.pnLoc.h;
      scratch.left := scratch.right + width;
   end;
   scratch.right += thePort^.pnSize.h;
   scratch.top := thePort^.pnLoc.v;
   scratch.bottom := scratch.top + thePort^.pnSize.v;
   if (thePort^.pnMode.e = srcCopy) then
      FillRect(scratch, thePort^.pnPat) // This's faster than PaintRect
   else                         // and they are the same operation when
      PaintRect(scratch);       // using srcCopy mode.
end;

//
// Bresenham's algorithm for the optimal cases.
//
// Line argument is 2 independent points, not a rect.
//
// We can only draw the line if it's a 1x1 pen, and is entirely
// unclipped, especially if it might be clipped by a complex region.
// (Our region traversal code only works top-to-bottom, left-to-right,
//  whereas we can draw our fast lines in any direction.)
//
function FastLine(br, line : Rect) : Boolean;
var tp : GrafPtr;
    pbr, cbr, vbr, br2 : Rect;
    dx, dy, sx, sy, err, e2 : smallInt;
    patBit : Boolean;
begin
   tp := thePort;
   if (tp^.pnSize.h <> 1) or (tp^.pnSize.v <> 1) then exit(False);
   pbr := tp^.portBits.bounds;  // We need the logical portBits.bounds for
   if tp^.device = 1 then       // clipping, and moving to global coordinates.
      RotRect(pbr, pbr);
   if not SectRect(br, pbr, br2) then exit(True); // Nothing left, done.
   if not EqualRect(br, br2) then exit(False); // If clipped, we can't do it.
   if (tp^.clipRgn <> nil) and (tp^.clipRgn^ <> nil) then begin
      cbr := tp^.clipRgn^^.rgnBBox;       // Has Clip region.
      if not SectRect(br2, cbr, br2) then exit(True); // Nothing left, done.
      if not EqualRect(br, br2) then exit(False); // Clipped, can't do it.  Or if
      if tp^.clipRgn^^.rgnSize <> sizeof(RegionHdr) then exit(False); // complex.
   end;
   if (tp^.visRgn <> nil) and (tp^.visRgn^ <> nil) then begin
      vbr := tp^.visRgn^^.rgnBBox;        // Has Visible region.
      if not SectRect(br2, vbr, br2) then exit(True); // Nothing left, done.
      if not EqualRect(br, br2) then exit(False); // Clipped, can't do it.  Or if
      if tp^.visRgn^^.rgnSize <> sizeof(RegionHdr) then exit(False); // complex.
   end;
   OffsetRect(line, -pbr.left, -pbr.top); // Move to global coordinates.

   // We survived the gauntlet, draw the (unclipped) line.
   if line.left < line.right then begin // dx := abs(x1-x0);
      sx := 1;
      dx := line.right - line.left;
   end else begin
      sx := -1;
      dx := line.left - line.right;
   end;
   if line.top < line.bottom then begin // dy := -abs(y1-y0);
      sy := 1;
      dy := line.top - line.bottom;
   end else begin
      sy := -1;
      dy := line.bottom - line.top;
   end;
   err := dx + dy;

   while (true) do begin
      patBit := (tp^.pnPat.bytes[line.top and 7] and
                 (1 << (7 - (line.left and 7)))) <> 0;
      if tp^.pnMode.i > ord(srcBic) then
         patBit := not patBit;
      case tp^.pnMode.e of
         notSrcCopy,
         srcCopy:    if patBit then   iSetBit(line.left, line.top)
                               else iClearBit(line.left, line.top);
         notSrcOr,
         srcOr:      if patBit then iSetBit(line.left, line.top);
         notSrcXor,
         srcXor:     if patBit then iFlipBit(line.left, line.top);
         notSrcBic,
         srcBic:     if patBit then iClearBit(line.left, line.top);
      end;
      if (line.left = line.right) and (line.top = line.bottom) then break;
      e2 := 2 * err;
      if e2 >= dy then begin // e_xy+e_x > 0
         err += dy;
         line.left += sx;
      end;
      if e2 <= dx then begin // e_xy+e_y < 0
         err += dx;
         line.top += sy;
      end;
   end;
   FastLine := True;            // We handled the line.
end;

// Draws a line from the current drawing pen position to (NEWX, NEWY).
// The pen is moved to (NEWX, NEWY) afterwards.  The pen size,
// transfer mode, and pattern are all used when drawing the line.
// A line distance of 0 will result in a single 'touch' of the pen to
// the drawing surface (but no motion), so a pen-sized rectangle will
// be drawn at the very least.
//
procedure LineTo(newx, newy : smallInt);
var tp : GrafPtr;
    ii, deltah : integer;
    p, p1, p2 : PsmallInt;
    negative : Boolean;
    size1, size2 : longint;
    dummyRgn : RgnPtr;
    line, scratch : Rect;
    scr : Point;
    ph : PolyHandle;
type
    TPoly = packed record       // Temporary polygon for making fat
       phdr : PolyHdr;          // penned lines.  We build a polygon
       pts  :array [0..6] of Point; // ourself to avoid recursion bugs that
    end;
var temppoly : TPoly;           // we'd get by calling the collection
    pop : PolyPtr;              // routines with MoveTo/LineTo.
    pp : ^Point;
    jj, wid, hei : integer;
begin
   tp := thePort;
   if tp = nil then exit;

   if tp^.rgnSave <> nil then begin // Accumulating a region?
      ph := tp^.polySave;
      if ph <> nil then begin   // If accumulating a polygon,
         SetPt(scr, newx, newy);
         if (ph^^.polyPoints[0].h = scr.h) and
            (ph^^.polyPoints[0].v = scr.v) then begin
            ClosePoly();        // Closure?  Yes, close it & accum it.
            if ph^^.polySize > sizeof(Polygon) then // Is it
               iPolyIntoRgn(ph); // big enough to use?
            KillPoly(ph); // Yes, put it in too.
            OpenPoly();
            iAdd2Poly(newx, newy); // Start again with this point.
         end;
      end;
   end;
   if tp^.polySave <> nil then // If accumulating a polygon,
      iAdd2Poly(newx, newy);  // add in this point.

   if (tp^.pnVis < 0) then begin
      tp^.pnLoc.h := newx;
      tp^.pnLoc.v := newy;
      exit;
   end;
   if (newy = tp^.pnLoc.v) then begin // Horizontal line?
      ii := newx - tp^.pnLoc.h;
      HLine(ii);                // Special case.
   end else if (newx = tp^.pnLoc.h) then begin // Vertical line?
      ii := newy - tp^.pnLoc.v;
      VLine(ii);                // Special case.
   end else begin               // Angled line.
      line.top := tp^.pnLoc.v;
      line.left := tp^.pnLoc.h;
      line.bottom := newy;
      line.right := newx;
      Pt2Rect(line.topLeft, line.botRight, scratch);
      scratch.bottom += tp^.pnSize.v;
      scratch.right += tp^.pnSize.h;
      if (SectRect(scratch, tp^.portRect, scratch) and not
          FastLine(scratch, line)) then begin // Can optimizer do it for us?
         negative := True xor (line.right  < line.left)
                          xor (line.bottom > line.top);
         Pt2Rect(line.topLeft, line.botRight, line);
      // Make sure line's not too big (tall) for us to draw.  We can run
      // out of stack space quite easily on these big honkers.  Trim down
      // the line in a cheesy fashion for safety.  The results won't be
      // exact, but it's close enough for us.
         ii := tp^.portRect.top - line.top - 200;
         if (ii > 0) then begin // Too far off top?  Trim it down some.
            deltah := ScaleS(line.right - line.left,
                             ii, line.bottom - line.top);
            if (negative) then
               line.right -= deltah
            else
               line.left += deltah;
            line.top += ii;
         end;
         ii := line.bottom - tp^.portRect.bottom - 200;
         if (ii > 0) then begin // Too far off bottom?  Trim it down some.
            deltah := ScaleS(line.right - line.left,
                             ii, line.bottom - line.top);
            if (negative) then
               line.left += deltah
            else
               line.right -= deltah;
            line.bottom -= ii;
         end;
         ii := tp^.pnSize.v;
         if ii > 1 then begin   // Use polygons for tall pens.
            jj := tp^.pnSize.h;
            wid := line.right + jj;
            hei := line.bottom + ii;
            temppoly.phdr.polySize := sizeof(temppoly);
            temppoly.phdr.polyBBox := line;
            temppoly.phdr.polyBBox.right := wid;
            temppoly.phdr.polyBBox.bottom := hei;
            pp := @temppoly.pts[0];
            if negative then begin // Negative slope, downleft to upright.
               pp^.v := line.bottom; pp^.h := line.left; inc(pp);
               pp^.v := line.top; pp^.h := line.right; inc(pp);
               pp^.v := line.top; pp^.h := wid; inc(pp);
               pp^.v := line.top + ii; pp^.h := wid; inc(pp);
               pp^.v := hei; pp^.h := line.left + jj; inc(pp);
               pp^.v := hei; pp^.h := line.left; inc(pp);
               pp^.v := line.bottom; pp^.h := line.left;
            end else begin      // Positive slope, upleft to downright.
               pp^ := line.topLeft; inc(pp);
               pp^.v := line.top; pp^.h := line.left + jj; inc(pp);
               pp^.v := line.bottom; pp^.h := wid; inc(pp);
               pp^.v := hei; pp^.h := wid; inc(pp);
               pp^.v := hei; pp^.h := line.right; inc(pp);
               pp^.v := line.top + ii; pp^.h := line.left; inc(pp);
               pp^ := line.topLeft;
            end;
            pop := PolyPtr(@temppoly);
            PaintPoly(@pop);
         end else begin      // 1x1 or 1xWide pens, quicker this way.
            size1 := (1 + line.bottom - line.top) * sizeof(smallInt);
            p1 := GetMem(size1); // # of line pts (max 1 per scan).
            iBuildLine(line, negative, p1);
            size2 := (size1 * 4) + sizeof(RegionHdr) + sizeof(smallInt);
            p2 := GetMem(size2); // Size of dummy region.
            iBuildLRgn(line, negative, p1, p2);
            dummyRgn := RgnPtr(p2);
            if tp^.pnSize.h > 1 then begin // Must we make region wider?
               ii := tp^.pnSize.h - 1;
               dummyRgn^.rgnBBox.right += ii;
               p := @dummyRgn^.rgnData;
               while p^ <> EndMarker do begin // Do each scan of the region.
                  p[2] += ii;   // Stretch width by bumping the
                  p += 4;       // ending X coordinate.
               end;
            end;
            PaintRgn(@dummyRgn);
            FreeMem(p2);
            FreeMem(p1);
         end;
      end;
   end;
   tp^.pnLoc.h := newx;
   tp^.pnLoc.v := newy;
end;

// CopyBits transfers a bit image between any two bit maps and clips the
// result to the area specified by the maskRgn parameter.  The transfer may be
// performed in any of the eight source transfer modes.  The result is always
// clipped to the maskRgn and the boundary rectangle of the destination bit
// map; if the destination bit map is the current grafPort's portBits, it's
// also clipped to the intersection of the grafPort's clipRgn and visRgn.  If
// you don't want to clip to a maskRgn, just pass NIL for the maskRgn parm.
// The dstRect and maskRgn are in terms of the dstBits.bounds coordinate
// system, and the srcRect coordinates are in terms of the srcBits.bounds
// coordinates.
//
// Warning:  If you perform a CopyBits between two grafPorts that overlap, you
// must first convert them both to the same coordinate system (global
// coordinates are good), and then specify the dstBits bitMap for both srcBits
// _and_ dstBits.  (Otherwise CopyBits won't realize they overlap, and will
// most likely mangle the copy.)  This is most commonly found when both
// bitMaps are part of the screen, and you are moving things around.
//
// CopyBits can shrink or expand by giving it different-sized rectangles.
// When shrinking, all source bits end up reflected in the destination.
// That tends to make things a bit blobby and dark under extreme scales.
//
// NB: The main arguments are var so that they are pass-by-reference (pointers),
//     but we do NOT modify them at any time!
//
procedure CopyBits(var srcBits, dstBits : BitMap; var srcRect, dstRect : Rect;
                   mode : xferMode; maskRgn : RgnHandle);
var sorigwid : smallInt; // Original width of source rect.
   sorighei : smallInt; // Original height of source rect.
   dorigwid : smallInt; // Original width of dest rect.
   dorighei : smallInt; // Original height of dest rect.
   clipSRect : Rect;    // Clipped source rectangle.
   clipDRect : Rect;    // Clipped destination rectangle.
   scratch : Rect;      // Scratch rectangle.
   dbr : Rect;          // Destination BitMap bounding rectangle.
   mps, cps, vps : ^smallInt; // Region data pointers; Nil if inactive.
   cStartX, cEndX, vStartX, vEndX, mStartX, mEndX : smallInt;
   co, vo, mo : integer;// Region data offsets.
   sOnScrn : Boolean;   // True if source is on screen (thePort).
   dOnScrn : Boolean;   // True if destination is on screen (thePort).
   srcBit : Boolean;
   smooth2x, smooth3x : Boolean;
   hratio : Fixed;      // Horizontal stretching ratio (fixed point).
   vratio : Fixed;      // Vertical stretching ratio (fixed point).
   delta, isx, tsx, rsx, isy, tsy, dx, dy : smallInt;
   sx, sy : Fixed;
   tp, savePort : GrafPtr;

   function srcBitIsSet(x, y : integer) : Boolean;
   var mask : byte;
       src : Pbyte;
       sm : ^BitMap;
   begin
      sm := @srcBits;
      if sOnScrn and (tp <> nil) and (tp^.device = 1) then begin // 3.7 inch E-paper?
         x := sm^.bounds.bottom - 1 - x;
         y := y - sm^.bounds.left;
         if (y < 0) then        // Stay on the e-Paper!
            y := 0;
         if (x < 0) then
            x := 0;
         if y >= sm^.bounds.right then
            y := sm^.bounds.right - 1;
         if x >= sm^.bounds.bottom then
            x := sm^.bounds.bottom - 1;
         x *= sm^.bounds.right div 8;
         src := sm^.baseAddr + x + (y div 8);
         mask := 1 << (7 - (y and 7));
      end else begin            // Normal (offscreen) monochrome bitmap.
{$IF 0}
         if (y < 0) or (y >= sm^.bounds.bottom) or  // Stay on the bitmap!
            (x < 0) or (x >= sm^.bounds.right) then begin
            srcBitIsSet := False;
            exit;
         end;
{$ELSE}
         // The smoothers might try to go off the reservation.
         // Better edge effects pinned at edge rather than null.
         if y >= sm^.bounds.bottom then y := sm^.bounds.bottom - 1;
         y -= sm^.bounds.top;
         if x >= sm^.bounds.right then x := sm^.bounds.right - 1;
         x -= sm^.bounds.left;
         if y < 0 then y := 0;
         if x < 0 then x := 0;
{$ENDIF}
         src := sm^.baseAddr;   // Base address of source image.
         src += y * sm^.rowBytes;
         src += x div 8;
         mask := 1 << (7 - (x and 7));
      end;
      srcBitIsSet := (src^ and mask) <> 0;
   end;

   // These are for working in offscreen destinations.  Global coordinates.
   procedure setDstBit(x, y : smallInt);
   var mask : byte;
       d : Pbyte;
       bm : ^BitMap;
   begin
      bm := @dstBits;
      if (y < 0) or (y >= bm^.bounds.bottom - bm^.bounds.top) or       // Stay on
         (x < 0) or (x >= bm^.bounds.right - bm^.bounds.left) then exit; // BMap!
      d := bm^.baseAddr + (y * bm^.rowBytes) + (x div 8);
      mask := 1 << (7 - (x and 7));
      d^ := d^ or mask;
   end;

   procedure clearDstBit(x, y : smallInt);
   var mask : byte;
       d : Pbyte;
       bm : ^BitMap;
   begin
      bm := @dstBits;
      if (y < 0) or (y >= bm^.bounds.bottom - bm^.bounds.top) or       // Stay on
         (x < 0) or (x >= bm^.bounds.right - bm^.bounds.left) then exit; // BMap!
      d := bm^.baseAddr + (y * bm^.rowBytes) + (x div 8);
      mask := 1 << (7 - (x and 7));
      d^ := d^ and not mask;
   end;

   procedure flipDstBit(x, y : smallInt);
   var mask : byte;
       d    : Pbyte;
       bm   : ^BitMap;
   begin
      bm := @dstBits;
      if (y < 0) or (y >= bm^.bounds.bottom - bm^.bounds.top) or       // Stay on
         (x < 0) or (x >= bm^.bounds.right - bm^.bounds.left) then exit; // BMap!
      d := bm^.baseAddr + (y * bm^.rowBytes) + (x div 8);
      mask := 1 << (7 - (x and 7));
      d^ := d^ xor mask;
   end;

   // See: http://www.scale2x.it/algorithm for details.
   // Smoothing algorithm for exactly 2x blowup.
   procedure doSmooth2;
   var B, D, F, H : Boolean;
       pixNo : smallInt;
   begin
      // Source/destination arrays for Scale2x:
      //   A B C   The center pixel E
      //   D E F   is expanded into      E0 E1
      //   G H I   4 new pixels E0..E3   E2 E3
      B := srcBitIsSet(isx, isy - 1); // Get our source neighbors.
      D := srcBitIsSet(isx - 1, isy); // (EPX/Scale2x algorithm.)
      F := srcBitIsSet(isx + 1, isy);
      H := srcBitIsSet(isx, isy + 1);
      if (B <> H) and (D <> F) then begin
         pixNo := ((dy and 1) << 1) or (dx and 1);
         case pixNo of // E0..E3 in algorithm.
           0: if (D = B) then
                 srcBit := D;
           1: if (B = F) then
                 srcBit := F;
           2: if (D = H) then
                 srcBit := D;
           3: if (H = F) then
                 srcBit := F;
         end;
      end;
   end;

   // See: http://www.scale2x.it/algorithm for details.
   // Smoothing algorithm for exactly 3x blowup.
   procedure doSmooth3;
   var A, B, C, D, E, F, G, H, I, J, K, L, M : Boolean;
       pixNo : smallInt;
   begin
{$IF 0} // Scale3x does kind of a crappy job on 45-degree diagonals.
      // Source/destination arrays for Scale3x:
      //   A B C   The center pixel E    E0 E1 E2
      //   D E F   is expanded into      E3 E4 E5
      //   G H I   9 new pixels E0..E8   E6 E7 E8
      B := srcBitIsSet(isx,     isy - 1); // Get our source neighbors.
      D := srcBitIsSet(isx - 1, isy);     // (Scale3x algorithm.)
      F := srcBitIsSet(isx + 1, isy);
      H := srcBitIsSet(isx,     isy + 1);
      if (B <> H) and (D <> F) then begin
         A := srcBitIsSet(isx - 1, isy - 1); // Get the rest of the
         C := srcBitIsSet(isx + 1, isy - 1); // neighbors' values.
         E := srcBit;
         G := srcBitIsSet(isx - 1, isy + 1);
         I := srcBitIsSet(isx + 1, isy + 1);
         pixNo := (((dy - clipDRect.top) mod 3) * 3) +
                   ((dx - clipDRect.left) mod 3);
         case pixNo of // E0..E8 in algorithm.
           0: if (D = B) then
                 srcBit := D;
           1: if ((D = B) and (E <> C)) or
                 ((B = F) and (E <> A)) then
                 srcBit := B;
           2: if (B = F) then
                 srcBit := F;
           3: if ((D = B) and (E <> G)) or
                 ((D = H) and (E <> A)) then
                 srcBit := D;
           4:    srcBit := E;
           5: if ((B = F) and (E <> I)) or
                 ((H = F) and (E <> C)) then
                 srcBit := F;
           6: if (D = H) then
                 srcBit := D;
           7: if ((D = H) and (E <> I)) or
                 ((H = F) and (E <> G)) then
                 srcBit := H;
           8: if (H = F) then
                 srcBit := F;
         end;
      end;
{$ENDIF}
{$IF 0} // Scale3x does kind of a crappy job on 45-degree diagonals.
      // Source/destination arrays for Scale3x:
      //   A B C   The center pixel E    E0 E1 E2
      //   D E F   is expanded into      E3 E4 E5
      //   G H I   9 new pixels E0..E8   E6 E7 E8
      A := srcBitIsSet(isx - 1, isy - 1); // Get our source neighbors.
      B := srcBitIsSet(isx,     isy - 1); // (Scale3x algorithm.)
      C := srcBitIsSet(isx + 1, isy - 1);
      D := srcBitIsSet(isx - 1, isy);
      E := srcBit;
      F := srcBitIsSet(isx + 1, isy);
      G := srcBitIsSet(isx - 1, isy + 1);
      H := srcBitIsSet(isx,     isy + 1);
      I := srcBitIsSet(isx + 1, isy + 1);
      pixNo := ((dy mod 3) * 3) + (dx mod 3);
      case pixNo of // E0..E8 in algorithm.
// E0 = B=D & B!=F & D!=H ? 0.5*(B+D) : E
        0: if (B=D) and (B<>F) and (D<>H) then
              srcBit := D;
// E1 = (B=D & B!=F & D!=H & E!=C) | (B=F & B!=D & F!=H & E!=A) ? B : E;
        1: if ((B=D) and (B<>F) and (D<>H))
           or ((B=F) and (B<>D) and (F<>H) and (E<>A)) then
              srcBit := B;
// E2 = B=F & B!=D & F!=H ? 0.5*(B+F) : E
        2: if (B=F) and (B<>D) and (F<>H) then
              srcBit := F;
// E3 = (B=D & B!=F & D!=H & E!=G) | (D=H & B!=D & F!=H & E!=A) ? D : E;
        3: if ((B=D) and (B<>F) and (D<>H) and (E<>G))
           or ((D=H) and (B<>D) and (F<>H) and (E<>A)) then
              srcBit := D;
// E4 = E;
        4:    srcBit := E;
// E5 = (B=F & B!=D & F!=H & E!=I) | (F=H & B!=F & D!=H & E!=C) ? F : E;
        5: if ((B=F) and (B<>D) and (F<>H) and (E<>I))
           or ((F=H) and (B<>F) and (D<>H) and (E<>C)) then
              srcBit := F;
// E6 = D=H & B!=D & F!=H ? 0.5*(D+H) : E
        6: if (D=H) and (B<>D) and (F<>H) then
              srcBit := D;
// E7 = (D=H & B!=D & F!=H & E!=I) | (F=H & B!=F & D!=H & E!=G) ? H : E;
        7: if ((D=H) and (B<>D) and (F<>H) and (E<>I))
           or ((F=H) and (B<>F) and (D<>H) and (E<>G)) then
              srcBit := H;
// E8 = F=H & B!=F & D!=H ? 0.5*(F+H) : E
        8: if (F=H) and (B<>F) and (D<>H) then
              srcBit := F;
      end;
{$ENDIF}
{$IF 1}
      // Source/destination arrays for Scale3xSFX:
      //     J
      //   A B C    The center pixel E    E0 E1 E2
      // K D E F L  is expanded into      E3 E4 E5
      //   G H I    9 new pixels E0..E8   E6 E7 E8
      //     M
      A := srcBitIsSet(isx - 1, isy - 1); // Get our source neighbors.
      B := srcBitIsSet(isx,     isy - 1); // (Scale3xSFX algorithm.)
      C := srcBitIsSet(isx + 1, isy - 1);
      D := srcBitIsSet(isx - 1, isy);
      E := srcBit;
      F := srcBitIsSet(isx + 1, isy);
      G := srcBitIsSet(isx - 1, isy + 1);
      H := srcBitIsSet(isx,     isy + 1);
      I := srcBitIsSet(isx + 1, isy + 1);
      J := srcBitIsSet(isx,     isy - 2);
      K := srcBitIsSet(isx - 2, isy);
      L := srcBitIsSet(isx + 2, isy);
      M := srcBitIsSet(isx,     isy + 2);
      pixNo := ((dy mod 3) * 3) + (dx mod 3);
      case pixNo of // E0..E8 in algorithm.
// E0 = (B=D & B!=F & D!=H & (E!=A | E=C | E=G | A=J | A=K))
//    | (B=D & C=E & C!=J & A!=E)
//    | (B=D & E=G & A!=E & G!=K) ? 0.5*(B+D) : E
        0: if ((B=D) and (B<>F) and (D<>H) and
               ((E<>A) or (E=C) or (E=G) or (A=J) or (A=K)))
           or ((B=D) and (C=E) and (C<>J) and (A<>E))
           or ((B=D) and (E=G) and (A<>E) and (G<>K)) then
              srcBit := B;
// JEC: Poor logic: (A and (B or C) and not C) forms.  E vs C, E vs A.
//      Simplifies: (A and B and not C).
// E1 = (B=D & B!=F & D!=H & (E!=A | E=C | E=G | A=J | A=K) & E!=C)
//    | (B=F & B!=D & F!=H & (E!=C | E=A | E=I | C=J | C=L) & E!=A) ? B : E;
        1: if ((B=D) and (B<>F) and (D<>H) and
               ((E<>A) {or (E=C)} or (E=G) or (A=J) or (A=K)) and (E<>C))
           or ((B=F) and (B<>D) and (F<>H) and
               ((E<>C) {or (E=A)} or (E=I) or (C=J) or (C=L)) and (E<>A)) then
              srcBit := B;
// E2 = (B=F & B!=D & F!=H & (E!=C | E=A | E=I | C=J | C=L))
//    | (B=F & A=E & A!=J & C!=E)
//    | (B=F & E=I & C!=E & I!=L) ? 0.5*(B+F) : E
        2: if ((B=F) and (B<>D) and (F<>H) and
               ((E<>C) or (E=A) or (E=I) or (C=J) or (C=L)))
           or ((B=F) and (A=E) and (A<>J) and (C<>E))
           or ((B=F) and (E=I) and (C<>E) and (I<>L)) then
              srcBit := B;
// JEC: Poor logic: (A and (B or C) and not C) forms.  E vs G, E vs A.
//      Simplifies: (A and B and not C).
// E3 = (B=D & B!=F & D!=H & (E!=A | E=C | E=G | A=J | A=K) & E!=G)
//    | (D=H & B!=D & F!=H & (E!=G | E=A | E=I | G=K | G=M) & E!=A) ? D : E;
        3: if ((B=D) and (B<>F) and (D<>H) and
               ((E<>A) or (E=C) {or (E=G)} or (A=J) or (A=K)) and (E<>G))
           or ((D=H) and (B<>D) and (F<>H) and
               ((E<>G) {or (E=A)} or (E=I) or (G=K) or (G=M)) and (E<>A)) then
              srcBit := D;
        4:    srcBit := E;
// JEC: Poor logic: (A and (B or C) and not C) forms.  E vs C, E vs I.
//      Simplifies: (A and B and not C).
// E5 = (F=H & B!=F & D!=H & (E!=I | E=C | E=G | I=L | I=M) & E!=C)
//    | (B=F & B!=D & F!=H & (E!=C | E=A | E=I | C=J | C=L) & E!=I) ? F : E;
        5: if ((F=H) and (B<>F) and (D<>H) and
               ((E<>I) {or (E=C)} or (E=G) or (I=L) or (I=M)) and (E<>C))
           or ((B=F) and (B<>D) and (F<>H) and
               ((E<>C) or (E=A) {or (E=I)} or (C=J) or (C=L)) and (E<>I)) then
              srcBit := F;
// E6 = (D=H & B!=D & F!=H & (E!=G | E=A | E=I | G=K | G=M))
//    | (D=H & A=E & A!=K & E!=G)
//    | (D=H & E=I & E!=G & I!=M) ? 0.5*(D+H) : E
        6: if ((D=H) and (B<>D) and (F<>H) and
               ((E<>G) or (E=A) or (E=I) or (G=K) or (G=M)))
           or ((D=H) and (A=E) and (A<>K) and (E<>G))
           or ((D=H) and (E=I) and (E<>G) and (I<>M)) then
              srcBit := D;
// JEC: Poor logic: (A and (B or C) and not C) forms.  E vs G, E vs I.
//      Simplifies: (A and B and not C).
// E7 = (F=H & B!=F & D!=H & (E!=I | E=C | E=G | I=L | I=M) & E!=G)
//    | (D=H & B!=D & F!=H & (E!=G | E=A | E=I | G=K | G=M) & E!=I) ? H : E;
        7: if ((F=H) and (B<>F) and (D<>H) and
               ((E<>I) or (E=C) {or (E=G)} or (I=L) or (I=M)) and (E<>G))
           or ((D=H) and (B<>D) and (F<>H) and
               ((E<>G) or (E=A) {or (E=I)} or (G=K) or (G=M)) and (E<>I)) then
              srcBit := H;
// E8 = (F=H & B!=F & D!=H & (E!=I | E=C | E=G | I=L | I=M))
//    | (F=H & C=E & C!=L & E!=I)
//    | (F=H & E=G & E!=I & G!=M) ? 0.5*(F+H) : E
        8: if ((F=H) and (B<>F) and (D<>H) and
               ((E<>I) or (E=C) or (E=G) or (I=L) or (I=M)))
              or ((F=H) and (C=E) and (C<>L) and (E<>I))
              or ((F=H) and (E=G) and (E<>I) and (G<>M)) then
              srcBit := F;
      end;
{$ENDIF}
   end;

begin
   tp := thePort;
   sOnScrn := (tp <> nil) and (@srcBits = @tp^.portBits);
   dOnScrn := (tp <> nil) and (@dstBits = @tp^.portBits);
   if dOnScrn and (tp^.pnVis < 0) then exit;
   GetPort(savePort);
   vratio := Int2Fix(1);
   hratio := vratio;
   if EmptyRect(dstRect) or EmptyRect(srcRect) then exit;
   sorigwid := srcRect.right - srcRect.left;
   dorigwid := dstRect.right - dstRect.left;
   hratio := FixRatio(sorigwid, dorigwid);
   sorighei := srcRect.bottom - srcRect.top;
   dorighei := dstRect.bottom - dstRect.top;
   vratio := FixRatio(sorighei, dorighei);
   // Will we engage special smoothing logic?
   smooth2x := (hratio = FixRatio(1, 2)) and (vratio = FixRatio(1, 2));
   smooth3x := (hratio = FixRatio(1, 3)) and (vratio = FixRatio(1, 3));

// Trim the 'original' source & destination rectangles so that they're entirely
// within their bitmap bounds.  This is so we never touch any memory that's
// outside the bitmaps (which may well be read or write-protected).  We do our
// best to scale things so the right results will show.

   if sOnScrn then begin
      RotRect(srcBits.Bounds, scratch);
      if not SectRect(srcRect, scratch, clipSRect) then exit
   end else
      if not SectRect(srcRect, srcBits.Bounds, clipSRect) then exit;

// Original 68010 assy form, typical 1 of 4:
//         movw  top(clipSRect),d0
//         subw  top(srcRect),d0
//         beqs  clipd1            | Any of the top of the SrcRect clipped?
//         cmpw  dorighei,sorighei | Yes, trim off a proportionate amount of
//         beqs  clipd1            | the DstRect's top.
//         muls  dorighei,d0
//         divs  sorighei,d0
// clipd1: addw  top(dstRect),d0
//         movw  d0,top(scratch)

   delta := clipSRect.top - srcRect.top;   // Any of the top of srcRect clipped?
   if (delta <> 0) and (sorighei <> dorighei) then // Yes, trim off a
      delta := ScaleS(delta, dorighei, sorighei);  // proportionate amount
   scratch.top := delta + dstRect.top;             // of dstRect's top.

   delta := clipSRect.left - srcRect.left;     // Same for the left sides.
   if (delta <> 0) and (sorigwid <> dorigwid) then
      delta := ScaleS(delta, dorigwid, sorigwid);
   scratch.left := delta + dstRect.left;

   delta := clipSRect.bottom - srcRect.bottom; // Same for the bottoms.
   if (delta <> 0) and (sorighei <> dorighei) then
      delta := ScaleS(delta, dorighei, sorighei);
   scratch.bottom := delta + dstRect.bottom;

   delta := clipSRect.right - srcRect.right;   // Same for the right sides.
   if (delta <> 0) and (sorigwid <> dorigwid) then
      delta := ScaleS(delta, dorigwid, sorigwid);
   scratch.right := delta + dstRect.right;

// Find out how much of the (new) dest rectangle lies outside the dest
// BitMap, and trim the source rectangle by that amount so we never go
// off the dest BitMap.  Also scale by relative sizes for stretching.
//
// Any add'l clipping we do based on rectangles (that affects the size of the
// destination rectangle we will be traversing) also needs to be done here,
// so we can also trim the source traversal rectangle a proportionate amount.
// (Because of scaling.)
//
// No valid destination at all, and we're done.

   dbr := dstBits.bounds;
   if dOnScrn and (tp^.device = 1) then // Special destination coordinate
      RotRect(dbr, dbr);                // 90deg rotation hack.
   if not SectRect(scratch, dbr, clipDRect) then exit;
// Now take into account thePort's visRgn and clipRgn, and the maskRgn argument.
   mps := nil; cps := nil; vps := nil;
   co := 0; cEndX := 0;
   vo := 0; vEndX := 0;
   mo := 0; mEndX := 0;
   if dOnScrn then begin
      if not SectRect(clipDRect, tp^.portRect, clipDRect) then exit;
      if (tp^.clipRgn <> nil) and (tp^.clipRgn^ <> nil) then begin // Clip rgn?
         if not SectRect(clipDRect, tp^.clipRgn^^.rgnBBox, clipDRect) then exit;
         if tp^.clipRgn^^.rgnSize <> sizeof(RegionHdr) then
            cps := @tp^.clipRgn^^.rgnData;
         co := 0;               // Offset to Y marker.
      end;
      if (tp^.visRgn <> nil) and (tp^.visRgn^ <> nil) then begin // Vis region?
         if not SectRect(clipDRect, tp^.visRgn^^.rgnBBox, clipDRect) then exit;
         if tp^.visRgn^^.rgnSize <> sizeof(RegionHdr) then
            vps := @tp^.visRgn^^.rgnData;
         vo := 0;               // Offset to Y marker.
      end;
   end;
   if (maskRgn <> nil) and (maskRgn^ <> nil) then begin // Mask region?
      if not SectRect(clipDRect, maskRgn^^.rgnBBox, clipDRect) then exit;
      if maskRgn^^.rgnSize <> sizeof(RegionHdr) then
         mps := @maskRgn^^.rgnData;
      mo := 0;               // Offset to Y marker.
   end; // Now clipDRect holds our final dest rect, and scratch is the original.

// Original 68010 assy form, typical 1 of 4:
//         movw top(clipDRect),d0
//         subw top(scratch),d0
//         beqs clips1            | Any of the top of the new DstRect clipped?
//         cmpw dorighei,sorighei | Yes, trim off a proportionate amount of
//         beqs clips1            | the SrcRect's top.
//         muls sorighei,d0
//         divs dorighei,d0
// clips1: addw d0,top(clipSRect)

   delta := clipDRect.top - scratch.top; // Any of the top of new dest clipped?
   if (delta <> 0) and (sorighei <> dorighei) then // Yes, trim off a prop. amt
      clipSRect.top += ScaleS(delta, sorighei, dorighei); // of srcRect's top.

   delta := clipDRect.left - scratch.left;     // Same for the left sides.
   if (delta <> 0) and (sorigwid <> dorigwid) then
      clipSRect.left += ScaleS(delta, sorigwid, dorigwid);

   delta := clipDRect.bottom - scratch.bottom; // Same for the bottoms.
   if (delta <> 0) and (sorighei <> dorighei) then
      clipSRect.bottom += ScaleS(delta, sorighei, dorighei);

   delta := clipDRect.right - scratch.right;   // Same for the right sides.
   if (delta <> 0) and (sorigwid <> dorigwid) then
      clipSRect.right += ScaleS(delta, sorigwid, dorigwid);

   scratch := clipDRect;        // New-new Dest rect.
{$IF 0}
   // Print out the post-clip/scale stats when debugging.
   WriteLn('hratio: ', fixedS32ToStr(hratio, 65536, 6, False),
           ' vratio: ', fixedS32ToStr(vratio, 65536, 6, False),
           ' smooth2x: ', ord(smooth2x), ' smooth3x: ', ord(smooth3x));
   DumpRect(clipSRect, 's');
   DumpRect(clipDRect, 'd');
{$ENDIF}
// We run the outer loops as destination integers because due to the XOR
// transfer modes we never want to touch a destination pixel twice, even if
// we're scaling.  That makes (only) the source index logic tricky.  This
// also ensures that we touch every destination pixel that we should;
// Once and only once.  dX and dY are in global coordinates.
   dy := clipDRect.top - dbr.top;
   sy := Int2Fix(clipSRect.top) + (vratio div 4); // Epsilon avoids edge effects.
   while dy < clipDRect.bottom - dbr.top do begin // Move across the dest bitmap.
      if (cps <> nil) then begin
         if co = 0 then         // Find the first appropriate scanline.
            fastForward(cps, dy + dbr.top);
         if (cps[co] <> EndMarker) and (dy >= (cps[co] - dbr.top)) then
            cps += co;          // Time to move the Y?
         co := 1;               // For re-traversing X list for each Y.
         cStartX := -32767;     // Coords of active X segment of clip region;
         cEndX   := -32767;     // Minus infinity to start with.
      end;
      if (vps <> nil) then begin
         if vo = 0 then         // Find the first appropriate scanline.
            fastForward(vps, dy + dbr.top);
         if (vps[vo] <> EndMarker) and (dy >= vps[vo] - dbr.top) then
            vps += vo;          // Time to move the Y?
         vo := 1;               // For re-traversing X list for each Y.
         vStartX := -32767;     // Coords of active X segment of vis region;
         vEndX   := -32767;     // Minus infinity to start with.
      end;
      if (mps <> nil) then begin
         if mo = 0 then         // Find the first appropriate scanline.
            fastForward(mps, dy + dbr.top);
         if (mps[mo] <> EndMarker) and (dy >= mps[mo] - dbr.top) then
            mps += mo;          // Time to move the Y?
         mo := 1;               // For re-traversing X list for each Y.
         mStartX := -32767;     // Coords of active X segment of mask region;
         mEndX   := -32767;     // Minus infinity to start with.
      end;
      isy := Fix2Int(sy);
      sy += vratio;
      tsy := Fix2Int(sy) - 1;
      dx := clipDRect.left - dbr.left;
      sx := Int2Fix(clipSRect.left) + (hratio div 4); // Epsilon, ditto.
      while dx < clipDRect.right - dbr.left do begin
         isx := Fix2Int(sx);
         sx += hratio;
         tsx := Fix2Int(sx) - 1; // For traversing source X during shrinking.
         srcBit := srcBitIsSet(isx, isy);

         // Shrinking?  Incorporate, or at least traverse, all source bits.
         // Once the bit is set we can stop looking at the source bitmap.
         if isy < tsy then begin // Shrinking vertically?
            rsx := isx;         // Look at ALL pixels on each source line.
            while isy < tsy do begin
               inc(isy);        // Traverse to next needed source scanline.
               isx := rsx;      // Reload our X index.
               while isx < tsx do begin
                  inc(isx);     // Traverse to next needed source pixel.
                  if not srcBit then begin
                     srcBit := srcBitIsSet(isx, isy);
                  end;
               end;
            end;
         end else begin // Not shrinking vertically, shrinking horizontally?
            while isx < tsx do begin
               inc(isx); // Traverse to next needed source pixel.
               if not srcBit then begin
                  srcBit := srcBitIsSet(isx, isy);
               end;
            end;
         end;

         if smooth2x then doSmooth2;
         if smooth3x then doSmooth3;

         while (cps <> nil) and
               (dx >= cEndX) and (cps[co] <> EndMarker) do begin
            cStartX := cps[co] - dbr.left; // Get the X segment of clip rgn
            cEndX := cps[co+1] - dbr.left; // for where we are right now.
            co += 2;
         end;
         while (vps <> nil) and
               (dx >= vEndX) and (vps[vo] <> EndMarker) do begin
            vStartX := vps[vo] - dbr.left; // Get the X segment of vis rgn
            vEndX := vps[vo+1] - dbr.left; // for where we are right now.
            vo += 2;
         end;
         while (mps <> nil) and
               (dx >= mEndX) and (mps[mo] <> EndMarker) do begin
            mStartX := mps[mo] - dbr.left; // Get X segment of mask rgn
            mEndX := mps[mo+1] - dbr.left; // for where we are right now.
            mo += 2;
         end;
         if ((cps = nil) or ((dx >= cStartX) and (dx < cEndX))) and // Clip!
            ((vps = nil) or ((dx >= vStartX) and (dx < vEndX))) and
            ((mps = nil) or ((dx >= mStartx) and (dx < mEndx))) then begin

            if ord(mode) > ord(srcBic) then // Now transfer the bit to dest.
               srcBit := not srcBit;
            case mode of
              notSrcCopy,
              srcCopy:    if srcBit then begin
                             if dOnScrn then   iSetBit(dx, dy)
                                        else setDstBit(dx, dy);
                          end else begin
                             if dOnScrn then   iClearBit(dx, dy)
                                        else clearDstBit(dx, dy);
                          end;
              notSrcOr,
              srcOr:      if srcBit then begin
                             if dOnScrn then   iSetBit(dx, dy)
                                        else setDstBit(dx, dy);
                          end;
              notSrcXor,
              srcXOr:     if srcBit then begin
                             if dOnScrn then   iFlipBit(dx, dy)
                                        else flipDstBit(dx, dy);
                          end;
              notSrcBic,
              srcBic:     if srcBit then begin
                             if dOnScrn then   iClearBit(dx, dy)
                                        else clearDstBit(dx, dy);
                          end;
            end;
         end;
         inc(dx);
      end;
      while (cps <> nil) and (cps[co] <> EndMarker) do // Get to end of X
         co += 2;               // segments, if not already there.
      inc(co);                  // Point to next Y marker.
      while (vps <> nil) and (vps[vo] <> EndMarker) do // Get to end of X
         vo += 2;               // segments, if not already there.
      inc(vo);                  // Point to next Y marker.
      while (mps <> nil) and (mps[mo] <> EndMarker) do // Get to end of X
         mo += 2;               // segments, if not already there.
      inc(mo);                  // Point to next Y marker.
      inc(dy);
   end;
end;

// Icon Plotter.
//
// NB: The main arguments are var so that they are pass-by-reference (pointers),
//     but we do NOT modify them at any time!
//
procedure PlotIcon(var r : Rect; var iconptr : Icon);
var scratch : BitMap; // Dummy bitmap structure.
begin
   if thePort = nil then exit;
   scratch.bounds.left := 0;
   scratch.bounds.top := 0;
   scratch.bounds.right := 32;
   scratch.bounds.bottom := 32;
   scratch.baseAddr := @iconptr;
   scratch.rowBytes := 4;

   // Move the Icon.
   CopyBits(scratch, thePort^.portBits, scratch.bounds, r, srcCopy, nil);
end;

// Naive bitmap rotator.
//
function RotateBMap(srcBMap : BitMap; var dstBMap : BitMap; dir : smallInt) : Boolean;
var ImageSize : integer;
    sx, sy, dx, dy : smallInt;

   function getBit : Boolean;
   var mask : byte;
       s : Pbyte;
       xoff, yoff : smallInt;
   begin
      getBit := False;
      xoff := sx - srcBMap.bounds.left;
      yoff := sy - srcBMap.bounds.top;
      if (yoff < 0) or (sy >= srcBMap.bounds.bottom) or // Stay on the bitmap!
         (xoff < 0) or (sx >= srcBMap.bounds.right) then exit;
      s := srcBMap.baseAddr + (yoff * srcBMap.rowBytes) + (xoff div 8);
      mask := 1 << (7 - (xoff and 7));
      getBit := (s^ and mask) <> 0;
   end;

   procedure setBit;
   var mask : byte;
       d : Pbyte;
       xoff, yoff : smallInt;
   begin
      xoff := dx - dstBMap.bounds.left;
      yoff := dy - dstBMap.bounds.top;
      if (yoff < 0) or (dy >= dstBMap.bounds.bottom) or // Stay on the bitmap!
         (xoff < 0) or (dx >= dstBMap.bounds.right) then exit;
      d := dstBMap.baseAddr + (yoff * dstBMap.rowBytes) + (xoff div 8);
      mask := 1 << (7 - (xoff and 7));
      d^ := d^ or mask;
   end;

begin
   RotateBMap := False;
   RotRect(srcBMap.bounds, dstBMap.bounds);
   dstBMap.rowBytes := (RectWidth(dstBMap.bounds) + 7) div 8;
   imageSize := RectHeight(dstBMap.bounds) * dstBMap.rowBytes;
   dstBMap.baseAddr := GetMem(ImageSize);
   if dstBMap.baseAddr = nil then exit;
   RotateBMap := True;
   FillByte(dstBMap.baseAddr^, ImageSize, 0);
   if (dir > 0) then
      dx := dstBMap.bounds.left
   else
      dx := dstBMap.bounds.right - 1;
   for sy := srcBMap.bounds.top to srcBMap.bounds.bottom - 1 do begin
      if (dir > 0) then
         dy := dstBMap.bounds.bottom - 1
      else
         dy := dstBMap.bounds.top;
      for sx := srcBMap.bounds.left to srcBMap.bounds.right - 1 do begin
         if getBit then
            setBit;
         dy -= dir;
      end;
      dx += dir;
   end;
end;

// Given a source and destination bitMap, rotate it CCW 90 degrees.
// The destination bitmap will be _created_ by this function.  Returns
// TRUE if the operation succeeded, or FALSE if it failed (allocation error).
// The bitMap can be of any size.
//
// The 'interesting' area of the source bitMap (what was justified against
// the top & left of the source) will end up justified against the bottom
// left of the destination.
//
function RotateBMapCCW(srcBMap : BitMap; var dstBMap : BitMap) : Boolean;
begin
   RotateBMapCCW := RotateBMap(srcBMap, dstBMap, 1);
end;

// Ditto, but CW (clockwise).
//
function RotateBMapCW(srcBMap : BitMap; var dstBMap : BitMap) : Boolean;
begin
   RotateBMapCW := RotateBMap(srcBMap, dstBMap, -1);
end;

// Handle (indirect Pointer) logic.

// Unlike in the Mac, we currently only support one master pointer block.
// Mac will chain in new blocks as necessary.  We find that unnecessary.
//
function NewHandle(size : longInt) : Handle;
var ii : integer;
   p : ^Ptr;
begin
   NewHandle := nil;
   p := handles;
   if p = nil then begin        // No master pointer table for handles?
      ii := sizeof(Ptr) * NHANDLES;
      p := GetMem(ii);
      if p = nil then
         exit;
      FillByte(p^, ii, 0);
      handles := p;
   end;
   // Now that our initial housekeeping's done, search out an empty master
   // pointer for the request made of us.  Allocate a handle and return it
   // as well as filling that master pointer.
   for ii := 0 to NHANDLES - 1 do begin // Find an empty slot.
      p := @handles[ii];
      if p^ = nil then begin
         handles[ii] := GetMem(size);
         NewHandle := p;
         break;
      end;
   end;
end;

procedure DisposeHandle(var h : Handle); // Completely destroy the handle.
begin
   if h <> nil then begin               // Not if it isn't real.
      FreeMem(h^);
      h^ := nil;
      h := nil;
   end;
end;

// Resize the handle.
procedure SetHandleSize(h : handle; size : PtrUInt);
begin
   if (h <> nil) and (h^ <> nil) then // Not if it isn't real.
      ReAllocMem(h^, size);
end;

// How big is the handle?
function GetHandleSize(h : handle) : PtrUInt;
begin
   GetHandleSize := 0;
   if (h <> nil) and (h^ <> nil) then // Not if it isn't real.
      GetHandleSize := MemSize(h^);
end;

// We only have the basic region code here.  The more complex region handling
// is really only necessary for the optional Windowing or rounded-shapes
// packages, and so it has been pushed into a separate unit.

//
// NewRgn allocates space for a new, variable-sized region, initializes it to
// the empty region defined by the rectangle (0,0)(0,0), and returns a handle
// to the new region.
//
// WARNING:  Only this function creates new regions; all other routines just
//           alter the size and shape of existing regions.  Before a region's
//           handle can be passed to any drawing or calculation routine, space
//           must already have been allocated for the region.
//
function NewRgn : RgnHandle;
var p : RgnHandle;
begin
   p := RgnHandle(NewHandle(sizeof(RegionHdr)));
   if p <> nil then begin
      p^^.rgnSize := sizeof(RegionHdr);
      SetEmptyRgn(RgnHandle(p));        // Empty the region. */
   end;
   NewRgn := p;
end;

//
// OpenRgn tells Clunkdraw to allocate temporary space and start saving lines
// and framed shapes for later processing as a region definition.  While a
// region is open, all calls to Line, LineTo, and the procedures that draw
// framed shapes (except arcs and wedges) affect the outline of the region.
// Only the line endpoints and shape boundaries affect the region definition;
// the pen mode, pattern, and size do not affect it.  In fact, Clunkdraw calls
// HidePen, so no drawing occurs on the screen while the region is open (unless
// you called ShowPen just after OpenRgn, or you called ShowPen previously
// without balancing it by a call to HidePen).  Since the pen hangs below and
// to the right of the pen location, drawing lines with even the smallest pen
// will change bits that lie outside the region you define.
//
// The outline of a region is mathematically defined and infinitely thin, and
// separates the bit image into two groups of bits:  Those within the region
// and those outside it.  A region should consist of one or more closed loops.
// Each framed shape itself constitutes a loop.  Any lines draw with Line or
// LineTo should connect with each other or with a framed shape.  Even though
// the screen presentation of a region is clipped, the definition of a region
// is not; you can define a region anywhere on the coordinate plane with
// complete disregard for the location of various grafPort entities on
// that plane.
//
// When a region is open, the current grafPort's rgnSave field contains a
// handle to information related to the region definition.  If you want to
// temporarily disable the collection of lines and shapes, you can save the
// current value of this field, set the field to NULL, and later restore the
// saved value to resume the region definition.  Also, changing the current
// port by calling SetPort while a region is being formed will discontinue
// formation of the region until another call to SetPort resets the region's
// original grafPort.
//
// WARNING:  Do not call OpenRgn while another region or polygon is already
//           open.  All open regions but the most recent will behave strangely.
//
procedure OpenRgn;
begin
   HidePen();
   thePort^.rgnSave := NewRgn;
end;

//
// CloseRgn stops the collection of lines and framed shapes, organizes them
// into a region definition, and saves the resulting region in the region
// indicated by dstRgn.  CloseRgn does _not_ create the destination region;
// space must already have been allocated for it.  You should perform one and
// only one CloseRgn for every OpenRgn.  CloseRgn calls ShowPen, balancing the
// HidePen call made by OpenRgn.
//
procedure CloseRgn(var dstRgn : RgnHandle);
var tp : GrafPtr;
    ph : PolyHandle;
begin
   tp := thePort;
   if tp^.polySave <> nil then begin
      ph := tp^.polySave;
      ClosePoly();           // Any leftover polygon to finish off?
      if ph^^.polySize > sizeof(Polygon) then
         iPolyIntoRgn(ph);   // Yes, put it in too. */
      KillPoly(ph);
   end;
   ShowPen();
   DisposeRgn(dstRgn);          // Kill off the old destination.
   dstRgn := tp^.rgnSave;       // Make dest the accumulated region.
   tp^.rgnSave^ := nil;         // Kill off the accumulator's references to
   tp^.rgnSave := nil;          // the region.
end;


//
// DisposeRgn releases the memory occupied by a given region.  Use this only
// after you're completely through with a temporary region.
//
procedure DisposeRgn(rgn : RgnHandle); // Completely destroy the region.
begin
    DisposeHandle(Handle(rgn));
end;

//
// CopyRgn copies the mathematical structure of srcRgn into dstRgn; that is, it
// makes a duplicate copy of srcRgn.  Once this is done, srcRgn may be altered
// (or even disposed of) without affecting dstRgn.  CopyRgn does _not_ create
// the destination region; space must already have been allocated for it.
//
procedure CopyRgn(srcRgn : RgnHandle; dstRgn : RgnHandle); // Copy one region over the other.
var size : integer;
begin
   // Not if they aren't real
   if (dstRgn <> nil) and (srcRgn <> nil) and (dstRgn^ <> srcRgn^) then begin
      size := srcRgn^^.rgnSize;
      SetHandleSize(Handle(dstRgn), size);
      system.Move(srcRgn^^, dstRgn^^, size);
   end;
end;

procedure SetEmptyRgn(rgn : RgnHandle); // Make the region empty.
begin
   if (rgn <> nil) and (rgn^ <> nil) then begin
      if rgn^^.rgnSize <> sizeof(RegionHdr) then begin
         SetHandleSize(Handle(rgn), sizeof(RegionHdr));
         rgn^^.rgnSize := sizeof(RegionHdr);
      end;
      SetRect(rgn^^.rgnBBox, 0, 0, 0, 0);
   end;
end;

procedure SetRectRgn(rgn : RgnHandle; left, top, right, bottom : smallInt);
begin
   if (rgn <> nil) and (rgn^ <> nil) then begin
      SetEmptyRgn(rgn);
      SetRect(rgn^^.rgnBBox, left, top, right, bottom);
   end;
end;

procedure RectRgn(rgn : RgnHandle; r : Rect);
begin
   if (rgn <> nil) and (rgn^ <> nil) then begin
      SetEmptyRgn(rgn);
      rgn^^.rgnBBox := r;
   end;
end;

procedure SetClip(rgn : RgnHandle);
var
   tp : GrafPtr;
begin
   tp := thePort;
   if tp <> nil then
      CopyRgn(rgn, tp^.clipRgn);
end;

procedure GetClip(rgn : RgnHandle);
var
   tp : GrafPtr;
begin
   tp := thePort;
   if tp <> nil then
      CopyRgn(tp^.clipRgn, rgn)
   else
      SetEmptyRgn(rgn);
end;

procedure ClipRect(r : Rect);
var
   tp : GrafPtr;
begin
   tp := thePort;
   if tp <> nil then
      RectRgn(tp^.clipRgn, r);
end;

procedure OffsetRgn(rgn : RgnHandle; dh, dv : smallInt);
var
   p : ^smallInt;
begin
   if (rgn <> nil) and (rgn^ <> nil) then begin
      OffsetRect(rgn^^.rgnBBox, dh, dv);
      if rgn^^.rgnSize <> sizeof(RegionHdr) then begin
         p := @rgn^^.rgnData;
         while p^ <> EndMarker do begin // Do each scanline of the region.
            p^ += dv;           // Bump Y coordinate.
            inc(p);
            while p^ <> EndMarker do begin // Do each X pair of the scanline.
               p^ += dh;        // Bump X coordinate.
               inc(p);
               p^ += dh;        // Bump X coordinate.
               inc(p);
            end;
            inc(p);             // Bump to next Y coordinate.
         end;
      end;
   end;
end;

//
// PtInRgn checks whether the pixel _below_and_to_the_right_ of the given
// coordinate point is within the specified region, and returns TRUE if so or
// FALSE if not.
//
function PtInRgn(pt : Point; rgn : RgnHandle) : Boolean;
var
   p, q : ^smallInt;
begin
   PtInRgn := False;    // Assume we did not find the point in the region.
   if (rgn <> nil) and (rgn^ <> nil) then begin
      if PtInRect(pt, rgn^^.rgnBBox) then begin // In bounding box?
         if rgn^^.rgnSize <> sizeof(RegionHdr) then begin // Complex region?
            p := @rgn^^.rgnData;
            q := p;
            while p^ <> EndMarker do begin // Find each scanline of the region.
               q := p;          // Record a pointer to its scanline record.
               inc(p);
               while p^ <> EndMarker do // Do each X pair of the scanline.
                  p += 2;       // Bump past X coordinate pairs.
               inc(p);          // Bump to next Y coordinate.
               if (pt.v >= q^) and (pt.v < p^) then begin // Found right line?
                  inc(q);       // (Bump to X coordinate pairs.)
                  while q^ <> EndMarker do begin // Yes, check each X pair in it.
                     if (pt.h >= q[0]) and (pt.h < q[1]) then begin
                        PtInRgn := True;
                        exit;   // Found it.
                     end;
                     q += 2;
                  end;
                  break;        // Had to be here, wasn't -- we failed.
               end;
            end;
         end else begin
            PtInRgn := True;    // Simple region, was in it, passed.
         end;
      end;
   end;
end;

//
// RectInRgn checks whether the given rectangle intersects the specified
// region, and returns TRUE if the intersection encloses at least one bit,
// or FALSE if not.
//
// Note: RectInRgn will sometimes return TRUE when the rectangle merely
//       intersects the region's enclosing rectangle.  If you need to know
//       exactly whether a given rectangle intersects the actual region, you
//       can use RectRgn to set the rectangle to a region, and call SectRgn to
//       see whether the two regions intersect.  If the result of SectRgn is an
//       empty region, then the rectangle doesn't intersect the region.
//
function RectInRgn(r : Rect; rgn : RgnHandle) : Boolean;
var scratch : Rect;
begin
   RectInRgn := False;
   if (rgn = nil) or (rgn^ = nil) then exit;
   RectInRgn := SectRect(r, rgn^^.rgnBBox, scratch);
end;


function EqualRgn(rgn1, rgn2 : RgnHandle) : Boolean;
var
   p, q : ^smallInt;
   size : smallInt;
begin
   if (rgn1 = rgn2) or (rgn1^ = rgn2^) then begin
      EqualRgn := True;
      exit;
   end;
   EqualRgn := False;
   if (rgn1 = nil) and (rgn2 = nil) then begin // Both non-existant, then equal.
      EqualRgn := True;
      exit;
   end;
   if (rgn1 = nil) or (rgn2 = nil) then // Either non-existant, then not equal.
      exit;
   if EmptyRgn(rgn1) and EmptyRgn(rgn2) then begin
      EqualRgn := True;        // Both empty, then equal.
      exit;
   end;
   size := rgn1^^.rgnSize;
   if size <> rgn2^^.rgnSize then
      exit;                    // Different sizes, not equal.
   size := size div sizeof(smallInt);
   p := @rgn1^^.rgnSize;
   q := @rgn2^^.rgnSize;
   while size <> 0 do begin     // Compare structures for exact match.
      if p^ <> q^ then          // Any mismatch means not equal.
         exit;
      inc(p); inc(q);
   end;
   EqualRgn := True;            // Ran the gauntlet, then equal.
end;

function EmptyRgn(rgn : RgnHandle) : Boolean;
begin
   if (rgn <> nil) and (rgn^ <> nil) then
      EmptyRgn := EmptyRect(rgn^^.rgnBBox)
   else                 // Non-existant, then empty.
      EmptyRgn := True;
end;

//
// Fills the specified region with the pnPat stored in the GrafPort.  Clips
// the region to the intersection of the region and the GrafPort's portRect,
// portBits.bounds, clipRgn, and visRgn.  All 8 transfer modes are supported.
//
procedure PaintRgn(rgn : RgnHandle);
var br : Rect;          // Bounding rectangle.
    pbr, cbr, vbr : Rect;
    co, vo, ro, x, y : integer;
    patBit : Boolean;
    tp : GrafPtr;
    rps, vps, cps : ^smallInt; // Region data pointers; Nil if inactive.
    vStartX, vEndX, cStartX, cEndX, rStartX, rEndX : smallInt;
begin
   tp := thePort;
   if (tp = nil) or (rgn = nil) or (rgn^ = nil) then exit;
   if (tp^.pnVis < 0) then exit;
   co := 0; cEndX := 0;
   vo := 0; vEndX := 0;
   pbr := tp^.portBits.bounds;  // We need the logical portBits.bounds for
   if tp^.device = 1 then       // clipping, and moving to global coordinates.
      RotRect(pbr, pbr);
   if SectRect(rgn^^.rgnBBox, tp^.portRect, br) then begin
      if SectRect(br, pbr, br) then begin
         cps := nil;
         if (tp^.clipRgn <> nil) and (tp^.clipRgn^ <> nil) then begin
            cbr := tp^.clipRgn^^.rgnBBox;       // Has Clip region.
            if not SectRect(br, cbr, br) then exit;
            if tp^.clipRgn^^.rgnSize <> sizeof(RegionHdr) then
               cps := @tp^.clipRgn^^.rgnData;
            co := 0;               // Offset to Y marker.
         end;
         vps := nil;
         if (tp^.visRgn <> nil) and (tp^.visRgn^ <> nil) then begin
            vbr := tp^.visRgn^^.rgnBBox;       // Has Visible region.
            if not SectRect(br, vbr, br) then exit;
            if tp^.visRgn^^.rgnSize <> sizeof(RegionHdr) then
               vps := @tp^.visRgn^^.rgnData;
            vo := 0;               // Offset to Y marker.
         end;
         rps := @rgn^^.rgnData;
         ro := 0;                  // Offset to Y marker.
         for y := br.top - pbr.top to br.bottom - pbr.top - 1 do begin
            if (cps <> nil) then begin
               if co = 0 then      // Find the first appropriate scanline.
                  fastForward(cps, y + pbr.top);
               if (cps[co] <> EndMarker) and (y >= (cps[co] - pbr.top)) then
                  cps += co;       // Time to move the Y?
               co := 1;            // For re-traversing X list for each Y.
               cStartX := -32767;  // Coords of active X segment of clip region;
               cEndX   := -32767;  // Minus infinity to start with.
            end;
            if (vps <> nil) then begin
               if vo = 0 then      // Find the first appropriate scanline.
                  fastForward(vps, y + pbr.top);
               if (vps[vo] <> EndMarker) and (y >= vps[vo] - pbr.top) then
                  vps += vo;       // Time to move the Y?
               vo := 1;            // For re-traversing X list for each Y.
               vStartX := -32767;  // Coords of active X segment of vis region;
               vEndX   := -32767;  // Minus infinity to start with.
            end;
            if ro = 0 then         // Find the first appropriate scanline.
               fastForward(rps, y + pbr.top);
            if (rps[ro] <> EndMarker) and (y >= rps[ro] - pbr.top) then begin
               rps += ro;          // Time to move the Y?
            end;
            ro := 1;               // For re-traversing X list for each Y.
            rStartX := -32767;     // Coords of active X segment of region;
            rEndX   := -32767;     // Minus infinity to start with.
            for x := br.left - pbr.left to br.right - pbr.left - 1 do begin
               while (cps <> nil) and
                     (x >= cEndX) and (cps[co] <> EndMarker) do begin
                  cStartX := cps[co] - pbr.left; // Get the X segment of clip rgn
                  cEndX := cps[co+1] - pbr.left; // for where we are right now.
                  co += 2;
               end;
               while (vps <> nil) and
                     (x >= vEndX) and (vps[vo] <> EndMarker) do begin
                  vStartX := vps[vo] - pbr.left; // Get the X segment of vis rgn
                  vEndX := vps[vo+1] - pbr.left; // for where we are right now.
                  vo += 2;
               end;
               while (x >= rEndX) and (rps[ro] <> EndMarker) do begin
                  rStartX := rps[ro] - pbr.left; // Get X segment of target rgn
                  rEndX := rps[ro+1] - pbr.left; // for where we are right now.
                  ro += 2;
               end;
               if ((cps = nil) or ((x >= cStartX) and (x < cEndX))) and // Clip.
                  ((vps = nil) or ((x >= vStartX) and (x < vEndX))) and
                                  ((x >= rStartX) and (x < rEndX)) then begin
                  patBit := (tp^.pnPat.bytes[y and 7] and
                             (1 << (7 - (x and 7)))) <> 0;
                  if tp^.pnMode.i > ord(srcBic) then
                     patBit := not patBit;
                  case tp^.pnMode.e of
                     notSrcCopy,
                     srcCopy:    if patBit then iSetBit(x, y)
                                           else iClearBit(x, y);
                     notSrcOr,
                     srcOr:      if patBit then iSetBit(x, y);
                     notSrcXor,
                     srcXor:     if patBit then iFlipBit(x, y);
                     notSrcBic,
                     srcBic:     if patBit then iClearBit(x, y);
                  else exit;
                  end;
               end;
            end;
            while (cps <> nil) and (cps[co] <> EndMarker) do // Get to end of X
               co += 2;            // segments, if not already there.
            inc(co);               // Point to next Y marker.
            while (vps <> nil) and (vps[vo] <> EndMarker) do // Get to end of X
               vo += 2;            // segments, if not already there.
            inc(vo);               // Point to next Y marker.
            while rps[ro] <> EndMarker do // Get to end of X segments,
               ro += 2;            // if not already there.
            inc(ro);               // Point to next Y marker.
         end;
      end;
   end;
end;

procedure EraseRgn(rgn : RgnHandle);
begin
   if thePort = nil then exit;
   FillRgn(rgn, thePort^.bkPat);
end;

procedure InvertRgn(rgn : RgnHandle);
var penSave : PenState;
begin
   if thePort = nil then exit;
   GetPenState(penSave);
   PenMode(srcXor);
   PenPat(black);
   PaintRgn(rgn);
   SetPenState(penSave);
end;

procedure FillRgn(rgn : RgnHandle; pat : Pattern);
var br : Rect;          // Bounding rectangle, in global coordinates.
    pbr, cbr, vbr : Rect;
    co, vo, ro, x, y : integer;
    patBit : Boolean;
    tp : GrafPtr;
    rps, cps, vps : ^smallInt; // Region data pointers; Nil if inactive.
    cStartX, cEndX, vStartX, vEndX, rStartX, rEndX : smallInt;
begin
   tp := thePort;
   if (tp = nil) or (rgn = nil) or (rgn^ = nil) then exit;
   if (tp^.pnVis < 0) then exit;
   co := 0; cEndX := 0;
   vo := 0; vEndX := 0;
   pbr := tp^.portBits.bounds;  // We need the logical portBits.bounds for
   if tp^.device = 1 then       // clipping, and moving to global coordinates.
      RotRect(pbr, pbr);
   if SectRect(rgn^^.rgnBBox, tp^.portRect, br) then begin
      if SectRect(br, pbr, br) then begin
         cps := nil;
         if (tp^.clipRgn <> nil) and (tp^.clipRgn^ <> nil) then begin
            cbr := tp^.clipRgn^^.rgnBBox;       // Has Clip region.
            if not SectRect(br, cbr, br) then exit;
            if tp^.clipRgn^^.rgnSize <> sizeof(RegionHdr) then
               cps := @tp^.clipRgn^^.rgnData;
            co := 0;               // Offset to Y marker.
         end;
         vps := nil;
         if (tp^.visRgn <> nil) and (tp^.visRgn^ <> nil) then begin
            vbr := tp^.visRgn^^.rgnBBox;       // Has Visible region.
            if not SectRect(br, vbr, br) then exit;
            if tp^.visRgn^^.rgnSize <> sizeof(RegionHdr) then
               vps := @tp^.visRgn^^.rgnData;
            vo := 0;               // Offset to Y marker.
         end;
         rps := @rgn^^.rgnData;
         ro := 0;                  // Offset to Y marker.
         for y := br.top - pbr.top to br.bottom - pbr.top - 1 do begin
            if (cps <> nil) then begin
               if co = 0 then      // Find the first appropriate scanline.
                  fastForward(cps, y + pbr.top);
               if (cps[co] <> EndMarker) and (y >= (cps[co] - pbr.top)) then
                  cps += co;       // Time to move the Y?
               co := 1;            // For re-traversing X list for each Y.
               cStartX := -32767;  // Coords of active X segment of clip region;
               cEndX   := -32767;  // Minus infinity to start with.
            end;
            if (vps <> nil) then begin
               if vo = 0 then      // Find the first appropriate scanline.
                  fastForward(vps, y + pbr.top);
               if (vps[vo] <> EndMarker) and (y >= vps[vo] - pbr.top) then
                  vps += vo;       // Time to move the Y?
               vo := 1;            // For re-traversing X list for each Y.
               vStartX := -32767;  // Coords of active X segment of vis region;
               vEndX   := -32767;  // Minus infinity to start with.
            end;
            if ro = 0 then         // Find the first appropriate scanline.
               fastForward(rps, y + pbr.top);
            if (rps[ro] <> EndMarker) and (y >= rps[ro] - pbr.top) then
               rps += ro;          // Time to move the Y?
            ro := 1;               // For re-traversing X list for each Y.
            rStartX := -32767;     // Coords of active X segment of region;
            rEndX   := -32767;     // Minus infinity to start with.
            for x := br.left - pbr.left to br.right - pbr.left - 1 do begin
               while (cps <> nil) and
                     (x >= cEndX) and (cps[co] <> EndMarker) do begin
                  cStartX := cps[co] - pbr.left; // Get the X segment of clip rgn
                  cEndX := cps[co+1] - pbr.left; // for where we are right now.
                  co += 2;
               end;
               while (vps <> nil) and
                     (x >= vEndX) and (vps[vo] <> EndMarker) do begin
                  vStartX := vps[vo] - pbr.left; // Get the X segment of vis rgn
                  vEndX := vps[vo+1] - pbr.left; // for where we are right now.
                  vo += 2;
               end;
               while (x >= rEndX) and (rps[ro] <> EndMarker) do begin
                  rStartX := rps[ro] - pbr.left; // Get X segment of target rgn
                  rEndX := rps[ro+1] - pbr.left; // for where we are right now.
                  ro += 2;
               end;
               if ((cps = nil) or ((x >= cStartX) and (x < cEndX))) and // Clip.
                  ((vps = nil) or ((x >= vStartX) and (x < vEndX))) and
                                  ((x >= rStartx) and (x < rEndx)) then begin
                  patBit := (pat.bytes[y and 7] and
                             (1 << (7 - (x and 7)))) <> 0;
                  if patBit then
                     iSetBit(x, y)
                  else
                     iClearBit(x, y);
               end;
            end;
            while (cps <> nil) and (cps[co] <> EndMarker) do // Get to end of X
               co += 2;            // segments, if not already there.
            inc(co);               // Point to next Y marker.
            while (vps <> nil) and (vps[vo] <> EndMarker) do // Get to end of X
               vo += 2;            // segments, if not already there.
            inc(vo);               // Point to next Y marker.
            while rps[ro] <> EndMarker do // Get to end of X segments,
               ro += 2;            // if not already there.
            inc(ro);               // Point to next Y marker.
         end;
      end;
   end;
end;

initialization

thePort := nil;         // No GrafPort at the beginning.

end.
