//===========================================================================
//
// Polygons part of the old ISC Clunkdraw routines.  Separate because you
// don't always want or need them.
//
//===========================================================================

//{$DEFINE DEBUG}

unit clunkpoly;

{$mode objfpc}{$H+}

interface

   uses Clunkdraw;

   procedure DumpPoly(poly : PolyHandle; s : string);
   procedure MapPoly(poly : PolyHandle; srcRect, dstRect : Rect);
   function  OpenPoly : PolyHandle;
   procedure ClosePoly;
   procedure KillPoly(poly :  PolyHandle);
   procedure OffsetPoly(poly : PolyHandle; dh, dv : smallInt);
   procedure FramePoly(poly : PolyHandle);
   procedure PaintPoly(poly : PolyHandle);
   procedure ErasePoly(poly : PolyHandle);
   procedure InvertPoly(poly : PolyHandle);
   procedure FillPoly(poly : PolyHandle; pat : Pattern);
   procedure iPolyIntoRgn(poly : PolyHandle);

implementation

   uses Math, Clunkregions;

const
   EndMarker = $7FFF; // End-of-region/polygon-data marker value.

type
   Action = (Paint, Frame, Erase, Invert, Fill, Union); // Drawing verbs.

//
// Dump out a polygon.
//
procedure DumpPoly(poly : PolyHandle; s : string);
var p : ^Point;
   polySize : smallInt;
begin
   if (poly = nil) or (poly^ = nil) then exit;
   DumpRect(poly^^.polyBBox, s);
   polysize := poly^^.polySize - sizeof(PolyHdr);
   p := @poly^^.polyPoints;
   while polysize > 0 do begin
      DumpPt(p^, s);
      inc(p);
      polysize -= sizeof(p^);
   end;
end;


procedure MapPoly(poly : PolyHandle; srcRect, dstRect : Rect);
var p : ^Point;
    polysize : smallInt;
begin
   if (poly = nil) or (poly^ = nil) then exit;
   polysize := poly^^.polySize - sizeof(PolyHdr);
   p := @poly^^.polyPoints;
   while polysize > 0 do begin
      MapPt(p^, srcRect, dstRect);
      inc(p);
      polysize -= sizeof(p^);
   end;
   MapRect(poly^^.polyBBox, srcRect, dstRect);
end;

//
// OpenPoly returns a handle to a new polygon and tells Clunkdraw to start
// saving the polygon definition as specified by calls to line-drawing
// routines.  While a polygon is open, all calls to Line and LineTo affect the
// outline of the polygon.  Only the line endpoints affect the polygon
// definition; the pen mode, pattern, and size do not affect it.  In fact,
// OpenPoly calls HidePen, so no drawing occurs on the screen while the
// polygon is open (unless you call ShowPen just after OpenPoly, or you called
// ShowPen previously without balancing it with a call to HidePen).
//
// A polygon should consist of a sequence of connected lines.  Even though the
// on-screen presentation of a polygon is clipped, the definition of a polygon
// is not; you can define a polygon anywhere on the coordinate plane.
//
// When a polygon is open, the current grafPort's polySave field contains a
// handle to information related to the polygon definition.  If you want to
// temporarily disable the polygon definition, you can save the current value
// of this field, set the field to NULL, and later restore the saved value to
// resume the polygon definition.
//
// WARNING:  Do not call OpenPoly while a region or another polygon is
//           already open.
//
// WARNING:  The polygon may not be referred to or used until the ClosePoly.
//

function OpenPoly : PolyHandle;
var
   tp : GrafPtr;
   ph : PolyHandle;
begin
   tp := thePort;
   if tp = nil then exit(nil);
   if tp^.polySave <> nil then
      KillPoly(tp^.polySave);
   HidePen();
   ph := PolyHandle(NewHandle(sizeof(PolyHdr)));
   tp^.polySave := ph;
   if ph <> nil then
      ph^^.polySize := sizeof(PolyHdr);
   OpenPoly := ph;
end;

//
// ClosePoly tells Clunkdraw to stop saving the definition of the currently
// open polygon and computes the polyBBox rectangle.  You should perform one
// and only one ClosePoly for every OpenPoly.  ClosePoly calls ShowPen,
// balancing the HidePen call made by OpenPoly.
//
// This implementation only supports one collected polygon per grafport.  If
// you need more than one you have to copy it to someplace else in memory
// yourself.  Imaging of polygons has no such restriction, only the collection
// of polygons is limited.
//
procedure ClosePoly;
var bbox : Rect;
    pp : PolyPtr;
    p : ^Point;
    ii : integer;
    tp : GrafPtr;
begin
   tp := thePort;
   if tp^.polySave = nil then
      exit;
   pp := tp^.polySave^;
   tp^.polySave := nil;
   ShowPen();
   SetRect(bbox, 32767, 32767, -32768, -32768); // Calculate polygon's bnd box.
   p := @pp^.polyPoints;
   ii := (pp^.polySize - sizeof(PolyHdr)) div sizeof(Point);
   while ii > 0 do begin        // Number of points.
      dec(ii);
      if p^.h > bbox.right then
         bbox.right := p^.h;
      if p^.h < bbox.left then
         bbox.left := p^.h;
      if p^.v > bbox.bottom then
         bbox.bottom := p^.v;
      if p^.v < bbox.top then
         bbox.top := p^.v;
      inc(p);
   end;
   pp^.polyBBox := bbox;
end;


//
// KillPoly releases the memory occupied by the collected polygon.  Use this
// only when you're completely through with the polygon (for example after
// you've copied it somewhere into your own storage space).
//
procedure KillPoly(poly : PolyHandle);
begin
   DisposeHandle(Handle(poly));
end;


//
// OffsetPoly moves the polygon on the coordinate plane, a distance of dh
// horizontally and dv vertically.  This doesn't affect the screen unless you
// subsequently call a routine to draw the polygon.  If dh and dv are positive,
// the movement is to the right and down; if either is negative, the
// corresponding movement is in the opposite direction.  The polygon retains
// its shape and size.
//
procedure OffsetPoly(poly : PolyHandle; dh, dv : smallInt);
var
   pp : PolyPtr;
   p : ^Point;
   ii : integer;
begin
   if poly = nil then exit;
   pp := poly^;
   if pp = nil then exit;       // Protect against bozo actions.
   p := @pp^.polyPoints;
   ii := (pp^.polySize - sizeof(PolyHdr)) div sizeof(Point);
   OffsetRect(pp^.polyBBox, dh, dv);
   while (ii > 0) do begin      // Number of points.
      dec(ii);
      p^.h += dh;
      p^.v += dv;
      inc(p);
   end;
end;

//
// BuildPolyRgn(inPoly, outRgn)
//
// R1.8 Randy Nichols, April 9, 1992
//
// Converts an arbitrarily-shaped polygon into a region.
// Polygons are divided into three classes:
//   1) Rectangular,
//   2) Monotone-Vertical,
//   3) Complex.
// Each class requires more complex and slower code.
//

function PolygonIsMonotoneVertical(inPoly : PolyHandle; length : integer) : Boolean; forward;
procedure FillMonotoneVerticalPolygon(inPoly : PolyHandle; length : integer; outRgn : RgnPtr); forward;

procedure iBuildPolyRgn(inPoly : PolyHandle; outRgn : RgnPtr);
var ii, jj, kk, Y, start, stop, temp : integer;
    Length : integer;           // Number of line segments in polygon.
    negative : Boolean;         // Flag for negative-slope lines.
    tBmaxy : integer;           // Maximum  y entries, w/ EndMarker.
    tBmaxx : integer;           // Maximum  x entries for each y & EndMarker
    tBx : integer;
    tBy : integer;              // Indexes into tBuf.
    lastBindex, curBindex : ^smallInt; // Pointers for compressing output rgn.
    startLrow, startCrow : ^smallInt; // Pointers for compressing output rgn.
    tLine : ^smallInt;          // Buffer for x values of line points.
    tBuf : ^smallInt;           // Buffer to build x values for each scan line
                                // within polyBBox: x1, x2, ..., xn, $7FFF
                                // x1,..., xm, $7FFF $7FFF
    p, p1, tBp : ^smallInt;     // Scratch pointers.
    match : Boolean;            // True = y rows match.
    CurrentYSign, PreviousYSign : integer;
    StartIndex, StopIndex : integer;
    PStartIndex, PStopIndex : integer;

    VertexPtr : ^Point;
    pp : PolyPtr;
    tmpRect : Rect;             // Used for _buildLines().
begin
   // Determine the number of line segments in the polygon.
   // If the polygon isn't closed, then adjust the length.

   pp := inPoly^;
   Length := (pp^.polySize - sizeof (PolyHdr)) div sizeof (Point);

   VertexPtr := @pp^.polyPoints; // Point to the vertex list.
   if (VertexPtr[Length - 1].h = VertexPtr[0].h) and
      (VertexPtr[Length - 1].v = VertexPtr[0].v) then
       dec(Length);

   // Setup minimum region information.
   // Copy the bounding box from the input poly.
   outRgn^.rgnBBox := pp^.polyBBox;
   tBuf := @outRgn^.rgnData;
   tBuf^ := pp^.polyBBox.top; inc(tBuf);
   tBuf^ := EndMarker; inc(tBuf);
   tBuf^ := EndMarker; inc(tBuf);

   // If the polygon has less than two sides it is not visible.
   if Length <= 2 then begin
{$IFDEF DEBUG}
      WriteLn('Final region...Polygon is not visible.');
{$ENDIF}
      outRgn^.rgnSize := sizeof (RegionHdr) +
                         (tBuf - PsmallInt(@outRgn^.rgnData)) * sizeof(smallInt);
      exit;
   end;

   // If the polygon is a rectangle then use the compressed form.
   if ((Length = 4) and (((VertexPtr[0].h = VertexPtr[1].h) and
                          (VertexPtr[2].h = VertexPtr[3].h) and
                          (VertexPtr[0].v = VertexPtr[3].v) and
                          (VertexPtr[2].v = VertexPtr[1].v)) or
                         ((VertexPtr[0].h = VertexPtr[3].h) and
                          (VertexPtr[1].h = VertexPtr[2].h) and
                          (VertexPtr[0].v = VertexPtr[1].v) and
                          (VertexPtr[3].v = VertexPtr[2].v)))) then begin
{$IFDEF DEBUG}
      WriteLn('Final region...Rectangle.');
{$ENDIF}
      outRgn^.rgnSize := sizeof(RegionHdr);
      exit;
   end;

   // If polygon is Monotone-Vertical then pass it to fast polygon routine.
   if PolygonIsMonotoneVertical(inPoly, Length) then begin
      FillMonotoneVerticalPolygon(inPoly, Length, outRgn);
      exit;
   end;
    // This is a Complex polygon, use the slow code.
    //
    // Initialize tBmaxx, tBmaxy, and working buffer tBuf.
   tBmaxx := Length * 2 + 2;    // Room for y & end tag
   tBmaxy := pp^.polyBBox.bottom - pp^.polyBBox.top + 2; // Room for end tag
   tBuf := @outRgn^.rgnData;

   for ii := 0 to tBmaxy - 2 do begin
      tBuf[0] := pp^.polyBBox.top + ii;
      tBuf[1] := EndMarker;
      tBuf += tBmaxx;           // Bump to next scanline.
   end;
   tBuf[0] := EndMarker;
   tBuf[1] := EndMarker;

    // Process all line segments. Call _BuildLine for each line segment
    // and put the points into tBuf.
   tLine := GetMem(tBmaxy * sizeof(smallInt)); // Max points per line.

   for ii := 0 to Length - 1 do begin
      StartIndex := ii;
      PStartIndex := ii;
      StopIndex := ii;
      StopIndex := (StopIndex + 1) mod Length;
      tmpRect.topLeft := VertexPtr[StartIndex];
      tmpRect.botRight := VertexPtr[StopIndex];

      CurrentYSign := Sign(VertexPtr[StopIndex].v - VertexPtr[StartIndex].v);
{$IFDEF DEBUG}
      Write('Line Seg:', ii+1, ', ');
      Write('Start:', tmpRect.left, ',', tmpRect.top,
            ' Stop:', tmpRect.right, ',', tmpRect.bottom);
{$ENDIF}
      if CurrentYSign = 0 then continue;        // Skip horizontal lines.

      // The start point is used if there is a reversal in Y direction,
      // so determine Y direction of previous line.
      // Skip horizontal lines, they don't count.
      while True do begin
         PStopIndex := PStartIndex;
         PStartIndex := (PStartIndex - 1 + Length) mod Length;
         PreviousYSign := Sign(VertexPtr[PStopIndex].v-VertexPtr[PStartIndex].v);
         if PreviousYSign <> 0 then break;
      end;

      // Set up and call _BuildLine.
      negative := (tmpRect.right < tmpRect.left) <>
                  (tmpRect.bottom > tmpRect.top);
      Pt2Rect(tmpRect.topLeft, &tmpRect.botRight, tmpRect);
      iBuildLine(tmpRect, negative, tLine);
      negative := not negative;         // This line is to fix zig-zag problems.

{$IFDEF DEBUG}
      if negative then
         WriteLn('    (negative slope)')
      else
         WriteLn;
      Write('BUILDLINE:');
      kk := RectHeight(tmpRect);
      for jj := 0 to kk do
         Write(tLine[jj], ' ');
      WriteLn;
{$ENDIF}
      // Put the X values from tLine into tBuf for each scan line.
      // Must add both start & stop points for each scan line because
      // we don't know yet which one should be used until all lines
      // are scanned.

      stop := 0;
      kk := tmpRect.bottom - tmpRect.top;
      tBuf := @outRgn^.rgnData;

      for jj := 0 to kk do begin
         // Skip the start point if the Y direction hasn't changed.
         if ((CurrentYSign = PreviousYSign) and
             (((jj = 0) and (CurrentYSign > 0)) or
              ((jj = kk) and (CurrentYSign < 0)))) then begin
            stop := tLine[jj];
            continue;
         end;

         // Position to the correct 'y' line in tBuf.
         tBy := tmpRect.top - pp^.polyBBox.top + jj;

         // Find the end of line in tBuf.
         tBp := @tBuf[tBy * tBmaxx + 1];
         while tBp^ <> EndMarker do
            inc(tBp);
         if tmpRect.left = tmpRect.right then begin // Vertical line.
            tBp^ := tmpRect.left; inc(tBp); // Start scanline.
            tBp^ := tmpRect.left; inc(tBp); // Stop scanline.
         end else begin         // Not vertical.
            start := stop;
            stop := tLine[jj];
            if (stop - start) < 1 then
               start := stop;

            if negative then begin      // Handle reversed-slope lines.
               tBp^ := tmpRect.right - stop; inc(tBp);  // Start scan.
               tBp^ := tmpRect.right - start; inc(tBp); // Stop scan.
            end else begin      // Positive sloped line.
               tBp^ := tmpRect.left + start; inc(tBp); // Start scan.
               tBp^ := tmpRect.left + stop; inc(tBp);  // Stop scan.
            end;
         end;
         tBp^ := EndMarker;     // Add new end tag.
      end;                      // Done with tLine ^. tBuf.
   end;                         // Done with all line segaments.
   FreeMem(tLine);

   // Clean up tBuf - Sort X's in ascending order for each Y.
   for Y := 0 to tBmaxy - 1 do begin
      tBp := @tBuf[Y * tBmaxx + 1]; // Start of X entries.
      while True do begin       // Repeat line until no more swaps.
         jj := 0;
         p := tBp;
         while (p[0] <> EndMarker) and (p[1] <> EndMarker) do begin
            if p[1] < p[0] then begin // Then swap.
               jj := 1;
               temp := p[0]; p[0] := p[1]; p[1] := temp; // Swap.
            end;
            inc(p);
         end;                   // for entire line
         if jj = 0 then break;
      end;                      // repeat line until no more swaps
   end;                         // for all scan lines

   // Clean up tBuf - ...
   for Y := 0 to tBmaxy - 1 do begin
      tBy := Y * tBmaxx;

      // Both the start and stop points have been added for each line.
      // Keep the 1st point for the start, and the 2nd point for the stop.
      p := @tBuf[tBy + 1];      // Start of X entries.
      tBp := p;
      kk := 0;                  // 0 = start point.
      jj := 0;
      while (p[0] <> EndMarker) and (p[1] <> EndMarker) do begin
         tBp^ := p[kk];
         kk := kk xor 1;        // Toggle start/end flag.
         inc(tBp);
         p += 2;
      end;
      tBp^ := EndMarker;

      // Remove duplicate X pairs for each y line.
      while True do begin       // Do each line 'till no more dup. pairs.
         p := @tBuf[tBy + 1];   // Start of X entries.
         jj := 0;
         while (p[0] <> EndMarker) and (p[1] <> EndMarker) do begin
            if p[0] = p[1] then begin // Found duplicate pair.
               jj := 1;
               while p[1] <> EndMarker do begin // Delete pair.
                  p[0] := p[2];
                  inc(p);
               end;
            end;
            inc(p);
         end;                   // for entire line
         if jj = 0 then         // No more matches on this line.
            break;
      end;
   end;                         // for all scan lines

   // Fix vertical boundary conditions - Move startpoints and endpoints
   // vertically INSIDE the defining line segments, not below and to
   // the right like lines do.
   //
   // Start with the top line, the current line, p, is compared to
   // the one below it, p1.
   for Y := 0 to tBmaxy - 2 do begin // Do all scan lines except last.
      tBy := Y * tBmaxx;

      // Do all point pairs on current line (p).
      p := @tBuf[tBy + 1];
      while (p[0] <> EndMarker) and (p[1] <> EndMarker) do begin
         // Try all points on next line (p1) until p1[0]<*p<p1[1].
         p1 := @tBuf[tBy + tBmaxx + 1];
         kk := 1;
         while p1^ <> EndMarker do begin
            kk := kk xor 1;     // Toggle in/out flag for p1. (0 = start)
            if (p[0] < p1^) and (p1^ < p[1]) then begin
               // This is a kludge fix for self intersecting polys.
               // Fixes the transition from one point pair on p to
               // 2 pairs on p1.
               if (p1[1] <= p[1]) and (kk = 1) then
                  p1 += 2;
               p[kk] := p1^;
            end;
            inc(p1);
         end;   // All points on p1.
         p += 2;
      end;      // All point pairs on p.
   end;         // for all scan lines
   tBuf[(tBmaxy - 2) * tBmaxx] := EndMarker; // Add final end tag.

   // Build the final output region: Delete duplicate x pairs for each y.
   // Delete duplicate y rows. Compact region data.
   // 'p' points to the source region, and 'tBuf' points to the
   // destination region.

   curBindex := @tBuf[0];
   lastBindex := curBindex;
   startCrow := lastBindex;
   startLrow := startCrow;
   // Pointers into final compressed list.
   match := False;
   for Y := 0 to tBmaxy - 2 do begin
      // Do all lines.  Match = 0 so 1st line won't match.
      tBx := 0;
      while tBx < tBmaxx do begin // Entire line
         tBp := @tBuf[(Y * tBmaxx) + tBx];
         while True do begin
            if (tBx <> 0) and (tBp^ <> EndMarker) and (tBp[0] = tBp[1]) then begin
               tBp += 2;        // Skip past duplicate x pair.
               tBx += 2;        // Skip past duplicate x pair.
            end else
               break;
         end;

         curBindex^ := tBp^;    // Add value; y, x, or tag.
         if (tBx <> 0) and (lastBindex^ <> tBp^) then
            match := False;     // Last compressed row <> new uncompressed row.
         if (curBindex^ = EndMarker) then begin
            inc(curBindex);
            break;
         end;
         inc(curBindex);
         inc(lastBindex);
         inc(tBx);
      end;                      // for entire line

      if match then             // If the last 2 rows match, then reset pointers.
         curBindex := startCrow
      else begin
         startLrow := startCrow; // No match, update pointers.
         startCrow := curBindex;
      end;
      lastBindex := startLrow;
      match := True;
   end;
   curBindex^ := EndMarker;     // Add the final end tag.
   inc(curBindex);
    // Calculate region size.
   outRgn^.rgnSize := PByte(curBindex) - PByte(@tBuf[0])
                    + sizeof(RegionHdr);
{$IFDEF DEBUG}
   WriteLn('Final region...Complex Polygon');
   Write('outRgn^.rgnBBox = (', outRgn^.rgnBBox.left, ',',
         outRgn^.rgnBBox.top, '), (', outRgn^.rgnBBox.right, ',',
         outRgn^.rgnBBox.bottom, '), ');
   WriteLn('outRgn^.rgnSize = ', outRgn^.rgnSize);
   tBuf := @outRgn^.rgnData;
   while True do begin
      Write(tBuf^, ', ');
      if tBuf[0] = EndMarker then begin
         WriteLn;
         if tBuf[1] = EndMarker then begin
            WriteLn(tBuf[1]);
            break;
         end;
      end;
      inc(tBuf);
   end;
{$ENDIF}
end;

//
// PolygonIsMonotoneVertical()
//
// Returns True if polygon described by passed-in vertex list is monotone with
// respect to a vertical line, False otherwise.  Doesn't matter if the polygon
// is simple (non-self-intersecting) or not.
// (From a series of articles by Michael Abrash, Dr. Dobb's Journal,
//  February, March, June, and December 1991.)
//
function PolygonIsMonotoneVertical(inPoly : PolyHandle; Length : integer) : Boolean;
var
   ii, CurrentYSign, PreviousYSign, NumYReversals : integer;
   VertexPtr : ^Point;
   pp : PolyPtr;
begin
   NumYReversals := 0;
   pp := inPoly^;
   VertexPtr := @pp^.polyPoints;

   // Three or fewer points can't make a non-vertical-monotone polygon.
   if Length <= 3 then exit(True);

   // Scan to the first non-horizontal edge.
   PreviousYSign := Sign(VertexPtr[Length - 1].v - VertexPtr[0].v);
   ii := 0;
   while (PreviousYSign = 0) and (ii < (Length - 1)) do begin
      PreviousYSign := Sign(VertexPtr[ii].v - VertexPtr[ii + 1].v);
      inc(ii);
   end;

   if ii = (Length - 1) then exit(True); // Polygon is a flat line.

   // Now count Y reversals.  Might miss one reversal, at the last vertex, but
   // because reversal counts must be even, being off by one isn't a problem.
   repeat begin
      CurrentYSign := Sign(VertexPtr[ii].v - VertexPtr[ii+1].v);
      if CurrentYSign <> 0 then begin
         if CurrentYSign <> PreviousYSign then begin
            // Switched Y direction; not vertical-monotone if
            // reversed Y direction as many as three times.
            inc(NumYReversals);
            if NumYReversals > 2 then exit(False); // NOT MONOTONE.
            PreviousYSign := CurrentYSign;
         end;
      end;
      inc(ii);
   end until ((ii - 1) >= (Length - 1));
   PolygonIsMonotoneVertical := True;   // It's a vertical-monotone polygon.
end;


//
// FillMonotoneVerticalPolygon()
//
// Creates a scan list for vertical-monotone polygons.
// Vertical-monotone means every horizontal
// line drawn through the polygon at any point would cross exactly
// two active edges (neither horizontal lines nor zero-length edges
// count as active edges;  both are acceptable anywhere in the polygon).
// Right & left edges may cross (self-intersecting).
// (From a series of articles by Michael Abrash, Dr. Dobb's Journal,
//  February, March, June, and December 1991.)
//
procedure FillMonotoneVerticalPolygon(inPoly : PolyHandle; Length : integer; outRgn : RgnPtr);
var negative : Boolean;         // Flag for negative-slope lines.
    tLine : ^smallInt;          // Buffer for x values of line points.
    tBuf : ^smallInt;           // Buffer to build x values for each line.
    tBufNew, tBufNewP : ^smallInt;
    temp : smallInt;
    ii, Y, MinIndex, MaxIndex, MinPoint_Y, MaxPoint_Y : integer;
    CurrentIndex, PreviousIndex : integer;
    VertexPtr : ^Point;
    pp : PolyPtr;
    tmpRect : Rect;             // Used for _buildLines().
begin
   pp := inPoly^;
   VertexPtr := @pp^.polyPoints; // Point to the vertex list.

   // Get memory in which to store the line list we generate.
   tLine := GetMem((pp^.polyBBox.bottom - pp^.polyBBox.top + 2)
                   * sizeof (smallInt)); // Max 1 X per scan line).

    // Scan the list to find the top and bottom of the polygon.
   MinIndex := 0;
   MaxIndex := 0;
   MinPoint_Y := VertexPtr[0].v;
   MaxPoint_Y := MinPoint_Y;
   for ii := 1 to Length - 1 do begin
      if (VertexPtr[ii].v < MinPoint_Y) then begin
         MinIndex := ii;
         MinPoint_Y := VertexPtr[ii].v // New top.
      end else if (VertexPtr[ii].v > MaxPoint_Y) then begin
         MaxIndex := ii;
         MaxPoint_Y := VertexPtr[ii].v; // New bottom.
      end;
   end;

    // Scan the first edge and store the boundary points in the list.
    // Scan convert each line in the first edge from top to bottom.

   PreviousIndex := MinIndex;
   CurrentIndex := MinIndex;
   tBuf := @outRgn^.rgnData;
   repeat begin
      CurrentIndex := (CurrentIndex - 1 + Length) mod Length;

      tmpRect.topLeft := VertexPtr[PreviousIndex];
      tmpRect.botRight := VertexPtr[CurrentIndex];

      negative := (tmpRect.right < tmpRect.left);
      Pt2Rect(tmpRect.topLeft, tmpRect.botRight, tmpRect);

      iBuildLine(tmpRect, negative, tLine);

      Y := 0;
      while Y < VertexPtr[CurrentIndex].v - VertexPtr[PreviousIndex].v do begin
         tBuf^ := Y + VertexPtr[PreviousIndex].v; // Add Y.
         inc(tBuf);
         if negative then
            tBuf^ := tmpRect.right - tLine[Y] // Add XStart.
         else
            tBuf^ := tmpRect.left + tLine[Y]; // Add XStart.
         tBuf += 3;             // Skip past XEnd and EndMarker.
         inc(Y);
      end;

      PreviousIndex := CurrentIndex;
   end until (CurrentIndex = MaxIndex);

   // Scan the second edge from top to bottom and store the
   // boundary points in the list.

   PreviousIndex := MinIndex;
   CurrentIndex := MinIndex;
   tBuf := @outRgn^.rgnData;
   repeat begin
      CurrentIndex := (CurrentIndex + 1) mod length;

      tmpRect.topLeft := VertexPtr[PreviousIndex];
      tmpRect.botRight := VertexPtr[CurrentIndex];

      negative := (tmpRect.right < tmpRect.left);
      Pt2Rect(tmpRect.topLeft, tmpRect.botRight, tmpRect);

      iBuildLine(tmpRect, negative, tLine);


      for Y := 0 to VertexPtr[CurrentIndex].v - VertexPtr[PreviousIndex].v - 1 do begin
         tBuf += 2;             // Skip past Y and XStart.
         if negative then
            tBuf^ := tmpRect.right - tLine[Y] // Stop scan.
         else
            tBuf^ := tmpRect.left + tLine[Y]; // Stop scan.
         inc(tBuf);
         tBuf^ := EndMarker;    // Add line EndMarker.
         inc(tBuf);
      end;
      PreviousIndex := CurrentIndex;
   end until (CurrentIndex = MaxIndex);
   tBuf^ := EndMarker;          // Add final EndMarker to terminate region.
   inc(tBuf);

   // Compress region. Sort and reduce X entries.

   tBuf := @outRgn^.rgnData;
   tBufNew := tBuf;
   for Y := 0 to MaxPoint_Y - MinPoint_Y - 1 do begin
      tBufNew^ := tBuf^;        // Copy Y entry.
      inc(tBufNew); inc(tBuf);
      if tBuf[0] = tBuf[1] then begin // If XStart = XEnd replace with EndMarker.
         tBufNew^ := EndMarker;
         inc(tBufNew);
         tBuf += 3;
      end else begin
         if tBuf[0] > tBuf[1] then begin // Sort so XStart <= XEnd.
            temp := tBuf[0]; tBuf[0] := tBuf[1]; tBuf[1] := temp;
         end;
         tBufNew^ := tBuf^; inc(tBufNew); inc(tBuf); // Copy XStart.
         tBufNew^ := tBuf^; inc(tBufNew); inc(tBuf); // Copy XEnd.
         tBufNew^ := EndMarker; inc(tBufNew); inc(tBuf); // Line terminator.
      end;
   end;
   tBufNew^ := EndMarker;       //  Region terminator.
   inc(tBufNew);

   // Compress region. Delete duplicate rows.
   tBuf := @outRgn^.rgnData;
   tBufNew := tBuf;
   tBufNewP := tBuf;

   if tBuf[1] = EndMarker then begin // Advance pointers to 2nd row.
      tBuf += 2;
      inc(outRgn^.rgnBBox.top); // First row is empty, shrink BBox.
   end else begin
      tBuf += 4;
      tBufNew += 4;
   end;

   for Y := 1 to MaxPoint_Y - MinPoint_Y - 1 do begin
      if (tBuf[1] = tBufNewP[1]) and (tBuf[1] = EndMarker) then
         tBuf += 2
      else if (tBuf[1] = tBufNewP[1]) and (tBuf[2] = tBufNewP[2]) then
         tBuf += 4
      else begin
         tBufNewP := tBufNew;
         tBufNew^ := tBuf^;     // Copy Y entry to compressed location.
         inc(tBufNew); inc(tBuf);
         if tBuf^ <> EndMarker then begin
            tBufNew^ := tBuf^; // Copy XStart.
            inc(tBufNew); inc(tBuf);
            tBufNew^ := tBuf^; // Copy XEnd.
            inc(tBufNew); inc(tBuf);
         end;
         tBufNew^ := EndMarker; // Line terminator.
         inc(tBufNew); inc(tBuf);
      end;
   end;
   tBufNew^ := EndMarker;       // Mark end of region.
   inc(tBufNew);

   // Calculate region size.
   outRgn^.rgnSize := PByte(tBufNew) - PByte(@outRgn^.rgnData)
                    + sizeof(RegionHdr);
end;

procedure doPoly(poly : PolyHandle; verb : Action; pat : Pattern);
var tBmaxx : smallInt;          // Maximum number of x entries for each y,
                                // including room for EndMarker.
    tBmaxy : smallInt;          // Maximum number of y entries, w/ EndMarker.
    size1 : PtrUInt;
    pp : PolyPtr;
    polyRgn : RgnPtr;
    resultRgn : RgnHandle;
    tp :GrafPtr;
begin
   tp := thePort;
   if (tp = nil) or (poly = nil) or (poly^ = nil) then exit;
   pp := poly^;                 // Create space for output buffer.
   tBmaxx := (2 * (pp^.polySize - sizeof(PolyHdr)))
             div sizeof(Point) + 2;     // Room for y & end tag.
   tBmaxy := pp^.polyBBox.bottom
           - pp^.polyBBox.top + 2; // Room for end tag.
   size1 := MulU(tBmaxy, tBmaxx);
   size1 *= sizeof(smallInt);
   size1 += sizeof(PolyHdr);
   polyRgn := GetMem(size1); // Space for output region.
   iBuildPolyRgn(poly, polyRgn); // Build region for filling polygons
   resultRgn := @polyRgn;
   if verb = Union then begin
      UnionRgn(resultRgn, tp^.rgnSave, tp^.rgnSave);
   end else begin
      case verb of
        Paint:  PaintRgn(resultRgn);
        Frame:  FrameRgn(resultRgn);
        Erase:  EraseRgn(resultRgn);
        Invert: InvertRgn(resultRgn);
        Fill:   FillRgn(resultRgn, pat);
      end;
   end;
   FreeMem(polyRgn);
end;

//
// iPolyIntoRgn accumulates the given polygon into the currently accumulating
// region definition.
//
procedure iPolyIntoRgn(poly : PolyHandle);
begin
   if (poly = nil) or (poly^ = nil) then exit; // Protect from bozo actions.
   doPoly(poly, Union, white);
end;

//
// FramePoly plays back the line-drawing routine calls that define the given
// polygon, using the current grafPort's pen pattern, mode, and size.  The pen
// will hang below and to the right of each point on the boundary of the
// polygon; thus, the polygon drawn will extend beyond the right and bottom
// edges of *(*poly)^.polyBBox by the pen width and height, respectively.  All
// other graphic operations occur strictly within the boundary of the polygon,
// as for other shapes.
//
// If a polygon is open and being formed, FramePoly affects the outline of the
// polygon just as if the line-drawing routines themselves had been called.  If
// a region is open and being formed, the outside outline of the polygon being
// framed is mathematically added to the region's boundary.
//
procedure FramePoly(poly : PolyHandle);
var pp : PolyPtr;
    p : ^Point;
    ii : integer;
begin
   if poly = nil then exit;
   pp := poly^;
   if pp = nil then exit;       // Protect against bozo actions.
   if (thePort <> nil) and (thePort^.rgnSave <> nil) then
      iPolyIntoRgn(poly);
   p := @pp^.polyPoints;
   ii := (pp^.polySize - sizeof(PolyHdr)) div sizeof(Point); // # pts.
   if ii <> 0 then begin        // Anything in the polygon to draw?
      MoveTo(p^.h, p^.v);       // First point is the MoveTo,
      inc(p);
      dec(ii);
      while ii > 0 do begin     // the rest are the line segments.
         dec(ii);
         LineTo(p^.h, p^.v);
         inc(p);
      end;
   end;
end;

//
// PaintPoly paints the specified polygon with the current grafPort's pen
// pattern and mode.  The polygon is filled with the pnPat, according to the
// pattern transfer mode specified by pnMode.  The pen location is not changed
// by this procedure.
//
procedure PaintPoly(poly : PolyHandle);
begin
   doPoly(poly, Paint, white);
end;

//
// ErasePoly paints the specified polygon with the current grafPort's
// background pattern bkPat (in srcCopy mode).  The grafPort's pnPat and
// pnMode are ignored; the pen location is not changed.
//
procedure ErasePoly(poly : PolyHandle);
begin
   doPoly(poly, Erase, white);
end;


//
// InvertPoly inverts the pixels enclosed by the specified polygon:  Every
// white pixel becomes black and every black pixel becomes white.  The
// grafPort's pnPat, pnMode, and bkPat are all ignored; the pen location is
// not changed.
//
procedure InvertPoly(poly : PolyHandle);
begin
   doPoly(poly, Invert, white);
end;


//
// FillPoly fills the specified polygon with the given pattern (in srcCopy
// mode).  The grafPort's pnPat, pnMode and bkPat are all ignored; the pen
// location is not changed.
//
procedure FillPoly(poly : PolyHandle; pat : Pattern);
begin
   doPoly(poly, Fill, pat);
end;

initialization

end.
