//
// Arc, wedge, oval, and rounded rectangle part of the old ISC Clunkdraw
// routines.  Separate because you don't always want or need them.
//
// These work by generating regions that describe the desired shape and then
// calling the appropriate region drawing routine.  Not the quickest way to
// do it by any means, but fairly easy to implement.
//
// For arcs/wedges of less than 180 degrees of coverage this first builds an
// oval shaped region, then builds wedge-shaped (half a trapezoid) regions (2
// of 'em), and then trims the oval with these two regions.  For arcs/wedges
// of more than 180 degrees we build two oval regions, build the two trimming
// regions, trim the ovals with one each, then merge the results.
//

unit clunkrounds;

{$mode objfpc}{$H+}

interface
   uses Clunkdraw;

   procedure PtToAngle(r : Rect; pt : Point; var angle : smallInt);
   function  SlopeFromAngle(angle : smallInt) : Fixed;
   function  AngleFromSlope(slope : Fixed) : smallInt;

   procedure FrameOval(r : Rect);
   procedure PaintOval(r : Rect);
   procedure EraseOval(r : Rect);
   procedure InvertOval(r : Rect);
   procedure FillOval(r : Rect; pat : Pattern);

   procedure FrameArc(r : Rect; startAngle, arcAngle : smallInt);
   procedure FrameWedge(r : Rect; startAngle, arcAngle : smallInt);
   procedure PaintArc(r : Rect; startAngle, arcAngle : smallInt);
   procedure EraseArc(r : Rect; startAngle, arcAngle : smallInt);
   procedure InvertArc(r : Rect; startAngle, arcAngle : smallInt);
   procedure FillArc(r : Rect; startAngle, arcAngle : smallInt; pat : Pattern);

   procedure FrameRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
   procedure PaintRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
   procedure EraseRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
   procedure InvertRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
   procedure FillRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt; pat : pattern);

implementation

   uses Clunkregions;

const
   EndMarker = $7FFF; // End-of-region/polygon-data marker value.

const
{$IFDEF ENDIAN_LITTLE}  // Handle byte sex.
   MSW = 1;
   LSW = 0;
   MSWMSB = 3;
   MSWLSB = 2;
{$ENDIF}
{$IFDEF ENDIAN_BIG}
   MSW = 0;
   LSW = 1;
   MSWMSB = 0;
   MSWLSB = 1;
{$ENDIF}

type
   Action = (Paint, Frame, Erase, Invert, Fill); // Drawing verbs.

   fixtrig = packed record
      case integer of
        0 : (fixval : Fixed);   // Full fixed value.
        1 : (words : array [0..1] of smallInt); // Integer and fraction words.
        2 : (bytes : array [0..4] of Byte);     // Split even further.
   end;

procedure tlookup(var retval : fixtrig; angle : smallInt);
//
// Tables used to look up angles.
//
const

   bytetab : array[0..27] of Byte = (
                $01,$02,$02,$02,$02,$02,$02,$02,$02,$03,$03,$03,$03,$04,
                $04,$04,$05,$05,$06,$07,$08,$09,$0B,$0E,$13,$1C,$39,$FF);

   angletab : array[0..90] of Word = (
                $0000,$0478,$08F1,$0D6B,$11E7,$1666,$1AE8,$1F6F,
                $23FA,$288C,$2D24,$31C3,$366A,$3B1A,$3FD4,$4498,
                $4968,$4E44,$532E,$5826,$5D2D,$6245,$676E,$6CAA,
                $71FB,$7760,$7CDC,$8270,$881E,$8DE7,$93CD,$99D2,
                $9FF7,$A640,$ACAD,$B341,$B9FF,$C0E9,$C802,$CF4E,
                $D6CF,$DE8A,$E681,$EEB9,$F737,$0000,$0919,$1287,
                $1C51,$267F,$3117,$3C22,$47AA,$53B9,$605B,$6D9B,
                $7B89,$8A35,$99AF,$AA0E,$BB68,$CDD6,$E177,$F66E,
                $0CE1,$24FE,$3EFC,$5B19,$799F,$9AE7,$BF5B,$E77A,
                $13E3,$4556,$7CC7,$BB68,$02C2,$54DB,$B462,$2501,
                $ABD9,$5051,$1D88,$24F3,$83AD,$6E17,$4CF5,$14BD,
                $A2D7,$4A30,$FFFF);
begin
   if angle > 90 then begin     // Fold into 0..90 degrees. */
      angle := -(angle - 180);
      retval.words[MSW] := 0;
   end;
   if angle >= 45 then begin    // Fold into first octant for table lookup.
      inc(retval.words[MSW]);   // From 45 to 63 tan(angle) is >= 1 and < 2.
      if angle >= 64 then begin // At 64 degrees, we need real digits.
         retval.bytes[MSWLSB] := bytetab[angle - 63];
         if bytetab[angle - 63] = $FF then
            retval.bytes[MSWMSB] := retval.bytes[MSWMSB] or $7F;
      end;
   end;
   retval.words[LSW] := angletab[angle];
   if retval.fixval < 0 then begin
      retval.bytes[MSWMSB] := retval.bytes[MSWMSB] and not $80;
      retval.fixval := -retval.fixval;
   end;
end;

// PtToAngle calculates an integer angle between a line from the center of the
// rectangle to the given point and a line from the center of the rectangle
// pointing straight up (12 o'clock high).  The angle is in degrees from 0 to
// 359, measured clockwise from 12 o'clock, with 90 degrees at 3 o'clock, 180
// at 6 o'clock, and 270 at 9 o'clock.  Other angles are measured relative to
// the rectangle:  If the line to the given point goes through the top right
// corner of the rectangle, the angle returned is 45 degrees, even if the
// rectangle isn't square; if it goes through the bottom left corner, the
// angle is 135 degrees, and so on.
//
// The angle returned might be used as input to one of the procedures that
// manipulates arcs and wedges.
//
procedure PtToAngle(r : Rect; pt : Point; var angle : smallInt);
var deltav, deltah : smallInt;
    slope, squashby : Fixed;
begin
   deltav := pt.v - ((r.top + r.bottom) div 2);
   deltah := pt.h - ((r.left + r.right) div 2);
   if deltah = 0 then begin
      if deltav <= 0 then
         angle := 0
      else
         angle := 180;
   end else begin
      slope := FixRatio(deltah, deltav);
      squashby := FixRatio(r.bottom - r.top, r.right - r.left);
      slope := FixMul(slope, squashby);
      angle := AngleFromSlope(slope);
      if deltah < 0 then begin
         angle += 180;
         if angle = 360 then
            angle := 0;
      end;
   end;
end;

// Given an angle, SlopeFromAngle returns the slope dh/dv of the line forming
// that angle with the y-axis (dh/dv is the horizontal change divided by the
// vertical change between any two points on the line).  The angle is treated
// modulo 180, and is in degrees measured from 12 o'clock; positive degrees
// are measured clockwise, negative degrees are measured counterclockwise (for
// example, 90 degrees is a 3 o'clock and -90 degrees is at 9 o'clock).
// Positive y is down; positive x is to the right.
//
// SlopeFromAngle(45) == 0xFFFF0000 (read as -1 in Fixed type).
// AngleFromSlope(0xFFFF0000) == 45.
//
// This is more-or-less a fixed-point Tangent routine.
//
function SlopeFromAngle(angle : smallInt) : Fixed;
var
   retval : fixtrig;
begin
   angle := angle mod 180;
   if angle < 0 then
      angle += 180;
   retval.words[MSW] := 0;
   retval.bytes[MSWMSB] := $80; // Result is assumed negative at first.
   tlookup(retval, angle);
   SlopeFromAngle := retval.fixval;
end;


// Given the slope dh/dv of a line, AngleFromSlope returns the angle formed
// by that line and the y-axis.  The angle returned is between 1 and 180
// (inclusive), in degrees measured clockwise from 12 o'clock.
//
// AngleFromSlope is meant for use when speed is much more important than
// accuracy -- its integer result is guaranteed to be within one degree of
// the correct answer, but not necessarily within half a degree.  However,
// the equation
//
//   AngleFromSlope(SlopeFromAngle(x)) == x
//
// is true for all x except 0 (although its reverse is not).
//
// Note:  SlopeFromAngle(0) is 0, and AngleFromSlope(0) is 180.
//
function AngleFromSlope(slope : Fixed) : smallInt;
var
   workval : fixtrig;
   angle : smallInt;
   negative : Boolean;
begin
   negative := slope < 0;
   if negative then
      slope := -slope;
   slope -= 500;        // Subtract 0.007629
   angle := -1;
   while True do begin
      inc(angle);
      workval.words[MSW] := 0;
      tlookup(workval, angle);
      if slope <= workval.fixval then
         break;
   end;
   if negative then
      AngleFromSlope := angle
   else
      AngleFromSlope := 180 - angle;
end;

//
// iBuildArc builds a 90-degree section of an arc along an oval contained
// within rect.  It is used by the xxOval procedures to construct regions that
// are ovular in shape.
//
// Ellipse drawer from November 1988 Computer Language magazine, pp 39.
//
procedure iBuildArc(r : Rect;           // What arc segment to build.
                    p : PsmallInt);     // Where to build it.
var
   alpha, beta, u, k1, k3, a, b, d : longInt;
   x, y, xradius, yradius, movevia : smallInt;
begin
   x := 0;                      // Start from origin.
   y := 0;
   xradius := (r.right - r.left) div 2;
   yradius := (r.bottom - r.top) div 2;
   alpha := MulS(xradius, xradius);
   beta := MulS(yradius, yradius);
   u := alpha * yradius;
   k1 := -2 * beta;             // Assume first move is in X direction.
   k3 := k1 - 2 * alpha;
   b := beta;
   a := 2 * u - b;
   d := b - u + alpha div 4;
   movevia := 0;                // X is moving.
   if a < 0 then begin          // First move is in Y direction.
      k1 := 2 * alpha;
      k3 := k1 + 2 * beta;
      b := 2 * u - alpha;
      a := -b;
      d := b - beta div 4;
      inc(movevia);             // No, make that Y is moving.
   end;
   while True do begin          // Walk around one quadrant of the ellipse.
      p[y] := x;
      if d < 0 then begin       // Move horizontally or vertically.
         if movevia <> 0 then
            inc(y)
         else
            inc(x);
         b -= k1;
         a += k1;
         d += b;
      end else begin            // Move diagonally.
         inc(x);
         inc(y);
         b -= k1;
         a += k3;
         d -= a;
      end;
      if (x > xradius) or (y > yradius) then begin // Time to quit?
         if p^ = 0 then
            p^ := 1;            // Kludge.
         exit;
      end;
      if a < 0 then begin       // Time to change H vs V move type?
         movevia := movevia xor 1;
         k1 -= k3;
         k3 := -k3;
         b += a - k1 div 2;
         d := b - a div 2 - d + k3 div 8;
         a := k1 - k3 div 2 - a;
      end;
   end;
end;

//
// BuildORgn builds a fake region data structure from a list of data points
// (provided by BuildArc) that describe an oval. It's used by the xxOval
// drawing routines.
//
procedure iBuildORgn(r : Rect;          // What to make an oval region from.
                     arc,               // Arc control point list.
                     p : PsmallInt);    // Where to make it.
var
   ii, jj, centerx, fudgex, fudgey, nullPt : smallInt;
   lastpt, save1, save2 : ^smallInt;
begin
   save1 := p;
   inc(p);                      // Skip over the size field.
   PRect(p)^ := r;              // Dummy region's bounding box.
   p += 4;
   save2 := p;
   ii := r.bottom - r.top;
   jj := ii div 2;
   if odd(ii) then              // Do we have to heighten the oval
      fudgey := 1               // by 1 pixel 'cause it's an odd height?
   else
      fudgey := 0;
   if odd(r.right - r.left) then // Ditto for widening it.
      fudgex := 1
   else
      fudgex := 0;

   nullpt := EndMarker;         // So we get the first region segment.
   lastpt := @nullpt;
   centerx := r.left + ((r.right - r.left) div 2);
   for ii := 0 to jj - 1 do begin // Top half of oval.
      if arc^ <> lastpt^ then begin // Don't output unnecessary region segments.
         lastpt := arc;
         p^ := r.top + ii;              inc(p);
         p^ := centerx - arc^;          inc(p);
         p^ := centerx + arc^ + fudgex; inc(p); inc(arc);
         p^ := EndMarker;               inc(p);
      end else begin
         inc(arc);
      end;
   end;

   for ii := 0 to jj - 1 do begin // Bottom half of oval.
      dec(arc);
      if arc^ <> lastpt^ then begin
         lastpt := arc;
         p^ := r.top + ii + jj + fudgey; inc(p);
         p^ := centerx - arc^;           inc(p);
         p^ := centerx + arc^ + fudgex;  inc(p);
         p^ := EndMarker;                inc(p);
      end;
   end;
   if save2 = p then            // No actual data?
      save1^ := sizeof(RegionHdr) // Re-size it then.
   else begin
      p^ := EndMarker;          // End-of-region marker.
      inc(p);
      save1^ := PByte(p) - PByte(save1); // Plug in real size.
   end;
end;


//
// doOval draws ovals with the specified drawing verb (via regions).
//
procedure doOval(r : Rect; verb : Action; pat : Pattern);
var
   p1 : ^smallInt;              // For creating arc points in.
   p2 : ^smallInt;              // For creating dummy region in.
   size1, size2 : longInt;
   dummyRgn : RgnPtr;
begin
   if EmptyRect(r) then
      exit;
   size1 := (1 + r.bottom - r.top) * sizeof(smallInt); // Number of arc
   p1 := GetMem(size1);         // points (maximum one per scanline).
   iBuildArc(r, p1);
   size2 := (size1 * 4) + sizeof(RegionHdr) + sizeof(smallInt); // Size of
   p2 := GetMem(size2);         // dummy region.
   iBuildORgn(r, p1, p2);
   dummyRgn := RgnPtr(p2);
   case verb of
     Paint:  PaintRgn(@dummyRgn);
     Frame:  FrameRgn(@dummyRgn);
     Erase:  EraseRgn(@dummyRgn);
     Invert: InvertRgn(@dummyRgn);
     Fill:   FillRgn(@dummyRgn, pat);
   end;
   FreeMem(p2);
   FreeMem(p1);
end;


//
// FrameOval draws an outline just inside the oval that fits into the specified
// rectangle, using the current grafPort's pen pattern, mode, and size.  The
// The outline is as wide as the pen width and as tall as the pen height.  It's
// drawn with the pnPat, according to the pattern transfer mode specified by
// pnMode.  The pen location is not changed by this procedure.
//
// Note:  For fat frames, We don't do FrameOval the same as the rest of them
//        because FrameRgn won't work with such a complex region (because of
//        the InsetRgn call).  Therefore we make our own framed region before
//        PaintRgn-ing it.
//
procedure FrameOval(r : Rect);
var
   p1 : ^smallInt;              // For creating arc points in.
   p2 : ^smallInt;              // For creating dummy region in.
   p3 : ^smallInt = nil;        // For creating dummy region in.
   size1, size2 : longInt;
   dummyRgn1, dummyRgn2 : RgnPtr;
   dummyRgn : RgnHandle;
   scratch : Rect;
   tp : GrafPtr;
begin
   tp := thePort;
   if EmptyRect(r) then
      exit;
   if (tp^.pnSize.h <= 2) or (tp^.pnSize.v <= 2) then
      doOval(r, Frame, white)
   else begin
      dummyRgn := NewRgn();
      size1 := (1 + r.bottom - r.top) * sizeof(smallInt);
      p1 := GetMem(size1);      // Number of arc points (one per scan).
      iBuildArc(r, p1);
      size2 := (size1 * 4) + sizeof(RegionHdr) + sizeof(smallInt); // Size of
      p2 := GetMem(size2);      // dummy regions.
      iBuildORgn(r, p1, p2);
      dummyRgn1 := RgnPtr(p2);
      if tp^.rgnSave <> nil then // Region accumulation in progress?
         UnionRgn(@dummyRgn1, tp^.rgnSave, tp^.rgnSave);
      if tp^.pnVis >= 0 then begin
         p3 := GetMem(size2);
         scratch := r;
         InsetRect(scratch, tp^.pnSize.h, tp^.pnSize.v);
         iBuildArc(scratch, p1);
         iBuildORgn(scratch, p1, p3);
         dummyRgn2 := RgnPtr(p3);
         DiffRgn(@dummyRgn1, @dummyRgn2, dummyRgn);
         PaintRgn(dummyRgn);
      end;
      if p3 <> nil then FreeMem(p3);
      FreeMem(p2);
      FreeMem(p1);
      DisposeRgn(dummyRgn);
   end;
end;


//
// PaintOval paints an oval just inside the specified rectangle with the
// current grafPort's pen pattern and mode.  The oval is filled with the
// pnPat, according to the pattern transfer mode specified by pnMode.  The
// pen location is not changed by this procedure.
//
procedure PaintOval(r : Rect);
begin
   doOval(r, Paint, white);
end;

//
// EraseOval paints an oval just inside the specified rectangle with
// the current grafPort's background pattern bkPat (in srcCopy mode).
// The grafPort's pnPat and pnMode are ignored; the pen location is
// not changed.
//
procedure EraseOval(r : Rect);
begin
   doOval(r, Erase, white);
end;


//
// InvertOval inverts the pixels enclosed by an oval just inside the specified
// rectangle:  Every white pixel becomes black and every black pixel becomes
// white.  The grafPort's pnPat, pnMode, and bkPat are all ignored; the pen
// location is not changed.
//
procedure InvertOval(r : Rect);
begin
   doOval(r, Invert, white);
end;


//
// FillOval fills an oval just inside the specified rectangle with the
// given pattern (in srcCopy mode).  The grafPort's pnPat, pnMode and
// bkPat are all ignored; the pen location is not changed.
//
procedure FillOval(r : Rect; pat : Pattern);
begin
   doOval(r, Fill, pat);
end;


// Build a line for trimming arc-shaped regions.

procedure lineForArc(angle : smallInt;  // Angle to build line for.
                     inclusive : Boolean; // TRUE to include angle itself in line.
                     r : Rect;          // Bounding rectangle.
                     p1,                // Where to build line control points.
                     p2 : PsmallInt     // Where to build line-shaped region.
                     );
var
   midx, midy, quadrant, ii, jj : smallInt;
   remaindr : smallInt;
   slope : Fixed;
   delta : smallInt;
   line : Rect;
   lineRgn : RgnPtr;
   divmod : longInt;
   p : ^smallInt;
begin
   divmod := DivModS(angle, 90);// Native divide instruction,
   quadrant := LoWord(divmod);  // pick up quotient
   remaindr := HiWord(divmod);  // and remainder.
   line := r;
   midx := (r.right + r.left) >> 1;
   midy := (r.bottom + r.top) >> 1;
   slope := SlopeFromAngle(angle);
   if not odd(quadrant) then    // Figure out proper slope.
      slope := -slope;
   delta := Fix2Int(slope * (r.right - midx));
   line.right := midx + delta;
   line.left := midx - delta;

   if inclusive then
      case quadrant of
        0: OffsetRect(line, -1, -1);
        1: OffsetRect(line,  1, -1);
        2: OffsetRect(line,  1,  1);
        3: OffsetRect(line, -1,  1);
      end;
   iBuildLine(line, not odd(quadrant), p1);
   iBuildLRgn(line, not odd(quadrant), p1, p2);
   lineRgn := RgnPtr(p2);
   lineRgn^.rgnBBox := r;
   if remaindr = 0 then begin   // Handle degenerate line cases.
      case quadrant of
        0: lineRgn^.rgnBBox.right  := midx; // 0 degrees.
        1: lineRgn^.rgnBBox.bottom := midy; // 90 degrees.
        2: lineRgn^.rgnBBox.left   := midx; // 180 degrees.
        3: lineRgn^.rgnBBox.top    := midy; // 270 degrees.
      end;
      lineRgn^.rgnSize := sizeof(RegionHdr); // For purely rectangular rgn.
   end else begin               // Smear the line region out into a wedge
      p := @lineRgn^.rgnData;   // shaped region for trimming purposes.
      if (quadrant = 1) or (quadrant = 2) then begin
         ii := r.right;
         jj := r.left;
         while p^ <> EndMarker do begin // Do each scanline of the region.
            if p[1] < jj then   // Force region into conformance.
               p[1] := jj;
            if p[1] > ii then
               p[1] := ii;
            p[2] := ii;         // Stretch width by resetting the
            p += 4;             // ending X coordinate.
         end;
      end else begin
         ii := r.left;
         jj := r.right;
         while p^ <> EndMarker do begin // Do each scanline of the region.
            if p[2] > jj then   // Force region into conformance.
               p[2] := jj;
            if p[2] < ii then
               p[2] := ii;
            p[1] := ii;         // Stretch width by resetting the
            p += 4;             // starting X coordinate.
         end;
      end;
   end;
end;

//
// doArc draws arcs with the specified drawing verb (via regions).
//
procedure doArc(r : Rect; doFrame : Boolean; verb : Action;
                startAngle, arcAngle : smallInt; pat : Pattern);
var
   p1 : ^smallInt;              // For creating arc points in.
   p2 : ^smallInt;              // For creating dummy oval region in.
   size1, size2 : longInt;      // Sizes of above areas.
   p3 : ^smallInt = nil;        // For creating line points in.
   p4 : ^smallInt = nil;        // For creating dummy line region in.
   p5 : ^smallInt = nil;        // For creating line points in.
   p6 : ^smallInt = nil;        // For creating dummy line region in.
   endangle : smallInt;
   ovlRgn, line1Rgn, line2Rgn : RgnPtr;
   resultRgn, tempRgn : RgnHandle;
   trimmer : Rect;
   quad1, quad2, midx, midy, didlx, didly : smallInt;
   quadsEq : Boolean;
begin
   if EmptyRect(r) then
      exit;
   if arcAngle < 0 then begin   // Fold arguments to positive angles.
      startAngle += arcAngle;
      arcAngle := -arcAngle;
   end;
   startAngle := startAngle mod 360;
   if startAngle < 0 then
      startAngle += 360;
   endangle := (startAngle + arcAngle + 180) mod 360;
                                // Now build the oval region.
   size1 := (1 + r.bottom - r.top) * sizeof(smallInt); // Number
   p1 := GetMem(size1);         // of arc points (maximum one per scanline).
   iBuildArc(r, p1);
   size2 := (size1 * 4) + sizeof(RegionHdr) + sizeof(smallInt); // Size of
   p2 := GetMem(size2);         // dummy region.
   iBuildORgn(r, p1, p2);
   ovlRgn := RgnPtr(p2);
   resultRgn := NewRgn();
   CopyRgn(@ovlRgn, resultRgn);
   if doFrame then begin
      tempRgn := NewRgn();
      CopyRgn(resultRgn, tempRgn);
      InsetRgn(tempRgn, thePort^.pnSize.h, thePort^.pnSize.v);
      DiffRgn(resultRgn, tempRgn, resultRgn);
      DisposeRgn(tempRgn);
   end;

   if arcAngle < 360 then begin // Now build the first line region.
      size1 := (1 + r.bottom - r.top) * sizeof(smallInt);
      p3 := GetMem(size1);      // # of line points (max one per scan).
      size2 := (size1 * 4) + sizeof(RegionHdr) + sizeof(smallInt); // Size of
      p4 := GetMem(size2);      // dummy region.
      p5 := GetMem(size1);      // Ditto for 2nd set of line control pts
      p6 := GetMem(size2);      // and its region.
      lineforarc(startAngle, True, r, p3, p4);
      line1Rgn := RgnPtr(p4);
      lineforarc(endangle, False, r, p5, p6);
      line2Rgn := RgnPtr(p6);

      if arcAngle > 180 then begin
         tempRgn := NewRgn();
         CopyRgn(resultRgn, tempRgn);
         DiffRgn(tempRgn, @line2Rgn, tempRgn);
         DiffRgn(resultRgn, @line1Rgn, resultRgn);
         UnionRgn(tempRgn, resultRgn, resultRgn);
         DisposeRgn(tempRgn);
      end else begin
         DiffRgn(resultRgn, @line1Rgn, resultRgn);
         DiffRgn(resultRgn, @line2Rgn, resultRgn);
         if arcAngle < 90 then begin // Trim off long points that may hang out
            tempRgn := NewRgn();     // as artifacts of especially acute angles.
            quad1 := startAngle div 90; // Find out if we trim
            quad2 := endangle div 90;   // with a quarter-section
            quadsEq := quad1 = quad2;   // or a half-section.
            trimmer := r;
            midx := (trimmer.right + trimmer.left) div 2;
            midy := (trimmer.bottom + trimmer.top) div 2;
            didlx := midx + ((trimmer.right - trimmer.left) and 1);
            didly := midy + ((trimmer.bottom - trimmer.top) and 1);
            case quad1 of
              0: begin
                    trimmer.left := midx;
                    if quadsEq then
                       trimmer.bottom := didly;
                 end;
              1: begin
                    trimmer.top := midy;
                    if quadsEq then
                       trimmer.left := midx;
                 end;
              2: begin
                    trimmer.right := didlx;
                    if quadsEq then
                       trimmer.top := midy;
                 end;
              3: begin
                    trimmer.bottom := didly;
                    if quadsEq then
                       trimmer.right := didlx;
                 end;
            end;
            RectRgn(tempRgn, trimmer);
            SectRgn(resultRgn, tempRgn, resultRgn);
            DisposeRgn(tempRgn);
         end;
      end;
   end;
   case verb of
     Paint:  PaintRgn(resultRgn);
     Frame:  FrameRgn(resultRgn);
     Erase:  EraseRgn(resultRgn);
     Invert: InvertRgn(resultRgn);
     Fill:   FillRgn(resultRgn, pat);
   end;
   if p6 <> nil then FreeMem(p6);
   if p5 <> nil then FreeMem(p5);
   if p4 <> nil then FreeMem(p4);
   if p3 <> nil then FreeMem(p3);
   DisposeRgn(resultRgn);
   FreeMem(p2);
   FreeMem(p1);
end;


//
// FrameArc draws an arc of the oval that fits inside the specified rectangle,
// using the current grafPort's pen pattern, mode, and size.  StartAngle
// indicates where the arc begins and is treated MOD 360.  ArcAngle defines the
// extent of the arc.  The angles are given in positive or negative degrees; a
// positive angle goes clockwise, while a negative angle goes counterclockwise.
// Zero degrees is at 12 o'clock high, 90 (or -270) is at 3 o'clock, 180 (or
// -180) is at 6 o'clock, and 270 (or -90) is at 9 o'clock.  Other angles are
// measured relative to the enclosing rectangle:  A line from the center of the
// rectangle through its top right corner is at 45 degrees, even if the
// rectangle isn't square; a line through the bottom right corner is at 135
// degrees, and so on.
//
// The arc is as wide as the pen width and as tall as the pen height.  It's
// drawn with the pnPat, according to the pattern transfer mode specified by
// pnMode.  The pen location is not changed by this procedure.
//
procedure FrameArc(r : Rect; startAngle, arcAngle : smallInt);
begin
   doArc(r, True, Paint, startAngle, arcAngle, white);
end;


//
// FrameWedge is the same as FrameArc, but it also frames the rest of the wedge
// of the oval using the current grafPort's pen pattern, mode, and size,
// according to the pattern transfer mode specified by pnMode.  The pen
// location is not changed by this procedure.
//
procedure FrameWedge(r : Rect; startAngle, arcAngle : smallInt);
begin
   doArc(r, False, Frame, startAngle, arcAngle, white);
end;


//
// PaintArc paints a wedge of the oval just inside the specified rectangle
// with the current grafPort's pen pattern and mode.  StartAngle and arcAngle
// define the arc of the wedge as in FrameArc.  The wedge is filled with the
// pnPat, according to the pattern transfer mode specified by pnMode.  The
// pen location is not changed by this procedure.
//
procedure PaintArc(r : Rect; startAngle, arcAngle : smallInt);
begin
   doArc(r, False, Paint, startAngle, arcAngle, white);
end;

//
// EraseArc paints a wedge of the oval just inside the specified rectangle
// with the current grafPort's background pattern bkPat (in srcCopy mode).
// StartAngle and arcAngle define the arc of the wedge as in FrameArc.  The
// grafPort's pnPat and pnMode are ignored; the pen location is not changed.
//
procedure EraseArc(r : Rect; startAngle, arcAngle : smallInt);
begin
   doArc(r, False, Erase, startAngle, arcAngle, white);
end;


//
// InvertArc inverts the pixels enclosed by a wedge of the oval just inside
// the specified rectangle:  Every white pixel becomes black and every black
// pixel becomes white.  StartAngle and arcAngle define the arc of the wedge
// as in FrameArc.  The grafPort's pnPat, pnMode, and bkPat are all ignored;
// the pen location is not changed.
//
procedure InvertArc(r : Rect; startAngle, arcAngle : smallInt);
begin
   doArc(r, False, Invert, startAngle, arcAngle, white);
end;


//
// FillArc fills a wedge of the oval just inside the specified rectangle with
// the given pattern (in srcCopy mode).  StartAngle and arcAngle define the
// arc of the wedge as in FrameArc.  The grafPort's pnPat, pnMode and bkPat
// are all ignored; the pen location is not changed.
//
procedure FillArc(r : Rect; startAngle, arcAngle : smallInt; pat : Pattern);
begin
   doArc(r, False, Fill, startAngle, arcAngle, pat);
end;

//
// BuildRRgn builds a fake region data structure that describes a rounded-
// corner rectangle.  It's used by the xxRoundRect drawing routines.
//
procedure BuildRRRgn(r,         // What to make a rounded region from.
                    ovlrect : Rect; // The oval's rectangle.
                    arc,        // The arc control points.
                    p : PsmallInt); // Where to make it.
var
   ii, jj, leftx, rightx, nullpt : smallInt;
   lastpt, save1, save2 : ^smallInt;
begin
   save1 := p;
   inc(p);                      // Skip over the size field.
   PRect(p)^ := r;              // Dummy region's bounding box.
   p += 4;
   save2 := p;
   jj := ovlrect.bottom div 2;
   ii := (r.bottom - r.top) div 2;
   if jj > ii then              // Make sure oval's not bigger than the
      jj := ii;                 // rectangle we're trimming.
   nullpt := EndMarker;         // So we get the first region segment.
   lastpt := @nullpt;
   leftx := ovlrect.right div 2;
   rightx := r.right - leftx;
   leftx += r.left;
   for ii := 0 to jj - 1 do begin // Top half of rect.
      if arc^ <> lastpt^ then begin // Don't output unnecessary region segments.
         lastpt := arc;
         p^ := r.top + ii;    inc(p);
         p^ := leftx - arc^;  inc(p);
         p^ := rightx + arc^; inc(p); inc(arc);
         p^ := EndMarker;     inc(p);
      end else begin
         inc(arc);
      end;
   end;

   for ii := 0 to jj - 1 do begin // Bottom half of rect.
      dec(arc);
      if arc^ <> lastpt^ then begin
         lastpt := arc;
         p^ := r.bottom - jj + ii; inc(p);
         p^ := leftx - arc^;       inc(p);
         p^ := rightx + arc^;      inc(p);
         p^ := EndMarker;          inc(p);
      end;
   end;
   if save2 = p then            // No actual data?
      save1^ := sizeof(RegionHdr) // Re-size it then.
   else begin
      p^ := EndMarker;          // End-of-region marker.
      inc(p);
      save1^ := PByte(p) - PByte(save1); // Plug in real size.
   end;
end;


//
// doRRect draws rounded rects with the specified drawing verb (via regions).
//
procedure doRRect(r, ovlrect : Rect; verb : Action; pat : Pattern);
var
   p1 : ^smallInt;              // For creating arc points in.
   p2 : ^smallInt;              // For creating dummy region in.
   size1, size2 : longInt;
   dummyRgn : RgnPtr;
   tp : GrafPtr;
begin
   tp := thePort;
   if tp^.pnVis < 0 then
      exit;
   if EmptyRect(r) then
      exit;
   size1 := (1 + ovlrect.bottom) * sizeof(smallInt); // Number of arc
   p1 := GetMem(size1);         // points (maximum one per scanline).
   iBuildArc(ovlrect, p1);
   size2 := (size1 * 4) + sizeof(RegionHdr) + sizeof(smallInt); // Size of
   p2 := GetMem(size2);         // dummy region. */
   BuildRRRgn(r, ovlrect, p1, p2);
   dummyRgn := RgnPtr(p2);
   case verb of
     Paint:  PaintRgn(@dummyRgn);
     Frame:  FrameRgn(@dummyRgn);
     Erase:  EraseRgn(@dummyRgn);
     Invert: InvertRgn(@dummyRgn);
     Fill:   FillRgn(@dummyRgn, pat);
   end;
   FreeMem(p2);
   FreeMem(p1);
end;


//
// FrameRoundRect draws an outline just inside the specified rounded-corner
// rectangle, using the current grafPort's pen pattern, mode, and size.
// OvalWidth and ovalHeight specify the diameters of curvature for the corners.
// The outline is as wide as the pen width and as tall as the pen height.  It's
// drawn with the pnPat, according to the pattern transfer mode specified by
// pnMode.  The pen location is not changed by this procedure.
//
procedure FrameRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
var
   ovlRect : Rect;
begin
   ovlRect.left := 0;
   ovlRect.top := 0;
   ovlRect.right := ovalWidth;
   ovlRect.bottom := ovalHeight;
   doRRect(r, ovlRect, Frame, white);
end;


//
// PaintRoundRect paints the specified rounded-corner rectangle with the
// current grafPort's pen pattern and mode.  OvalWidth and ovalHeight specify
// the diameters of curvature for the corners.
//
// The rounded-corner rectangle is filled with the pnPat, according to the
// pattern transfer mode specified by pnMode.  The pen location is not changed
// by this procedure.
//
procedure PaintRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
var
   ovlRect : Rect;
begin
   ovlRect.left := 0;
   ovlRect.top := 0;
   ovlRect.right := ovalWidth;
   ovlRect.bottom := ovalHeight;
   doRRect(r, ovlRect, Paint, white);
end;


//
// EraseRoundRect paints the specified rounded-corner rectangle with the
// current grafPort's background pattern bkPat (in srcCopy mode).
//
// OvalWidth and ovalHeight specify the diameters of curvature for the
// corners.  The grafPort's pnPat and pnMode are ignored; the pen location
// is not changed.
//
procedure EraseRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
var
   ovlRect : Rect;
begin
   ovlRect.left := 0;
   ovlRect.top := 0;
   ovlRect.right := ovalWidth;
   ovlRect.bottom := ovalHeight;
   doRRect(r, ovlRect, Erase, white);
end;


//
// InvertRoundRect inverts the pixels enclosed by the specified rounded-corner
// rectangle:  Every white pixel becomes black and every black pixel becomes
// white.  OvalWidth and ovalHeight specify the diameters of curvature for the
// corners.  The grafPort's pnPat, pnMode, and bkPat are all ignored; the pen
// location is not changed.
//
procedure InvertRoundRect(r : Rect; ovalWidth, ovalHeight : smallInt);
var
   ovlRect : Rect;
begin
   ovlRect.left := 0;
   ovlRect.top := 0;
   ovlRect.right := ovalWidth;
   ovlRect.bottom := ovalHeight;
   doRRect(r, ovlRect, Invert, white);
end;


//
// FillRoundRect fills the specified rounded-corner rectangle with the
// given pattern (in srcCopy mode).  OvalWidth and ovalHeight specify the
// diameters of curvature for the corners.  The grafPort's pnPat, pnMode and
// bkPat are all ignored; the pen location is not changed.
//
procedure FillRoundRect(r : Rect; ovalWidth,ovalHeight : smallInt; pat: pattern);
var
   ovlRect : Rect;
begin
   ovlRect.left := 0;
   ovlRect.top := 0;
   ovlRect.right := ovalWidth;
   ovlRect.bottom := ovalHeight;
   doRRect(r, ovlRect, Fill, pat);
end;

initialization

end.
