//===========================================================================
//
// Save a bitmap to a MacPaint file, for documentation purposes.  Bitmaps
// larger than 576x720 will be truncated to that size.  A file extension
// of .mac can be recognized by ImageMagick as a MacPaint file.
//
//===========================================================================

unit clunksave;

{$mode objfpc}{$H+}

interface

   uses Clunkdraw;

   procedure PackBits(var src, dest : PByte; srccount : smallInt);
   procedure UnpackBits(var src, dest : PByte; destcount : smallInt);
   procedure SaveToPaintFile(sbits : BitMap; fName : string);
   procedure SaveToPBMFile(sbits : BitMap; fName : string);

implementation

//
// Bit packing stuff.  See Macintosh Technical Note #171 for more information.
//
procedure PackBits(var src, dest : PByte; srccount : smallInt);
var ii, packcount, nonpackcount : smallInt;
    p : PByte;
    match : Byte;
begin
   while (srccount > 127) do begin // Too much for a single run at it,
      PackBits(src, dest, 127);    // do it in pieces.
      srccount -= 127;
   end;
   p := src;
   nonpackcount := 0;
   packcount := 0;
   while srccount > 0 do begin  // While there is data to pack.
      ii := srccount;
      dec(srccount);
      if (p[0] = p[1]) and (p[0] = p[2]) and (ii > 2) then begin // Run of >= 3?
         if nonpackcount <> 0 then begin // Any non-packable accumulated yet?
            dest^ := nonpackcount - 1;   // Stuff in a count byte
            inc(dest);
            while nonpackcount > 0 do begin // and blow out non-packed data.
               dec(nonpackcount);
               dest^ := src^;   // (To get all this out of the way of
               inc(dest);       //  the packed data first.)
               inc(src);
            end;
         end;
         match := src^;
         repeat                 // Find length of run.
            inc(p);
            dec(ii);
            if (p[-1] <> match) or (ii < 0) then break;
            inc(packcount);
         until False;
         dest^ := 1 - packcount; // Stuff in a repeat byte.
         inc(dest);
         srccount -= packcount - 1; // Decrement count by what's packed.
         nonpackcount := 0;
         packcount := 0;
         dest^ := match;     // Copy the byte to repeat.
         inc(dest);
         dec(p);
         src := p;           // Move to uncompressed data.
      end else begin         // Not a run.
         inc(nonpackcount);  // Increment the count of non-packeds.
         inc(p);
      end;
   end;
   if nonpackcount <> 0 then begin // Any leftover non-packable?
      dest^ := nonpackcount - 1; // Stuff in a count byte
      inc(dest);
      while nonpackcount > 0 do begin // and blow out non-packed data.
         dec(nonpackcount);
         dest^ := src^;
         inc(dest);
         inc(src);
      end;
   end;
end;

procedure UnpackBits(var src, dest : PByte; destcount : smallInt);
var
   repcnt : smallInt;
   copy : Byte;
begin
    while destcount > 127 do begin // Too much for a single run at it,
       UnpackBits(src, dest, 127); // do it in pieces.
       destcount -= 127;
    end;
    while destcount > 0 do begin   // While there is packed data.
       if Int8(src^) < 0 then begin // We have a repeat count.
          repcnt := 1 - Int8(src^);
          inc(src);
          destcount -= repcnt;
          if destcount >= 0 then begin // Sanity check.
             copy := src^;
             inc(src);
             while repcnt > 0 do begin
                dec(repcnt);
                dest^ := copy;
                inc(dest);
             end;
          end;
        end else begin          // We have a raw piece.
           repcnt := 1 + src^;
           inc(src);
           destcount -= repcnt;
           if destcount >= 0 then begin // Sanity check.
                while repcnt > 0 do begin
                   dec(repcnt);
                   dest^ := src^;
                   inc(dest);
                   inc(src);
                end;
           end;
        end;
    end;
end;

// Save the given bitmap to a MacPaint (576x720) file.
// Too much data and it will be trimmed, too little and it will be padded.
// File format taken from Macintosh Technical Note #24.
//
// More modern systems might be able to read these if you use a .mac extension.
//
procedure SaveToPaintFile(sbits : BitMap; fName : string);
const width  = 576;
      height = 720;
var header : packed array [0..511] of Byte;
    iline : packed array[1..width div 8] of Byte; // 576 pixel's worth.
    oline : packed array[0..width div 8] of Byte; // 576 pixel's worth, plus 1.
    x, y, ii : integer;
    s, d : PByte;
    f : File;
begin
   try
      AssignFile(f, fName);
   except
      exit;
   end;
   try
      ReWrite(f, 1);    // Byte orientation.
      FillByte(header, sizeof(header), 0);
      BlockWrite(f, header, sizeof(header)); // The ver/pat header.  Empty.

      for y := 1 to height do begin
{$IF 1}
         FillByte(iline, sizeof(iline), 0);
         s := sbits.baseAddr + ((y-1) * sbits.rowBytes);
         for x:= 1 to sizeof(iline) do begin
            if (y > RectHeight(sbits.bounds)) or (x > sbits.rowBytes) then
               iline[x] := 0
            else
               iline[x] := s^;
            inc(s);
         end;
         s := @iline;
         d := @oline;
         PackBits(s, d, sizeof(iline)); // Pack it down.  (Worst-case is +1.)
         ii := d - PByte(@oline);
{$ELSE} //Uncompressed
         FillByte(oline, sizeof(oline), 0);
         s := sbits.baseAddr + ((y-1) * sbits.rowBytes);
         d := @oline;
         d^ := 71;
         inc(d);
         for x:= 1 to sizeof(iline) do begin
            if (y > RectHeight(sbits.bounds)) or (x > sbits.rowBytes) then
               d^ := 0
            else
               d^ := s^;
            inc(s);
            inc(d);
         end;
         ii := sizeof(oline);
{$ENDIF}
         BlockWrite(f, oline, ii);
      end;
   finally
      Close(f);
   end;
end;

// Imagemagick sucks, and can't seem to eat MacPaint files reliably.
// Save the given bitmap to a PBM interchange file.  Bulky, but safe?
//
procedure SaveToPBMFile(sbits : BitMap; fName : string);
var x, y : integer;
    s : PByte;
    f : TextFile;
begin
   try
      AssignFile(f, fName);
   except
      exit;
   end;
   try
      ReWrite(f);
      WriteLn(f, 'P1');
      WriteLn(f, sbits.rowBytes * 8, ' ', RectHeight(sbits.bounds));
      for y := 1 to RectHeight(sbits.bounds) do begin
         WriteLn(f, '# Line ', y);
         s := sbits.baseAddr + ((y-1) * sbits.rowBytes);
         for x:= 1 to sbits.rowBytes do begin
            Write(f, BinStr(s^, 8));
            inc(s);
         end;
         WriteLn(f);
      end;
   finally
      CloseFile(f);
   end;
end;

initialization

end.
